(* TODO:
 *)


(* Note added due to LGPL terms.

This file was edited by Valerio Bigiani, AKA The Bigg, starting from
6 November 2005. All changes for this file are listed in
diffs/src.tp.ml.diff file, as the output of a diff -p -E -b command.

It was originally taken from Westley Weimer's WeiDU 185. *)

(* Talk-Patcher / Installer *)
open Util
open Diff

let always_yes = ref false
let always_uninstall = ref false
let sometimes_reinstall = ref false
let ask_all = ref false
let forced_language = ref ( 0-1 )
let force_install_these = ref []
let force_uninstall_these = ref []
let specified_specific_components = ref false
let skip_at_view = ref false
let quick_log = ref false
let debug_boiic = ref false
let debug_change = ref false
(* let omg_broken = ref false *)

let has_if_eval_bug = ref true
let continue_on_error = ref false
let debug_pe = ref false

type tp_flag =
| Auto_Tra of string
| TP_No_If_Eval of unit
| Menu_Style of string
| Ask_Every_Component
| Always of tp_action list
| Define_Action_Macro of string * tp_local_declaration list * tp_action list
| Define_Patch_Macro of string * tp_local_declaration list * tp_patch list
| Allow_Missing of string list
| Script_Style of Load.script_style
| Load_Macro of string list

and tp_file = {
  mutable tp_filename  : string ;
  backup    : string ;
  author    : string ;
  flags     : tp_flag list ;
  languages : tp_lang list ;
  module_list : tp_mod list ;
}

and tp_lang = {
  lang_name : string ;
  lang_dir_name : string ;
  lang_tra_files : string list ;
}

and tp_mod = {
  mod_name : Dlg.tlk_string ;
  mod_parts : tp_action list;
  mod_flags : tp_mod_flag list ;
}

and tp_mod_flag =
| TPM_Deprecated of Dlg.tlk_string (* should be uninstalled when encountered *)
| TPM_RequireComponent of string * int * Dlg.tlk_string
| TPM_ForbidComponent of string * int * Dlg.tlk_string
| TPM_RequirePredicate of tp_patchexp * Dlg.tlk_string
| TPM_SubComponents of Dlg.tlk_string * tp_patchexp * bool (* is forced? *)
| TPM_Designated of int
| TPM_NotInLog
| TPM_InstallByDefault
| TPM_Group of Dlg.tlk_string

and tp_copy_args = {
  copy_get_existing    : bool ;  (* get from biffs? *)
  copy_use_regexp      : bool ;
  copy_use_glob        : bool ;
  copy_file_list       : ( string * string ) list ; (* (source,dest) list *)
  copy_patch_list      : tp_patch list ;
  copy_constraint_list : tp_constraint list ;
  copy_backup          : bool ; (* normally TRUE *)
  copy_at_end          : bool ;
    (* write all of the output *after* doing all of the processing for
     * all souce files *)
  copy_save_inlined    : bool ; (* the result is not saved to the HD, but stored in the inlined Hashtable *)
}

and tp_copy_large_args = {
  copy_large_use_glob       : bool ;
  copy_large_file_list      : ( string * string ) list ; (* (source,dest) list *)
  copy_large_backup         : bool ; (* normally TRUE *)
}

and store_args = {
  overwrite_store_item  : bool ; (* normally TRUE *)
}

and tp_action =
  | TP_ActionBashFor of ((string * (bool option) * string) list) * (tp_action list)
  | TP_Action_For_Each of tp_pe_string * string list * tp_action list
  | TP_Action_ReadLN of tp_pe_string
  | TP_ClearMemory
  | TP_Clear_Ids_Map
  | TP_CopyAllGamFiles of tp_patch list * tp_constraint list
  | TP_Copy   of tp_copy_args
  | TP_CopyLarge of tp_copy_large_args
  | TP_CopyRandom of (string list list) * (tp_patch list) * (tp_constraint list)
  | TP_RandomSeed of tp_patchexp
  | TP_Compile of bool * (string list) * (string list) (* eval, DLG, TRA *)
  | TP_Launch_Action_Macro of string
  | TP_Reinclude of string list
  | TP_Include of string list
  | TP_Load_Tra of string list
  | TP_Inlined_File of (string * string)
  | TP_Define_Action_Macro of string * tp_local_declaration list * tp_action list
  | TP_Define_Patch_Macro of string * tp_local_declaration list * tp_patch list
  | TP_Biff of string * ((string * (bool option) * string) list)
  | TP_Mkdir of string list
  | TP_Outer_For of (tp_patch list) * tp_patchexp * (tp_patch list) * (tp_action list)
  | TP_Outer_Inner_Buff of string * (tp_patch list)
  | TP_Outer_Inner_Buff_Save of string * string * (tp_patch list)
  | TP_Outer_Set of tp_pe_string * tp_patchexp
  | TP_Outer_Sprint of tp_pe_string * tp_pe_tlk_string
  | TP_Outer_Text_Sprint of tp_pe_string * tp_pe_string
  | TP_Outer_While of tp_patchexp * (tp_action list)
  | TP_Require_File of string * (Dlg.tlk_string)
  | TP_Forbid_File of string * (Dlg.tlk_string)
  | TP_Append of string * string * (tp_constraint list) * bool
        (* When bool is true, load the file as bif or override; otherwise, it's loaded
         * as a normal file.
         *)
  | TP_Append_Col of string * string * int * (tp_constraint list)
        (* prepend to the string int blank entries (EG for ADD_KIT) *)
  | TP_Set_Col of string * (string list) * int
  | TP_Extend_Top of bool * string * string * (tp_patch list) * (string list)
  | TP_Extend_Bottom of bool * string * string * (tp_patch list) * (string list)
  | TP_At_Exit of string
  | TP_At_Interactive_Exit of string
  | TP_At_Uninstall of string
  | TP_At_Interactive_Uninstall of string
  | TP_At_Now of string
  | TP_At_Interactive_Now of string
  | TP_Add_Kit of tp_add_kit
  | TP_Add_Music of tp_add_music
  | TP_Add_Projectile of tp_add_projectile
  | TP_Add_Spell of string * tp_patchexp * tp_patchexp * string * (tp_patch list) * (tp_constraint list)
  | TP_Silent
  | TP_Verbose
  | TP_String_Set of ((string * Dlg.tlk_string) list) * (string option)
  | TP_String_Set_Evaluate of ((tp_patchexp * Dlg.tlk_string) list) * (string option)
  | TP_String_Set_Range of int * int * string
  | TP_Fail of Dlg.tlk_string
  | TP_Print of Dlg.tlk_string
  | TP_If of tp_patchexp * (tp_action list) * (tp_action list)
  | TP_Uninstall_Now of string * tp_patchexp

(*
and predicate =
  | Pred_And of predicate * predicate
  | Pred_Or of predicate * predicate
  | Pred_Not of predicate
  | Pred_True
  | Pred_Expr of tp_patchexp 
  *)

and tp_add_kit = {
  kit_name : string ;
  clasweap : string ;
  weapprof : string ;
  abclasrq : string ;
  abclsmod : string ;
  abdcdsrq : string ;
  abdcscrq : string ;
  dualclas : string ;
  alignmnt : string ;
  ability_file : string ; 
  include_in : string ;
  lower : Dlg.tlk_string ;
  mixed : Dlg.tlk_string ;
  help : Dlg.tlk_string ;
  unused_class : string ;
  tob_start : string list ;
  tob_abbr : string ; 
} 

and tp_add_map_note = {
  xcoord : int ;
  ycoord : int ;
  mstring : Dlg.tlk_string ;
  colour : string ;
}

and tp_add_cre_item = {
  item_name : string ;
  i_charge1 : int ;
  i_charge2 : int ;
  i_charge3 : int ;
  i_flags : string ;
  item_slot : string ;
  equip : bool ; (* Normally FALSE *)
  twohanded_weapon : bool ; (* Normally TRUE *)
}

and tp_add_music = {
  music_name : string ;
  music_file : string ;
}

and tp_add_projectile = {
  pro_file : string ;
}

and tp_local_declaration =
  | TP_LocalSet of tp_pe_string * tp_patchexp
  | TP_LocalSprint of tp_pe_string * tp_pe_tlk_string
  | TP_LocalTextSprint of tp_pe_string * tp_pe_string

and tp_patch =
  | TP_PatchBashFor of ((string * (bool option) * string) list) * (tp_patch list)
  | TP_PatchForEach of tp_pe_string * string list * tp_patch list
  | TP_PatchStrRef of tp_patchexp * Dlg.tlk_string (* offset + text *)
  | TP_PatchStrRefEvaluated of tp_patchexp * string
  | TP_PatchString of (bool option) * (bool option) * string * Dlg.tlk_string (* regexp + text *)
  | TP_PatchStringTextually of (bool option) * (bool option) * string * string * (tp_patchexp option) (* regexp + text *)
  | TP_PatchStringEvaluate of (bool option) * string * (tp_patch list) * string (* see below *)
  | TP_PatchReplaceBCSBlock of string * string (* old + new *)
  | TP_PatchReplaceBCSBlockRE of string * string (* old + new *)
  | TP_PatchApplyBCSPatch of string (* patch *) * (string option) (* copyover *)
  | TP_PatchByte of tp_patchexp * tp_patchexp
  | TP_PatchShort of tp_patchexp * tp_patchexp
  | TP_PatchLong of tp_patchexp * tp_patchexp
  | TP_PatchGetStrRef of tp_patchexp * string
  | TP_PatchReadAscii of tp_patchexp * string * (string option) * (tp_patchexp option)
  | TP_PatchReadStrRef of tp_patchexp * string * (string option)
  | TP_PatchReadByte of tp_patchexp * string * (tp_patchexp option)
  | TP_PatchReadSByte of tp_patchexp * string * (tp_patchexp option)
  | TP_PatchReadSShort of tp_patchexp * string * (tp_patchexp option)
  | TP_PatchReadShort of tp_patchexp * string * (tp_patchexp option)
  | TP_PatchReadLong of tp_patchexp * string * (tp_patchexp option)
  | TP_PatchReadSLong of tp_patchexp * string * (tp_patchexp option)
  | TP_PatchReadLN of tp_pe_string
  | TP_Launch_Patch_Macro of string
  | TP_Add_Known_Spell of string * int * string
  | TP_Read2DA of tp_patchexp * tp_patchexp * tp_patchexp * string
  | TP_Read2DANow of string * tp_patchexp
  | TP_Read2DAFormer of string * tp_patchexp * tp_patchexp * string
  | TP_Get2DARows of tp_patchexp * string
  | TP_CountRegexpInstances of (bool option) * (bool option) * string * string
  | TP_DescribeItem of string
  | TP_PatchPrint of Dlg.tlk_string
  | TP_PatchSprint of tp_pe_string * tp_pe_tlk_string
  | TP_PatchTextSprint of tp_pe_string * tp_pe_string
  | TP_PatchToLower of tp_pe_string
  | TP_PatchToUpper of tp_pe_string
  | TP_PatchSnprint of tp_patchexp * tp_pe_string * tp_pe_tlk_string
  | TP_PatchASCII of
      tp_patchexp (* where? *)
    * string (* what? *)
    * bool (* evaluate? *)
    * (int option) (* minimum size? *)
  | TP_PatchInsertBytes of tp_patchexp * tp_patchexp
  | TP_PatchDeleteBytes of tp_patchexp * tp_patchexp
  | TP_PatchSet of tp_pe_string * tp_patchexp
  | TP_PatchSetIdsSymOfInt of string * string * tp_patchexp
  | TP_PatchWhile of tp_patchexp * (tp_patch list)
  | TP_PatchFor of
    (tp_patch list) * tp_patchexp * (tp_patch list) * (tp_patch list)
  | TP_PatchIf of tp_patchexp * (tp_patch list) * (tp_patch list)
  | TP_PatchRandomSeed of tp_patchexp
  | TP_PatchInnerAction of tp_action list
  | TP_PatchInnerBuff of string * (tp_patch list)
  | TP_PatchInnerBuffFile of string * (tp_patch list)
  | TP_PatchInnerBuffSave of string * string * (tp_patch list)
  | TP_PatchInsert2DARow of tp_patchexp * tp_patchexp * tp_patchexp
  | TP_Patch2DA of tp_patchexp * tp_patchexp * tp_patchexp * tp_patchexp
  | TP_Patch2DALater of string * tp_patchexp * tp_patchexp * tp_patchexp
  | TP_Patch2DANow of string * tp_patchexp
  | TP_PatchSilent
  | TP_PatchVerbose


  | TP_Add_Map_Note of tp_add_map_note
  | TP_Patch_Gam of string * string * int * int
  | TP_Add_Cre_Item of tp_add_cre_item
  | TP_Replace_Cre_Item of tp_add_cre_item
  | TP_Add_S_Item of store_args * string * ((string * int) list) * int * int * int * string * int * (string option)
  | TP_Remove_Cre_Item of string list
  | TP_Remove_Known_Spell of string list
  | TP_Remove_Memorized_Spell of string list
  | TP_Remove_Store_Item of string list
  | TP_PatchWriteFile of tp_patchexp * string * bool (* where, what, insert? *)
  | TP_PatchAppendFile of string * bool (* evaluate the file to be appended? *)
  | TP_CompileBAFtoBCS
  | TP_CompileBCStoBAF
  | TP_CompileDtoDLG
  | TP_CompileDLGtoD
  | TP_EvaluateBuffer


and tp_constraint =
  | TP_Contains of string
  | TP_NotContains of string
  | TP_IfSizeIs of int
  | TP_Eval of tp_patchexp
  | TP_ButOnlyIfItChanges

and tp_pe_string =
  | PE_LiteralString of string
  | PE_Evaluate of tp_pe_string
  | PE_Lowercase of tp_pe_string
  | PE_Uppercase of tp_pe_string

and tp_pe_tlk_string =
  | PE_Tlk of Dlg.tlk_string
  | PE_Pe of tp_pe_string

and tp_patchexp =
  | Pred_True
  | PE_String of tp_pe_string
  | PE_StringEqual of tp_pe_string * tp_pe_string * bool * bool (* ignore-case? * returns bool vs. 1,-1,0 *)
  | PE_StringRegexp of tp_pe_string * tp_pe_string * bool (* match exactly?  *)
  | PE_Not of tp_patchexp
  | PE_Add of tp_patchexp * tp_patchexp
  | PE_Sub of tp_patchexp * tp_patchexp
  | PE_Mul of tp_patchexp * tp_patchexp
  | PE_Div of tp_patchexp * tp_patchexp
  | PE_Exp of tp_patchexp * tp_patchexp * tp_patchexp
  | PE_Equal of tp_patchexp * tp_patchexp
  | PE_And of tp_patchexp * tp_patchexp
  | PE_Or of tp_patchexp * tp_patchexp
  | PE_GT of tp_patchexp * tp_patchexp
  | PE_GTE of tp_patchexp * tp_patchexp
  | PE_LT of tp_patchexp * tp_patchexp
  | PE_LTE of tp_patchexp * tp_patchexp

  | PE_BAND of tp_patchexp * tp_patchexp
  | PE_BOR of tp_patchexp * tp_patchexp
  | PE_BNOT of tp_patchexp
  | PE_BXOR of tp_patchexp * tp_patchexp
  | PE_BLSL of tp_patchexp * tp_patchexp
  | PE_BLSR of tp_patchexp * tp_patchexp
  | PE_BASR of tp_patchexp * tp_patchexp

  | PE_Random of tp_patchexp * tp_patchexp
  | PE_FileContainsEvaluated of tp_pe_string * tp_pe_string

  | PE_If of tp_patchexp * tp_patchexp * tp_patchexp
  | PE_ModIsInstalled of string * int
  | PE_GameIs of string
  | PE_VariableIsSet of tp_pe_string
  | PE_IdsOfSymbol of string * string
  | PE_StateWhichSays of (Dlg.tlk_string option) * ((tp_patchexp * string) option) * string
  | PE_IsAnInt of tp_pe_string

  | Pred_File_MD5 of tp_pe_string * tp_pe_string
  | Pred_File_Exists of tp_pe_string
  | Pred_File_Is_In_Compressed_Bif of tp_pe_string
  | Pred_File_Exists_In_Game of tp_pe_string
  | Pred_File_Size of tp_pe_string * int
  | Pred_File_Contains of tp_pe_string * tp_pe_string

let strings_to_print_at_exit : string list ref = ref []

let rec get_menu_style fl = match fl with
    [] -> 0
  | Menu_Style(i) :: tl -> int_of_string i
  | hd :: tl -> get_menu_style tl

let rec pe_to_str pe = match pe with
  | Pred_File_MD5(s,_) -> Printf.sprintf "FILE_MD5 %s" (pe_str_str s)
  | Pred_File_Exists(s) -> Printf.sprintf "FILE_EXISTS %s" (pe_str_str s)
  | Pred_File_Exists_In_Game(s) -> Printf.sprintf "FILE_EXISTS_IN_GAME %s"
    (pe_str_str s)
  | Pred_File_Size(s,i) -> Printf.sprintf "FILE_SIZE %s %d" (pe_str_str s) i
  | Pred_File_Contains(s,i) -> Printf.sprintf "FILE_CONTAINS %s %s"
    (pe_str_str s) (pe_str_str i)
  | Pred_File_Is_In_Compressed_Bif(a) -> Printf.sprintf "FILE_IS_IN_COMPRESSED_BIF %s" (pe_str_str a)
  | PE_String(s) -> (pe_str_str s)
  | PE_StringEqual(s1,s2,b,c) -> Printf.sprintf "%s %s%s %s"
      (pe_str_str s1) (if c then "STRING_EQUAL" else "STRING_COMPARE") (if b then "_CASE" else "")
      (pe_str_str s2)
  | PE_StringRegexp(s1,s2,b) -> Printf.sprintf "%s %s %s"
      (pe_str_str s1) (if b then "STRING_MATCHES_REGEXP"
        else "STRING_CONTAINS_REGEXP")
      (pe_str_str s2)
  | PE_Not(e) -> Printf.sprintf "NOT (%s)" (pe_to_str e)
  | PE_Add(e1,e2) -> Printf.sprintf "(%s) + (%s)" (pe_to_str e1) (pe_to_str e2)
  | PE_Sub(e1,e2) -> Printf.sprintf "(%s) - (%s)" (pe_to_str e1) (pe_to_str e2)
  | PE_Mul(e1,e2) -> Printf.sprintf "(%s) * (%s)" (pe_to_str e1) (pe_to_str e2)
  | PE_Div(e1,e2) -> Printf.sprintf "(%s) / (%s)" (pe_to_str e1) (pe_to_str e2)
  | PE_Exp(e1,e2,Pred_True) -> Printf.sprintf "(%s) ** (%s)" (pe_to_str e1) (pe_to_str e2)
  | PE_Exp(e1,e2,e3) -> Printf.sprintf "(%s) ** ((%s) (%s))" (pe_to_str e1) (pe_to_str e2) (pe_to_str e3)
  | PE_Equal(e1,e2) -> Printf.sprintf "(%s) = (%s)" (pe_to_str e1) (pe_to_str e2)
  | PE_And(e1,e2) -> Printf.sprintf "(%s) AND (%s)" (pe_to_str e1) (pe_to_str e2)
  | PE_Or(e1,e2) -> Printf.sprintf "(%s) OR (%s)" (pe_to_str e1) (pe_to_str e2)
  | PE_GT(e1,e2) -> Printf.sprintf "(%s) > (%s)" (pe_to_str e1) (pe_to_str e2)
  | PE_GTE(e1,e2) -> Printf.sprintf "(%s) >= (%s)" (pe_to_str e1) (pe_to_str e2)
  | PE_LT(e1,e2) -> Printf.sprintf "(%s) < (%s)" (pe_to_str e1) (pe_to_str e2)
  | PE_LTE(e1,e2) -> Printf.sprintf "(%s) <= (%s)" (pe_to_str e1) (pe_to_str e2)

  | PE_BAND(e1,e2) -> Printf.sprintf "(%s) BAND (%s)" (pe_to_str e1) (pe_to_str e2)
  | PE_BOR(e1,e2) -> Printf.sprintf "(%s) BOR (%s)" (pe_to_str e1) (pe_to_str e2)
  | PE_BXOR(e1,e2) -> Printf.sprintf "(%s) BXOR (%s)" (pe_to_str e1) (pe_to_str e2)
  | PE_BLSL(e1,e2) -> Printf.sprintf "(%s) BLSL (%s)" (pe_to_str e1) (pe_to_str e2)
  | PE_BLSR(e1,e2) -> Printf.sprintf "(%s) BLSR (%s)" (pe_to_str e1) (pe_to_str e2)
  | PE_BASR(e1,e2) -> Printf.sprintf "(%s) BASR (%s)" (pe_to_str e1) (pe_to_str e2)
  | PE_BNOT(e1) -> Printf.sprintf "BNOT (%s)" (pe_to_str e1)
  | PE_GameIs(e1) -> Printf.sprintf "GAME_IS ~%s~" e1
  | PE_ModIsInstalled(e1,e2) -> Printf.sprintf "MOD_IS_INSTALLED ~%s~ #%d" e1 e2
  | Pred_True -> Printf.sprintf "1"

  | PE_Random(e1,e2) -> Printf.sprintf "RANDOM((%s) (%s))" (pe_to_str e1) (pe_to_str e2)

  | PE_FileContainsEvaluated(s1,s2) -> Printf.sprintf "FILE_CONTAINS_EVALUATED(%s %s)" (pe_str_str s1) (pe_str_str s2)

  | PE_If(e1,e2,e3) -> Printf.sprintf "(%s) ? (%s) : (%s)"
    (pe_to_str e1) (pe_to_str e2) (pe_to_str e3)
  
  | PE_VariableIsSet(s) -> Printf.sprintf "VARIABLE_IS_SET %s" (pe_str_str s)
  | PE_IdsOfSymbol(a,b) -> Printf.sprintf "IDS_OF_SYMBOL (%s %s)" a b
  | PE_StateWhichSays(Some(x),None,y) -> Printf.sprintf "STATE_WHICH_SAYS __ FROM %s.DLG"  y
  | PE_StateWhichSays(None,Some(a,b),y) -> Printf.sprintf "STATE_WHICH_SAYS (%s) IN %s FROM %s.DLG" (pe_to_str a) b y
  | PE_StateWhichSays(None,None,_)
  | PE_StateWhichSays(Some(_),Some(_),_) -> Printf.sprintf "INTERNAL ERROR"
  | PE_IsAnInt(x) -> Printf.sprintf "IS_AN_INT %s" (pe_str_str x)

and pe_str_str s = match s with
  | PE_LiteralString(s) -> s
  | PE_Evaluate(s) -> Printf.sprintf "EVALUATE_BUFFER %s" (pe_str_str s)
  | PE_Lowercase(s) -> Printf.sprintf "LOWERCASE %s" (pe_str_str s)
  | PE_Uppercase(s) -> Printf.sprintf "UPPERCASE %s" (pe_str_str s)

let action_to_str a = match a with
  | TP_Copy _ -> "COPY"
  | TP_ClearMemory -> "CLEAR_MEMORY"
  | TP_Clear_Ids_Map -> "CLEAR_IDS_MAP"
  | TP_CopyLarge _ -> "COPY_LARGE"
  | TP_CopyAllGamFiles _ -> "COPY_ALL_GAM_FILES"
  | TP_CopyRandom _ -> "COPY_RANDOM"
  | TP_RandomSeed _ -> "RANDOM_SEED"
  | TP_Compile _ -> "COMPILE"
  | TP_Inlined_File _ -> "<<<<<<<<"
  | TP_Load_Tra _-> "LOAD_TRA"
  | TP_Mkdir _ -> "MKDIR"
  | TP_Forbid_File _ -> "FORBID_FILE"
  | TP_Require_File _ -> "REQUIRE_FILE"
  | TP_Append _ -> "APPEND"
  | TP_Append_Col _ -> "APPEND_COL"
  | TP_Set_Col _ -> "SET_COL"
  | TP_Extend_Top _
  | TP_Extend_Bottom _ -> "EXTEND_*"
  | TP_At_Exit _
  | TP_At_Interactive_Exit _
  | TP_At_Uninstall _
  | TP_At_Interactive_Uninstall _
  | TP_At_Now _
  | TP_At_Interactive_Now _ -> "AT_*"
  | TP_Add_Kit _ -> "ADD_KIT"
  | TP_Add_Music _ -> "ADD_MUSIC"
  | TP_Add_Projectile _ -> "ADD_PROJECTILE"
  | TP_Add_Spell _ -> "ADD_SPELL"
  | TP_String_Set _ -> "STRING_SET"
  | TP_String_Set_Evaluate _ -> "STRING_SET_EVALUATE"
  | TP_String_Set_Range _ -> "STRING_SET_RANGE"
  | TP_Fail _ -> "FAIL"
  | TP_Print _ -> "PATCH_PRINT"
  | TP_If _ -> "PATCH_IF"
  | TP_Uninstall_Now _ -> "UNINSTALL"
  | TP_ActionBashFor _ -> "ACTION_BASH_FOR"
  | TP_Action_For_Each _ -> "ACTION_FOR_EACH"
  | TP_Biff _ -> "BIFF"
  | TP_Outer_For _ -> "OUTER_FOR"
  | TP_Outer_Inner_Buff _ -> "OUTER_INNER_PATCH"
  | TP_Outer_Inner_Buff_Save _ -> "OUTER_INNER_PATCH_SAVE"
  | TP_Outer_Set _ -> "OUTER_SET"
  | TP_Outer_Sprint _ -> "OUTER_SPRINT"
  | TP_Outer_Text_Sprint _ -> "OUTER_TEXT_SPRINT"
  | TP_Outer_While _ -> "OUTER_WHILE"
  | TP_Launch_Action_Macro _ -> "LAUNCH_ACTION_MACRO"
  | TP_Include _ -> "INCLUDE"
  | TP_Reinclude _ -> "REINCLUDE"
  | TP_Define_Action_Macro _ -> "DEFINE_ACTION_MACRO"
  | TP_Define_Patch_Macro _ -> "DEFINE_PATCH_MACRO"
  | TP_Silent -> "SILENT"
  | TP_Verbose -> "VERBOSE"
  | TP_Action_ReadLN _ -> "ACTION_READLN"

(************************************************************************
 * For handling the problem w/ REQUIRE_FILE & akin on bif
 * w/ multiple installs.
 ************************************************************************)
let bigg_file_exists file key =
  let test = ref false in
  if Str.string_match (Str.regexp_case_fold "data.25.*\\.bif") file 0 then
  begin
(*    log_and_print "\nbigg_file_exist special case\n" ; *)
    test := Key.bif_exists_in_key key (Str.global_replace (Str.regexp "[\\/:]")
            Arch.biff_path_separator file)
  end
  else begin
(*    log_and_print "\nbigg_file_exist special case not triggered: %s\n" file ; *)
    test := file_size file >= 0
  end ;
  !test

(************************************************************************
 * Common hashtables.
 ************************************************************************)
let handle_tph_filename_global = ref (fun x -> ([]) )
let handle_tra_filename_global = ref (fun x -> ignore(x) )
let loaded_tph = Hashtbl.create 5
let action_macros = Hashtbl.create 10
let patch_macros = Hashtbl.create 10
let readln_strings = ref []

(************************************************************************
 * Some shorthand for writing macroes.
 ************************************************************************)
let make_lse_internal male_string male_sound female_string female_sound =
  Dlg.Local_String ({ lse_male = male_string ; lse_male_sound = male_sound ;
                    lse_female = female_string ; lse_female_sound = female_sound; })

let get_lse the_string =
  make_lse_internal the_string "" the_string ""

let get_pe_int a_string =
  PE_String(PE_LiteralString(a_string))

let get_pe_string a_string =
  PE_LiteralString(a_string)

let get_pe_tlk_string a_string =
  PE_Pe(get_pe_string a_string)
(************************************************************************
 * For handling lists of modules.
 ************************************************************************)
let get_nth_module tp_file n print_why =
  let lst = tp_file.module_list in
  let last = ref (-1) in
  let rec process lst = match lst with
  | [] -> 
    if print_why then 
      log_or_print "%s's %dth component not found." tp_file.tp_filename n;
    raise Not_found
  | hd :: tl ->
    begin
      List.iter (fun x -> match x with
      | TPM_Designated(i) -> last := pred i
      | _ -> ()) hd.mod_flags ;
      incr last;
      if (!last = n) then hd else process tl
    end
  in
  process lst 

let get_last_module_index tp_file = 
  let last = ref 0 in
  List.fold_left (fun acc elt -> 
    let this_one = 
      let rec process lst = match lst with
      | TPM_Designated(i) :: tl -> i
      | hd :: tl -> process tl 
      | [] -> !last + 1
      in process elt.mod_flags
    in 
    last := this_one ;
    max !last acc
  ) 0 tp_file.module_list


(************************************************************************
 * Set some default strings that can be overwritten. 
 ************************************************************************)
let init_default_strings () = 
  let add i s = 
    Dc.add_trans_strings [i,(Dlg.Local_String({
      lse_male = s; 
      lse_male_sound = s;
      lse_female = s;
      lse_female_sound = s;
    }))]
  in
  add (-1000) "The %TP2_FILE_NAME% mod has" ;
  add (-1001) "distinct optional components.\nTo save time, you can choose what to do with them at a high level rather\nthan being asked about each one.\n" ;
  add (-1002) "What should be done with all components that are NOT YET installed?\n[I]nstall them, [S]kip them, [A]sk about each one? " ;
  add (-1003) "What should be done with all components that are ALREADY installed?\n[R]e-install them, [U]ninstall them, [S]kip them, [A]sk about each one? " ;

  add (-1004) "PLEASE email the file" ;
  add (-1005) "to" ;

  add (-1006) "Install Component [" ;
  add (-1007) "]?\n[R]e-Install, [N]o Change or [U]ninstall or [Q]uit? " ;
  add (-1008) "]?\n[I]nstall, or [N]ot Install or [Q]uit? " ;

  add (-1009) "Enter part of a module name: " ;

  add (-1010) "Re-Installing [";
  add (-1011) "] component " ;
  add (-1012) "Re-Installing Using Language" ;

  add (-1013) "Removing old installation of [" ;
  add (-1014) "] first ..." ;
  add (-1015) "SUCCESSFULLY REMOVED OLD" ;

  add (-1016) "Installing" ;
  add (-1017) "ERROR Installing [" ;
  add (-1018) "], rolling back to previous state" ;
  add (-1019) "SUCCESSFULLY INSTALLED" ;
  add (-1020) "Skipping" ;
  add (-1021) "Removing [";
  add (-1022) "] (component #";
  add (-1023) ")" ;
  add (-1024) "SUCCESSFULLY REMOVED [" ;

  add (-1025) "]?\n[R]e-install, [N]o Change, [U]ninstall, [Q]uit or choose one:" ;
  add (-1026) "]?\n[N]o, [Q]uit or choose one:" ;
  add (-1027) " (currently installed)";

  add (-1028) "Would you like to display the components from [";
  add (-1029) "]?\n[Y]es, [N]o? " ;

  add (-1030) "]?\nchoose one:" ;
  add (-1031) "]?\n[R]e-install, [N]o Change, [Q]uit or choose one:" ;
  ()


(************************************************************************
 * Evaluate a TP2 Patch Expression
 ************************************************************************)
let log_match a b =
  Filename.basename (String.uppercase a) = Filename.basename (String.uppercase b)

type status = Installed | Temporarily_Uninstalled | Permanently_Uninstalled

let the_log = ref []

let already_installed tp2 i =
  let rec is_installed lst = match lst with
  | [] -> false
  | (a,b,c,sopt,d) :: tl when log_match a tp2 
        && c = i && d <> Permanently_Uninstalled -> true
  | hd :: tl -> is_installed tl 
  in is_installed !the_log 

let installed_lang_index tp2 =
  let rec is_installed lst = match lst with
  | [] -> None
  | (a,b,c,sopt,d) :: tl when log_match a tp2
        && d <> Permanently_Uninstalled -> Some(b)
  | hd :: tl -> is_installed tl
  in is_installed !the_log

let temporarily_uninstalled tp2 i =
  let rec is_installed lst = match lst with
  | [] -> false
  | (a,b,c,sopt,d) :: tl when log_match a tp2
        && c = i && d = Temporarily_Uninstalled -> true
  | hd :: tl -> is_installed tl
  in is_installed !the_log

let is_true i = i <> 0l
(* (Int32.compare i 0l) <> 0 *)
let if_true p = if p then 1l else 0l 

let eval_pe_warn = ref true  

let rec eval_pe_str s = match s with
  | PE_LiteralString(s) -> s 
  | PE_Evaluate(p) -> Var.get_string (eval_pe_str p)
  | PE_Uppercase(s) -> String.uppercase (eval_pe_str s)
  | PE_Lowercase(s) -> String.lowercase (eval_pe_str s)

let eval_pe_tlk_str game s = match s with
  | PE_Tlk(d) -> Dc.single_string_of_tlk_string game d
  | PE_Pe(p) -> eval_pe_str p

exception FoundInt of int

let rec eval_pe game p =
  match p with
  | Pred_True -> 1l 
  | PE_String(s) ->
      let s = eval_pe_str s in
    begin
      try
        Int32.of_string s
      with e -> 
      begin 
      try
        Var.get_int32 ("%" ^ s ^ "%")
      with e -> 
        begin
          try
            Var.get_int32 s
          with e ->
            begin
              (if !eval_pe_warn then log_and_print
                "ERROR: cannot convert %s or %%%s%% to an integer\n" s s) ;
              raise e
            end 
        end
      end 
    end
  | PE_VariableIsSet(s) ->
      let s = eval_pe_str s in
      if Hashtbl.mem Var.variables s then 1l
      else if Hashtbl.mem Var.variables ("%" ^ s ^ "%") then 1l
      else 0l
  
  | PE_IdsOfSymbol(file,entry) ->
      begin try
          let file = Var.get_string file in
          let entry = Var.get_string entry in
          Bcs.int_of_sym game file entry
        with _ -> -1l
      end


  | PE_Random(lb,ub) ->
    let lb = eval_pe game lb in
    let ub = eval_pe game ub in
    if lb > ub then 0l 
    else if lb = ub then lb
    (* random(3,5) = 3 + rand_zero_exclusive((5-3)+1) *)
    else Int32.add lb (Random.int32 (Int32.succ (Int32.sub ub lb)))

  | PE_FileContainsEvaluated(filename, reg) ->
    begin
    let filename = eval_pe_str filename in 
    let reg = eval_pe_str reg in
    let filename = Var.get_string filename in
    let reg = Var.get_string reg in
    let old_allow_missing = !Load.allow_missing in 
    Load.allow_missing := [String.uppercase filename] ; 
    let answer = 
    try
      let buf = 
        if file_exists filename then load_file filename 
        else 
          let a,b = split filename in 
          Load.skip_next_load_error := true; 
          let buff,path = 
            Load.load_resource "FILE_CONTAINS_EVALUATED" game true a b
          in buff
      in 
      if buf = "" then 0l
      else
      let regexp = Str.regexp_case_fold reg in
        let _ = Str.search_forward regexp buf 0 in 
        1l
      with Not_found -> 0l
    in
    Load.allow_missing := old_allow_missing ;
    answer
    end

  | Pred_File_MD5(f,s) -> if_true (
      let f = eval_pe_str f in
      let s = eval_pe_str s in 
      if file_exists f then begin
        let digest = Digest.file f in
        let hex = Digest.to_hex digest in
        log_only "File [%s] has MD5 checksum [%s]\n" f hex ; 
        (String.uppercase hex) = (String.uppercase s)
      end else begin
        log_only "WARNING: File [%s] not found for MD5 checksum.\n" f ;
        false
      end)

  | Pred_File_Exists(f) -> if_true (
      let f = eval_pe_str f in
      let filename = (Var.get_string f) in
      bigg_file_exists filename game.Load.key  )

  | Pred_File_Is_In_Compressed_Bif(f) -> if_true (
      let f = eval_pe_str f in
      let filename = (Var.get_string f) in
      let (a,b) = split filename in
      try
        let (a,b,c,this_biff) = Load.find_in_key game a b in
        this_biff.Biff.compressed
      with _ -> failwith (Printf.sprintf "couldn't find file %s in a biff" f) )

  | Pred_File_Exists_In_Game(f) -> if_true  (
      let f = eval_pe_str f in
      let old_allow_missing = !Load.allow_missing in
      let f = Var.get_string f in
      Load.allow_missing := [] ;
      let res =
      (try
        let a,b = split f in
        Load.skip_next_load_error := true;
        let buff,path = Load.load_resource "FILE_EXISTS_IN_GAME" game true a b in
          (String.length buff > 0) 
      with _ -> false ) in
      Load.allow_missing := old_allow_missing ;
      res )
  | Pred_File_Size(f,s) -> 
      let f = eval_pe_str f in 
      if_true (file_size (Var.get_string f) = s)
  | Pred_File_Contains(f,r) -> if_true (
      let f = eval_pe_str f in
      let r = eval_pe_str r in
      let buf = load_file f in
      let regexp = Str.regexp_case_fold r in 
      try 
        let _ = Str.search_forward regexp buf 0 in 
        true
      with Not_found -> false)

  | PE_StringRegexp(s1,s2,exact) ->
      let s1 = eval_pe_str s1 in
      let s2 = eval_pe_str s2 in
      let s1 = Var.get_string s1 in
      let s2 = Var.get_string s2 in
      let r = Str.regexp_case_fold s2 in
      let b = 
        if exact then Str.string_match r s1 0
        else 
          try
            let _ = Str.search_forward r s1 0 in
            true 
          with _ -> false
      in
      if b then 0l else 1l

  | PE_StringEqual(s1,s2,ignore_case,isbool) ->
      let s1 = eval_pe_str s1 in
      let s2 = eval_pe_str s2 in
      let s1 = Var.get_string s1 in
      let s2 = Var.get_string s2 in 
      let comparison = if ignore_case then 
        (fun s1 s2 -> String.compare (String.uppercase s1)
        (String.uppercase s2)) else String.compare
      in
      let result = Int32.of_int (comparison s1 s2) in
      let bigg_make_bool () =
        (if isbool then
          if result = Int32.of_int 0 then Int32.of_int 1 else Int32.of_int 0
        else
          result )
      in
      bigg_make_bool ()
  | PE_Add(a,b) -> let a = eval_pe game a in let b = eval_pe game b in Int32.add a b
  | PE_Sub(a,b) -> let a = eval_pe game a in let b = eval_pe game b in Int32.sub a b
  | PE_Mul(a,b) -> let a = eval_pe game a in let b = eval_pe game b in Int32.mul a b
  | PE_Div(a,b) -> let a = eval_pe game a in let b = eval_pe game b in 
    (try Int32.div a b with Division_by_zero -> Int32.zero)
  | PE_Exp(a,b,c) -> 
    (try 
    let a = Int32.to_float (eval_pe game a) in 
    let b = Int32.to_float (eval_pe game b) in
    let c = Int32.to_float (eval_pe game c) in
    let bc = b /. c in 
    let answer = a ** bc in
    Int32.of_float answer 
    with _ -> 0l)

  | PE_Not(a) ->   if is_true (eval_pe game a)
                   then Int32.zero else Int32.one
  | PE_And(a,b) -> if is_true (eval_pe game a) && is_true (eval_pe game b)
                  then Int32.one else Int32.zero 
  | PE_Or(a,b) -> if is_true (eval_pe game a) || is_true (eval_pe game b)
                  then Int32.one else Int32.zero 
  | PE_Equal(a,b) -> if Int32.compare (eval_pe game a) (eval_pe game b) = 0
                  then Int32.one else Int32.zero
  | PE_GT(a,b) -> if Int32.compare (eval_pe game a) (eval_pe game b) > 0 
                  then Int32.one else Int32.zero 
  | PE_GTE(a,b) -> if Int32.compare (eval_pe game a) (eval_pe game b) >= 0
                  then Int32.one else Int32.zero 
  | PE_LT(a,b) -> if Int32.compare (eval_pe game a) (eval_pe game b) < 0 
                  then Int32.one else Int32.zero 
  | PE_LTE(a,b) -> if Int32.compare (eval_pe game a) (eval_pe game b) <= 0
                  then Int32.one else Int32.zero

  | PE_BAND(a,b) -> Int32.logand (eval_pe game a) (eval_pe game b) 
  | PE_BOR(a,b) -> Int32.logor (eval_pe game a) (eval_pe game b) 
  | PE_BXOR(a,b) -> Int32.logxor (eval_pe game a) (eval_pe game b) 
  | PE_BNOT(a) -> Int32.lognot (eval_pe game a)
  | PE_BLSL(a,b) -> Int32.shift_left (eval_pe game a) (Int32.to_int (eval_pe game b))
  | PE_BLSR(a,b) -> Int32.shift_right_logical (eval_pe game a) (Int32.to_int (eval_pe game b) )
  | PE_BASR(a,b) -> Int32.shift_right (eval_pe game a) (Int32.to_int (eval_pe game b) )

  | PE_If(p,t,e) -> if is_true (eval_pe game p) then eval_pe game t else eval_pe game e
  | PE_ModIsInstalled(filename,number) -> if_true (already_installed filename number &&
             not (temporarily_uninstalled filename number))
  | PE_GameIs(game_list) -> begin
      let game_list = Str.split many_whitespace_regexp game_list in
      eval_pe game (List.fold_left (fun acc this -> if acc = (get_pe_int "1") then (get_pe_int "1") else begin
          match String.uppercase this with
          | "BG2"
          | "SOA"    -> PE_And(PE_And(Pred_File_Exists_In_Game(get_pe_string "flythr01.mve"),
                      PE_Not(Pred_File_Exists_In_Game(get_pe_string "saradush.mve"))),
                      PE_Not(Pred_File_Exists_In_Game(get_pe_string "_eregost.mve")))
          | "TOB"    -> PE_And(Pred_File_Exists_In_Game(get_pe_string "saradush.mve"),
                      PE_Not(Pred_File_Exists_In_Game(get_pe_string "_eregost.mve")))
          | "IWD2"   -> Pred_File_Exists_In_Game(get_pe_string "subrace.ids")
          | "PST"    -> Pred_File_Exists_In_Game(get_pe_string "faction.ids")
          | "BG1"    -> PE_And(Pred_File_Exists_In_Game(get_pe_string "beregost.mve"),
                        PE_Not(Pred_File_Exists_In_Game(get_pe_string "durlag.mve")))
          | "TOTSC"  -> Pred_File_Exists_In_Game(get_pe_string "durlag.mve")
          | "IWD"
          | "IWD1"   -> PE_And(Pred_File_Exists_In_Game(get_pe_string "avalanch.mve"),
                        PE_Not(Pred_File_Exists_In_Game(get_pe_string "howdrag.mve")))
          | "HOW"    -> Pred_File_Exists_In_Game(get_pe_string "howdrag.mve")
          | "TUTU"   -> PE_And(Pred_File_Exists_In_Game(get_pe_string "_eregost.mve"),
                        PE_Not(Pred_File_Exists_In_Game(get_pe_string "_durlag.mve")))
          | "TUTU_TOTSC"
          | "TUTU+TOTSC"  -> Pred_File_Exists_In_Game(get_pe_string "_durlag.mve")
          | _ -> failwith (Printf.sprintf "No rule to identify %s" (String.uppercase this))
        end
      ) (get_pe_int "0") game_list )
    end
  | PE_IsAnInt(x) -> (eval_pe_warn := false ;
    try
      let _ = (eval_pe game (PE_String(x))) in
      1l
    with _ ->
      0l
    )
  | PE_StateWhichSays(lse,traref,file) -> begin
      let rec lookforit game lse lst =
        if !debug_ocaml then log_and_print "in lookforit \n";
        try (match lst with
          | [] -> []
          | hd :: tl -> looksingle game lse hd; lookforit game lse tl
        ) with 
        | FoundInt(i) -> i :: lookforit game lse (List.tl lst)
        | e -> raise e
      and looksingle game lse this =
        if !debug_ocaml then log_and_print "looking in %s\n" this;
        Dc.push_copy_trans () ;
        !handle_tra_filename_global this ;
        try
            ( match Dc.resolve_tlk_string_internal false false game lse
              with
                | Dlg.TLK_Index(i) -> Dc.pop_trans (); if !debug_ocaml then begin
                  let str = Dc.single_string_of_tlk_string game (Dlg.TLK_Index(i)) in
                  let str = Var.get_string str in
                  log_and_print "\n%d - %s\n" i str
                end ; raise(FoundInt i)
                | _ -> Dc.pop_trans ();
            )
          with
            FoundInt(i) -> raise (FoundInt i)
            | a -> if !debug_ocaml then log_and_print "%s\n" (Printexc.to_string a)
      in
      let theref = match traref with
        | None (*going by tra + @ref*) ->
          begin try
              ( match Dc.resolve_tlk_string_internal false true game (value_of_option lse)
                with
                  | Dlg.TLK_Index(i) -> [i]
                  | _ -> log_and_print "ERROR: cannot resolve SAY patch\n" ; failwith "resolve"
              )
            with Not_found -> [0 - 3]
              | e -> raise e
          end
        | Some(num,file) (* going by hard string *) ->
          begin
            let lst = try
                let regexp = Str.regexp "%s" in
                let addr = Str.search_forward regexp file 0 in
                let parts = Str.split regexp file in
                if List.length parts > 2 then begin
                  log_and_print "Can't use more than one %%s in STATE_WHICH_SAYS ... IN\n" ;
                  failwith "STATE_WHICH_SAYS %s/%s"
                end ;
                let dh = Case_ins.unix_opendir (List.nth parts 0) in
                let rec readdir dh =
                  try
                    let x = Unix.readdir dh in
                    if !debug_ocaml then log_and_print "%s-\n" x;
                    if (Case_ins.unix_stat ((List.nth parts 0) ^ x)).Unix.st_kind = Unix.S_DIR &&
                        (List.nth parts 0) <> "." && (List.nth parts 0) <> ".." then
                      ((List.nth parts 0) ^ x ^ (List.nth parts 1)) :: readdir dh
                    else readdir dh
                  with End_of_file -> Unix.closedir dh ;[]
                in
                readdir dh
              with
              | Not_found -> [file]
              | e -> raise e
            in
            if !debug_ocaml then List.iter (fun x -> log_and_print "%s\n" x) lst;
            let x = begin try
              lookforit game (Dlg.Trans_String(Int32.to_int(eval_pe game num))) lst
              with
                | Not_found -> [0 - 3]
                | e -> Dc.pop_trans ();  raise e
            end
            in
            x
          end
      in
      if !debug_ocaml then log_and_print "Looking for these: \n" ;
      if !debug_ocaml then List.iter (fun x -> log_and_print "%d\n" x) theref ;
      if theref = [] then log_and_print "Couldn't find any suitable string in STATE_WHICH_SAYS\n";
      if theref = [] then failwith "resolve" ;
      if theref = [0 - 3] then Int32.of_int (0 - 3) else begin
        let (a,b) = split file in
        let buff,path = Load.load_resource "STATE_WHICH_SAYS" game true a b in
        if !debug_ocaml then log_and_print "%s \n" path ;
        let numstates = int_of_str_off buff 8  in
        let stateoff  = int_of_str_off buff 12 in
        let whichstate = ref (0 - 1) in
        let howmany = ref 0 in
        for i = 0 to numstates - 1 do
          let this_say = int_of_str_off buff ((16 * i) + stateoff) in
          if List.mem this_say theref then whichstate := i ;
          if List.mem this_say theref && !debug_ocaml then log_and_print "Found number %d at state %d\n" this_say i;
          if List.mem this_say theref then incr howmany ;
        done ;
        if !howmany = 0 then Int32.of_int (0 - 1) else if !howmany > 1 then Int32.of_int (0 - 2) else
          Int32.of_int !whichstate
      end
    end

let eval_pe game pe =
  let res = Stats.time "eval_pe" (fun () -> eval_pe game pe) () in
  (if !debug_pe then log_and_print "Value [%s] = %ld\n"
    (pe_to_str pe) res ) ;
  res 

(************************************************************************
 * Uninstall STRSET 
 ************************************************************************)
let uninstall_strset game filename = 
  if (file_exists filename) then begin
    (try
      let infile = Case_ins.perv_open_in_bin filename in
      let record : Load.str_set_record list =
        Marshal.from_channel infile
      in
      close_in infile ;
      List.iter (fun (i,m,f) ->
        if (i < 0 || i > Array.length game.Load.dialog) then begin
          log_only "WARNING: Cannot uninstall STRING_SET #%d, out of range 0 -- %d\n" i (Array.length game.Load.dialog) 
        end else begin
        (*
          log_only "Un-SET_STRING #%d from %s back to %s\n"
            i
            (Tlk.short_print game.Load.dialog.(i) 18)
            (Tlk.short_print m 18) ;
            *)
          game.Load.dialog.(i) <- m ;
          game.Load.dialog_mod <- true; 
          match game.Load.dialogf with
          | Some(a) -> a.(i) <- f ; game.Load.dialogf_mod <- true; 
          | None -> ()
        end
      ) record
    with e ->
      log_and_print "WARNING: Unable to uninstall STRING_SET references from [%s]: %s\n" filename (Printexc.to_string e));
    begin
      try
        Case_ins.unix_unlink filename
      with e ->
        log_or_print "Unable to Unlink [%s]: %s"
          filename (Printexc.to_string e)
    end ;
  end else log_only "[%s] SET_STRING uninstall info not found\n" filename

let record_strset_uninstall_info game filename =
  if (game.Load.str_sets = []) then
    ()
  else begin
    try 
      let outchan = Case_ins.perv_open_out_bin filename in
      Marshal.to_channel outchan game.Load.str_sets [];
      close_out outchan ;
      game.Load.str_sets <- []
    with e ->
      log_and_print "WARNING: Unable to write STRING_SET uninstall info to [%s]: %s\n" filename (Printexc.to_string e)
  end

let handle_recursive_tp2 = ref (fun s -> failwith "cannot handle recursive
TP2 invocation at this time")

(************************************************************************
 * Uninstall a TP2 component
 ************************************************************************)
let handle_at_uninstall tp2 m do_uninstall do_interactive_uninstall game =
    let rec handle_al al =
      List.iter (fun a -> match a with
        | TP_At_Uninstall(str) ->
          if do_uninstall then begin
          let str = Var.get_string str in
          match (split (String.uppercase str)) with
          | _,"TP2" -> (!handle_recursive_tp2) str
          | _,_ -> let str = Arch.handle_view_command str !skip_at_view in
                 ignore (Unix.system(Arch.slash_to_backslash str))
          end
        | TP_At_Interactive_Uninstall(str) ->
          if do_interactive_uninstall then begin
            let str = Var.get_string str in
            match (split (String.uppercase str)) with
            | _,"TP2" -> (!handle_recursive_tp2) str
            | _,_ -> let str = Arch.handle_view_command str !skip_at_view in
                   ignore(Unix.system(Arch.slash_to_backslash str))
          end
        | TP_If(p,al1,al2) ->
          begin try
              eval_pe_warn := false ;
              let res = is_true (eval_pe game p) in
              (* log_or_print "IF evaluates to %b\n" res ; *)
              if res then begin
                handle_al al1
              end else begin
                handle_al al2
              end
            with _ -> () ;
            eval_pe_warn := true ;
          end
        | TP_Biff _ ->
            (* re-load the chitin *)
            let keyname = Load.find_file_in_path "." "^chitin.key$" in
            let keybuff = load_file keyname in
            game.Load.key <- Key.load_key keyname keybuff ;
            game.Load.loaded_biffs <- Hashtbl.create 5 ;
        | TP_Include(string_list) ->
          List.iter (fun file ->
            let incl_al = if Hashtbl.mem loaded_tph file
              then begin
                if !debug_ocaml then log_and_print "%s is already loaded, restoring the image...\n" file ;
                Hashtbl.find loaded_tph file
              end
              else begin
                if !debug_ocaml then log_and_print "Loading %s...\n" file ;
                let x = !handle_tph_filename_global file in
                Hashtbl.replace loaded_tph file x ;
                x
              end
            in
            handle_al incl_al ;
          ) string_list ;
        | TP_Reinclude(string_list) ->
          List.iter (fun file ->
            let incl_al =
              if !debug_ocaml then log_and_print "Loading %s...\n" file ;
              let x = !handle_tph_filename_global file in
              Hashtbl.replace loaded_tph file x ;
              x
            in
            handle_al incl_al ;
          ) string_list ;
(* for now, no AT_* in macroes. Too shifty.
        | TP_Launch_Action_Macro(str) ->
          begin try
              let (decl, actions) =
                Hashtbl.find action_macros str
              in
              handle_al actions ;
            with _ ->
              log_and_print "WARNING: can't find AT_*UNINSTALL in macro\n"
          end
        | TP_Define_Action_Macro(str,decl,al) ->
            Hashtbl.replace action_macros str (decl, al)
        | TP_Define_Patch_Macro(str,decl,al) ->
            Hashtbl.replace patch_macros str (decl, al) *)
        | _ -> ()
      ) al
    in
    let handle_flag fl =
      List.iter (fun flag ->
        match flag with
        | Always(al) -> handle_al al ;
        | _ -> ()
      ) fl
    in
    handle_flag tp2.flags ;
    handle_al m.mod_parts

let uninstall_tp2_component game tp2 tp_file i interactive =
  Stats.time "tp2 uninstall" (fun () ->
  try
    let result = tp2 in
    let d = result.backup ^ "/" ^ (string_of_int i) in
    let u_filename = (Printf.sprintf "%s/UNINSTALL.%d" d i) in
    let u_strset_filename = (Printf.sprintf "%s/UNSETSTR.%d" d i) in
    uninstall_strset game u_strset_filename ;
(*     let buff = load_file u_filename in
    let file_list = Str.split one_newline_regexp buff in *)
    let file_list = ref [] in
    let inchan = Case_ins.perv_open_in_bin u_filename in
    try
      while true do
        file_list := (input_line inchan) :: !file_list ;
      done
    with End_of_file -> () ;
    close_in inchan;
    let file_list = List.rev (!file_list) in
    log_and_print "Will uninstall %3d files for [%s] component %d.\n"
      (List.length file_list) tp_file i; 
    List.iter (fun override_filename -> 
      if !debug_ocaml then log_and_print "Restoring %s\n" override_filename ;
      let base = Filename.basename override_filename in 
      let backup_filename = d ^ "/" ^ base in 
      (* did we save a backup copy? *) 
      if file_exists backup_filename then begin
        (* copy it from Backup to Override *)
        (* let buff = load_file backup_filename in
         * let restore_chn = Case_ins.perv_open_out_bin override_filename in
         * output_string restore_chn buff ;
         * close_out restore_chn ; *)
        begin
          try
            Case_ins.unix_unlink override_filename
          with e ->
            log_or_print "Unable to Unlink [%s]: %s"
              override_filename (Printexc.to_string e)
        end ;
        if !debug_ocaml then log_and_print "Calling Util.copy_large_file() to restore a backup\n" ;
        log_or_print "  Restoring backed-up [%s]\n" backup_filename ;
        copy_large_file backup_filename override_filename "restoring a backup" ;
        begin
          try
            Case_ins.unix_unlink backup_filename
          with e ->
            log_or_print "Unable to Unlink [%s]: %s"
              backup_filename (Printexc.to_string e)
        end
      end else begin
        (* no backup copy, so just delete it from override *)
        log_or_print "  Deleting [%s] (to restore original game condition)\n"
          override_filename ;
        try
          Case_ins.unix_unlink override_filename
        with e ->
          log_or_print "Unable to Unlink [%s]: %s"
            override_filename (Printexc.to_string e)
      end 
    ) (List.rev file_list)  ;
    begin try
      Case_ins.unix_unlink u_filename ;
    with e ->
      log_or_print "Unable to Unlink [%s]: %s"
        u_filename (Printexc.to_string e)
    end ;
    let m = get_nth_module result i true in
    handle_at_uninstall tp2 m true interactive game ;
    log_and_print "Uninstalled    %3d files for [%s] component %d.\n" 
      (List.length file_list) tp_file i; 
  with e ->
    log_and_print "Error Uninstalling [%s] component %d:\n%s\n"
      tp_file i (Printexc.to_string e)
  ) ()

(************************************************************************
 * Determine what has been installed. 
 ************************************************************************)
type installed_mods = (string * int * int * (string option) * status) list
  (* module : language : component : status *) 
let log_name = "WeiDU.log"

let str_of_str_opt sopt = match sopt with
    Some(str) -> "~" ^ str ^ "~"
  | None -> ""

let print_log () = 
  List.iter (fun (n,i1,i2,sopt,st) ->
    log_or_print "%s %2d %2d %s %s\n" n i1 i2 (match st with
      Installed -> "Installed"
    | Temporarily_Uninstalled -> "Temporarily_Uninstalled"
    | Permanently_Uninstalled -> "Permanently_Uninstalled")
    (str_of_str_opt sopt)
  ) !the_log

let save_log game handle_tp2_filename handle_tra_filename get_tra_list_filename =
  let out = Case_ins.perv_open_out log_name in
  log_or_print "Saving This Log:\n" ; print_log () ;
  Printf.fprintf out "// Log of Currently Installed WeiDU Mods\n" ;
  Printf.fprintf out "// The top of the file is the 'oldest' mod\n" ;
  Printf.fprintf out "// ~TP2_File~ #language_number #component_number%s"
    (if !quick_log then "\n" else " // [Subcomponent Name -> ] Component Name\n") ;
  let newline_regexp = one_newline_or_cr_regexp in
  let tp2_ht = Hashtbl.create 511 in
  let tra_ht = Hashtbl.create 511 in
  List.iter (fun (a,b,c,sopt,d) ->
    let str =
      if not !quick_log then begin
        let component_name, subcomponent_name =
          try
            (* log_or_print "*** Looking Up %s.\n" a; *)
            let tp2 =
              try Hashtbl.find tp2_ht a
              with _ ->
                (* log_or_print "*** Parsing %s.\n" a; *)
                let res = handle_tp2_filename a in
                Hashtbl.add tp2_ht a res ; res
            in
            Dc.clear_state () ;
            Dc.push_trans ();
            let a_dir = Filename.dirname a in
            (try
              let l = List.nth tp2.languages b in
              List.iter (fun s ->
                let x =
                (* log_or_print "*** Loading %s for %s.\n" s a; *)
                  try Hashtbl.find tra_ht s
                  with _ ->
                    (let x = get_tra_list_filename (Arch.backslash_to_slash s) in
                    Hashtbl.add tra_ht s x; x
                    )
                in
                Stats.time "adding translation strings" Dc.add_trans_strings x
              ) l.lang_tra_files ;
            with _ -> ()) ;
            let m = get_nth_module tp2 c true in
            let comp_str = Dc.single_string_of_tlk_string game m.mod_name in
            let subcomp_group the_comp =
              let rec walk lst = match lst with
              | TPM_SubComponents(ts,a,b) :: tl -> Some(ts)
              | hd :: tl -> walk tl
              | [] -> None
              in walk the_comp.mod_flags
            in
            let subcomp_str = (
              match subcomp_group m with
              | None    -> ""
              | Some(x) -> "" ^ (Dc.single_string_of_tlk_string game x) ^ " -> ") in
            Dc.clear_state () ;
            Dc.pop_trans ();
            (comp_str, subcomp_str)
          with _ ->
            ( "??? -> ", "???" )
        in
        let    component_name = Str.global_replace newline_regexp " "    component_name in
        let subcomponent_name = Str.global_replace newline_regexp " " subcomponent_name in
        Printf.sprintf "~%s~ #%d #%d // %s%s\n"
          (String.uppercase a) b c subcomponent_name component_name
      end
      else begin
        Printf.sprintf "~%s~ #%d #%d\n"
          (String.uppercase a) b c
      end
    in
  match d with
    Installed -> output_string out str
  | Temporarily_Uninstalled ->
      log_and_print "Internal Error: saving a log with temporarily uninstalled module %s" str 
  | Permanently_Uninstalled ->
      Printf.fprintf out "// Recently Uninstalled: %s" str
  ) !the_log ;
  close_out out

let temp_to_perm_uninstalled tp2 i handle_tp2_filename game =
  let rec is_installed lst = match lst with
  | [] -> []
  | (a,b,c,sopt,d) :: tl when log_match a tp2
        && c = i && d = Temporarily_Uninstalled ->
        (* if there were any "at_uninstall" actions here, do them! *) 
        let tp_file = a in
        let tp2 = handle_tp2_filename tp_file in 
        let lang_name = 
        (try
          let l = List.nth tp2.languages b in
          l.lang_dir_name ;
        with _ -> "" ) in 
        Var.set_string "TP2_AUTHOR" tp2.author ;
        Var.set_string "LANGUAGE" lang_name ;
        Var.set_string "TP2_FILE_NAME" tp2.tp_filename ;
        let m = get_nth_module tp2 c true in
        log_only "Running AT_INTERACTIVE_EXITs in ~%s~ %d %d %s\n"
          (String.uppercase a) b c
          (str_of_str_opt sopt) ;
        handle_at_uninstall tp2 m
          false (* "AT_UNINSTALL" was already done! *)
          true (* but the user just asked for this to be explicit *) game ;
        (* Var.remove_var "LANGUAGE" ; *) 
        (a,b,c,sopt,Permanently_Uninstalled) :: tl
  | hd :: tl -> hd :: (is_installed tl)
  in the_log := is_installed !the_log 

(************************************************************************
 * Do everything necessary to uninstall the given tp2 component. This
 * may require temporarily uninstalling stuff that has been installed
 * since our target was installed. 
 ************************************************************************)
let rec find_best_file lst =
  match lst with
    hd :: tl -> if file_exists hd then hd
                else if file_exists (Filename.basename hd) then
                  (Filename.basename hd)
                else find_best_file tl
  | [] -> failwith "TP2 not found!"  

let uninstall game handle_tp2_filename tp2 i interactive =
  log_or_print "uninstall: %s %d\n" tp2 i ;
  let worked = ref true in 
  if not (already_installed tp2 i) then begin
    log_and_print "Internal Error: trying to uninstall non-installed mod component [%s] %d\n" tp2 i ;
    false
  end else if (temporarily_uninstalled tp2 i) then begin
    log_or_print "uninstall: %s %d already temporarily uninstalled\n" tp2 i ;
    temp_to_perm_uninstalled tp2 i handle_tp2_filename game ;
    true
  end else begin
    let rec prepare lst = match lst with
    | [] -> [] (* end of the line *)

    (* this is the entry in the list *)
    | (a,b,c,sopt,d) :: tl when log_match a tp2 && c = i -> 
      begin match d with
      | Permanently_Uninstalled -> (* some sort of error *) 
                  log_and_print "Internal Error: mod component [%s] %d already uninstalled\n" tp2 i ; (a,b,c,sopt,Permanently_Uninstalled) :: tl
      | Temporarily_Uninstalled -> (* we just won't restore it! *)
                      (a,b,c,sopt,Permanently_Uninstalled) :: tl
      | Installed -> 
          begin 
          try
            let best = find_best_file [a ; tp2] in
            uninstall_tp2_component game (handle_tp2_filename best) a c interactive ;
            (a,b,c,sopt,Permanently_Uninstalled) :: tl
          with _ ->
            log_and_print "ERROR: This Mod is too old (or too new) to uninstall that component for you.\nUpgrade to the newest versions of this mod and that one and try again.\n" ;
            worked := false ; 
            lst
          end 
      end

    | (a,b,c,sopt,d) :: tl -> 
      begin match d with
      | Permanently_Uninstalled    
      | Temporarily_Uninstalled -> (* keep going *)
                      (a,b,c,sopt,d) :: (prepare tl)
      | Installed -> 
          log_or_print "We must temporarily uninstall [%s] component %d\n"
            a c ; 
          begin
          try
            let best = find_best_file [ a] in
            uninstall_tp2_component game (handle_tp2_filename best) a c false ;
            (* take away for now *)
            (a,b,c,sopt,Temporarily_Uninstalled) :: (prepare tl)
          with e ->
            log_and_print "ERROR: This Mod is too old (or too new) to uninstall that component for you.\nUpgrade to the newest versions of this mod and that one and try again.\n[%s]\n"
            (Printexc.to_string e);
            worked := false ; 
            lst 
          end
      end
    in
    let new_log = List.rev (prepare (List.rev !the_log)) in
    the_log := new_log ;
    !worked
  end

let body_of_script buff = 
  if buff = "" then "" else 
  try 
    let first_nl = String.index buff '\n' in
    let last_nl = (String.rindex buff 'S') - 1 in 
    if first_nl = last_nl then
      "" 
    else 
      let length = (last_nl - first_nl) - 1 in 
      String.sub buff (first_nl + 1) length 
  with e -> 
    log_and_print "ERROR: not a BCS script\n" ;
    failwith "not a BCS script"

type default_action = TP_Install | TP_Uninstall | TP_Skip | TP_Ask

(************************************************************************
 * change location "where" in "buff" to point to str-ref "what"
 ************************************************************************)
let process_patch1 patch_filename game buff p =
  Stats.time "READ_*" (fun () -> 
  let bounds_check idx size retfun eo ef =
    let len = String.length buff in
    let out_of_bounds = (idx < 0 || (idx + size) > len) in
    match out_of_bounds, eo with
    | false, _ -> (retfun ())
    | true, Some(v) -> ef v
    | true, None -> 
    begin 
      log_and_print "ERROR: illegal %d-byte read from offset %d of %d-byte file %s\n" size idx len patch_filename ;
      failwith (patch_filename ^ ": read out of bounds")
    end 
  in 
  match p with 
  | TP_PatchReadByte(where,name,eo) ->
      let where = Int32.to_int (eval_pe game where) in
      let value = bounds_check where 1  
        (fun () -> Int32.of_int (byte_of_str_off buff where)) eo 
        (fun v -> (eval_pe game v)) in
      Var.set_int32 name value
  | TP_PatchReadSByte(where,name,eo) ->
      let where = Int32.to_int (eval_pe game where) in
      let value = bounds_check where 1  
        (fun () -> Int32.of_int (signed_byte_of (byte_of_str_off buff where))) eo 
        (fun v -> (eval_pe game v)) in
      Var.set_int32 name value
  | TP_PatchReadShort(where,name,eo) ->
      let where = Int32.to_int (eval_pe game where) in 
      let value = bounds_check where 2
        (fun () -> Int32.of_int (short_of_str_off buff where)) eo
        (fun v -> (eval_pe game v)) in
      Var.set_int32 name value
  | TP_PatchReadSShort(where,name,eo) ->
      let where = Int32.to_int (eval_pe game where) in
      let value = bounds_check where 2 
        (fun () -> Int32.of_int (signed_short_of (short_of_str_off buff where))) eo
        (fun v -> eval_pe game v) in
      Var.set_int32 name value
  | TP_PatchReadLong(where,name,eo) ->
      let where = Int32.to_int (eval_pe game where) in
      let value = bounds_check where 4 
        (fun () -> int32_of_str_off buff where) eo 
        (fun v -> (eval_pe game v)) in
      Var.set_int32 name value
  | TP_PatchGetStrRef(which,name) ->
      let which = Int32.to_int (eval_pe game which) in
      let value = Dc.pretty_print_no_quote game.Load.dialog which in
      Var.set_string name value
  | TP_PatchReadStrRef(where,name,eo) ->
      let where = Int32.to_int (eval_pe game where) in
      let value = bounds_check where 4
        (fun () ->
          let i = int_of_str_off buff where in
          Dc.pretty_print_no_quote game.Load.dialog i) eo
        (fun v -> v) in
      Var.set_string name value
  | TP_PatchReadAscii(where,name,eo,size_opt) ->
      let where = Int32.to_int (eval_pe game where) in
      let size, forced = match size_opt with
      | None -> 8, false
      | Some(pe) ->
          let size = Int32.to_int (eval_pe game pe) in
          size, true
      in
      let value = bounds_check where size
        (fun () -> 
          if forced then String.sub buff where size 
          else get_string_of_size buff where size) eo
        (fun v ->  Var.get_string v)
      in 
      Var.set_string name value

  | TP_Get2DARows(req_col, var_name) ->
      let req_col = Int32.to_int (eval_pe game req_col) in
      let lines = Str.split many_newline_or_cr_regexp buff in
      let entries = List.map (fun line ->
        Str.split many_whitespace_regexp line
      ) lines in
      let num_rows = List.fold_left (fun acc elt ->
        acc + if List.length elt >= req_col then 1 else 0
      ) 0 entries in
      Var.set_int32 var_name (Int32.of_int num_rows)

  | TP_CountRegexpInstances(case_sens,match_exact,find, var_name) ->
      let find = Var.get_string find in
      let case_sens = match case_sens with
          None -> false
        | Some(x) -> x
      in
      let match_exact = match match_exact with
          None -> false
        | Some(x) -> x
      in
      let my_regexp = match case_sens, match_exact with
          false, false -> Str.regexp_case_fold        find
        | true , false -> Str.regexp                  find
        | false, true  -> Str.regexp_string_case_fold find
        | true , true  -> Str.regexp_string           find
      in
      let split = Str.split my_regexp buff in
      let length = List.length split in
      Var.set_int32 var_name (Int32.of_int (length - 1))

  | TP_Read2DA(row, col, req_col, var_name) ->
      let row = Int32.to_int (eval_pe game row) in
      let col = Int32.to_int (eval_pe game col) in
      let req_col = Int32.to_int (eval_pe game req_col) in
      let lines = Str.split many_newline_or_cr_regexp buff in
      let entries = List.map (fun line ->
        let entry_list = Str.split many_whitespace_regexp line in
        Array.of_list entry_list) lines in
      let rec process line_list lines_left =
        match line_list, lines_left with
          [], 0 -> ()
        | [], _ ->
          begin
            log_and_print "ERROR: Cannot find %d rows with at least %d columns." row req_col ;
            failwith "Cannot Read 2DA Entry"
          end
        | hd :: tl, i  when Array.length hd < req_col -> process tl i
        | hd :: tl, 0 -> Var.set_string var_name hd.(col)
        | hd :: tl, i -> process tl (i - 1)
      in
      process entries row ;

  | _ -> ()
  ) ()

(*************************************************************************
 *************************************************************************
                                 handle_tp
 *************************************************************************
 *************************************************************************)

let rec handle_tp
      game
      handle_tp2_filename
      handle_tra_filename
      handle_d_filename
      handle_tph_filename
      compile_baf_filename
      handle_script_buffer
      handle_dlg_buffer
      handle_d_buffer
      get_tra_list_filename
      enqueue_tp2_filename
      emit_dlg_files
      this_tp2_filename
      tp
= begin
      (* allow me to use this stuff before *)
      handle_tph_filename_global := handle_tph_filename ;
      handle_tra_filename_global := handle_tra_filename ;

(*************************************************************************
                           2da helper functions
 *************************************************************************)
      let split_apart str = Str.split many_whitespace_regexp  str
      in
      let get_next_col_number file =
        let (a,b) = split file in
        let buff,path = Load.load_resource "getting 2DA columnns" game true a b in 
        try 
          let lst = Str.split many_newline_or_cr_regexp buff in
          let elt = List.nth lst 5 in
          let cols = split_apart elt in
          let num_cols = List.length cols in 
          let last_col = num_cols - 2 in 
          last_col + 1
        with e -> 
          log_and_print "ERROR: cannot find col numbers in %s\n" file ;
          raise e
      in 
      let get_next_line_number file =
        let (a,b) = split file in 
        let buff,path = Load.load_resource "getting 2DA lines" game true a b in 
        try 
          let idx = Str.search_backward (Str.regexp "[\r\n][0-9]") buff 
            ((String.length buff) - 1) in 
          let minibuff = Str.string_after buff (idx+1) in
          let lst = split_apart minibuff in
          let elt = List.hd lst in
          let last_number = int_of_string elt in
          last_number + 1
        with e -> 
          log_and_print "ERROR: cannot find line numbers in %s\n" file  ;
          raise e
      in
      (* 2da helper functions end *)

  handle_recursive_tp2 := enqueue_tp2_filename ;

  let old_allow_missing = !Load.allow_missing in
  Load.allow_missing := 
    List.fold_left (fun acc elt -> match elt with
      Allow_Missing(lst) -> lst @ acc
      | _ -> acc) [] tp.flags ;

  let old_script_style = game.Load.script_style in
  List.iter (fun f -> match f with
  | Script_Style(s) -> game.Load.script_style <- s
  | _ -> ()
  ) tp.flags ;

  let interactive = ref true in

(*************************************************************************
                            pick your language
 *************************************************************************)
  let our_lang = ref None in
  let our_lang_index = ref 0 in
  begin
    match tp.languages with
      [] -> ()
    | [l] -> our_lang := Some(l)
    | _ ->
      let arr = Array.of_list tp.languages in
      let answer, answer_index =
        match !always_uninstall || !always_yes || !sometimes_reinstall || (!force_uninstall_these <> [] && !force_install_these = [])
                  , (installed_lang_index this_tp2_filename) with
        | true, Some(i) when i >= 0 && (i < Array.length arr) ->
            Some(arr.(i)), i
        | _, _ -> None, 0
      in
      let answer, answer_index = ref answer, ref answer_index in
      if !forced_language >= 0 && (!forced_language < Array.length arr) then begin
        answer := Some(arr.(!forced_language)) ;
        answer_index := !forced_language ;
      end ;
      while !answer = None do
        log_and_print "\nChoose your language:\n" ;
        Array.iteri (fun i l ->
          log_and_print "%2d [%s]\n" i l.lang_name) arr ;
        try
          let i = read_int () in
          if i >= 0 && i < Array.length arr then begin
            answer := Some(arr.(i)) ;
            answer_index := i
          end
        with _ -> ()
        done ;
      our_lang := !answer ;
      our_lang_index := !answer_index ;
  end ;
  let lang_init () =
    init_default_strings () ;
    begin
      match !our_lang with
        None -> ()
      | Some(l) ->
          Var.set_string "LANGUAGE" l.lang_dir_name ;
          log_and_print "Using Language [%s]\n" l.lang_name ;
          Var.set_string "TP2_FILE_NAME" tp.tp_filename;
          log_or_print "[%s] has %d top-level TRA files\n"
            l.lang_name (List.length l.lang_tra_files) ;
          List.iter handle_tra_filename (List.map Arch.backslash_to_slash l.lang_tra_files)
    end
  in
  Var.set_string "TP2_AUTHOR" tp.author ;
  Var.set_string "TP2_FILE_NAME" tp.tp_filename;

  lang_init () ; 

(*************************************************************************
                           process_patch2
 *************************************************************************)
let rec process_patch2 patch_filename game buff p =
  process_patch1 patch_filename game buff p ;

  Stats.time "process_patch2" (fun () ->
  let bounds_check_write idx size (str : string) = 
    let len = String.length buff in 
    let out_of_bounds = (idx < 0 || (idx + size) > len) in
    if out_of_bounds then begin 
      log_and_print "ERROR: illegal %d-byte write (%s) offset %d of %d-byte file %s\n" size str idx len patch_filename ;
      failwith (patch_filename ^ ": write out of bounds")
    end ;
    ()
  in 
  match p with
    TP_PatchStrRef(where,what) -> 
      let where = Int32.to_int (eval_pe game where) in
      let new_index = match Dc.resolve_tlk_string game what with
        Dlg.TLK_Index(i) -> i
      | _ -> log_and_print "ERROR: cannot resolve SAY patch\n" ; failwith "resolve"
      in
      bounds_check_write where 4 (Dc.single_string_of_tlk_string game what) ;
      let new_string = str_of_int new_index in
      String.blit new_string 0 buff where 4 ;
      buff

  | TP_DescribeItem(dest) ->
      let str = Eff_table.describe_itm buff in
      Var.set_string dest str ;
      buff 

  | TP_PatchStrRefEvaluated(where,what) -> 
      let where = Int32.to_int (eval_pe game where) in
      let str = Var.get_string what in
      let lse = {
        lse_male = str ;
        lse_male_sound = "";
        lse_female = str ;
        lse_female_sound = "";
      } in 
      let new_index =
        match Dc.resolve_tlk_string game (Dlg.Local_String(lse)) with
        Dlg.TLK_Index(i) -> i
      | _ -> log_and_print "ERROR: cannot resolve SAY_EVALUATED patch\n" ; 
             failwith "SAY_EVALUATED resolve"
      in
      let new_string = str_of_int new_index in
      bounds_check_write where 4 what ;
      String.blit new_string 0 buff where 4 ;
      buff

  | TP_PatchReplaceBCSBlock(old_file, new_file) -> begin
      let old_file = Arch.backslash_to_slash old_file in
      let new_file = Arch.backslash_to_slash new_file in
      let bcs_buff_of_baf_or_bcs file =
        let a,b = split (String.uppercase file) in
        if b = "BAF" then begin
          try
            let script = parse_file file "parsing .baf files"
              (Bafparser.baf_file Baflexer.initial) in
            let buff = Buffer.create 1024 in
            Bcs.save_bcs game (Bcs.Save_BCS_Buffer(buff)) script ;
            Buffer.contents buff
          with e ->
            log_and_print "ERROR: error compiling [%s]: %s\n"
              file (Printexc.to_string e) ;
            raise e
        end else begin
          load_file file
        end
      in
      let old_file_buff = bcs_buff_of_baf_or_bcs old_file in
      let string_to_find = body_of_script old_file_buff in
      let new_file_buff = bcs_buff_of_baf_or_bcs new_file in
      let string_to_sub_in = body_of_script new_file_buff in
      let my_regexp = Str.regexp_string string_to_find in
      try
        let _ = Str.search_forward my_regexp buff 0 in
        Str.global_replace my_regexp string_to_sub_in buff
      with Not_found -> 
        log_and_print "WARNING: cannot find block matching [%s]\n"
          old_file ;
        buff
      end 

  | TP_PatchReplaceBCSBlockRE(old_file, new_file) -> begin
      let old_file = Arch.backslash_to_slash old_file in
      let new_file = Arch.backslash_to_slash new_file in
      let string_to_find = load_file old_file in
      let string_to_sub_in = load_file new_file in
      let my_regexp = Str.regexp string_to_find in
      try 
        let _ = Str.search_forward my_regexp buff 0 in
        Str.global_replace my_regexp string_to_sub_in buff 
      with Not_found -> 
        log_and_print "WARNING: cannot find block matching [%s]\n"
          old_file ;
        buff
      end 

  | TP_PatchApplyBCSPatch(patch_file,opt_overwrite) -> begin
      let patch_file = Arch.backslash_to_slash patch_file in
      let patch_buff = load_file patch_file in
      try
        let new_buff, bad_chunks, app_chunks = 
            Diff.do_patch buff patch_buff false in begin
          if ( bad_chunks > 0 ) then begin
            log_or_print 
              "ERROR: Cannot apply patch %s (%d bad chunks).\n" 
              patch_file bad_chunks ;
            match opt_overwrite with
            | Some(newfile) -> load_file (Arch.backslash_to_slash newfile)
            | None -> failwith "Cannot Apply Patch"
          end else begin 
          if ( app_chunks > 0 ) then begin
            log_or_print "WARNING: %d chunks in patch file %s already applied.\n" app_chunks patch_file 
          end ;
          new_buff
          end 
        end
      with Not_found ->
        log_and_print "Error: applying patch %s failed.\n" patch_file ;
        buff
  end

  | TP_PatchInsertBytes(where,how_many) ->
      let how_many = Int32.to_int (eval_pe game how_many) in 
      let blanks = String.make how_many ('\000') in
      let where = Int32.to_int (eval_pe game where) in
      let before = Str.string_before buff where in
      let after = (* try  *)
        Str.string_after buff where
(*         with _ -> "" *)
      in
      before ^ blanks ^ after

  | TP_PatchDeleteBytes(where,how_many) ->
      let where = Int32.to_int (eval_pe game where) in 
      let before = Str.string_before buff where in
      let how_many = Int32.to_int (eval_pe game how_many) in 
      let after = Str.string_after buff (where + how_many) in
      before ^ after

  | TP_PatchSetIdsSymOfInt(var,ids,value) -> 
      let value = (eval_pe game value) in
      let ids = Var.get_string ids in 
      let str = Bcs.sym_of_int game ids value in
      Var.set_string var str ;
      buff

  | TP_PatchSet(name,value) ->
      let value = (eval_pe game value) in
      let name = eval_pe_str name in 
      Var.set_int32 name value ;
      buff

  | TP_PatchInnerAction(al) ->
      List.iter (process_action tp) al ;
      buff

  | TP_PatchInnerBuff(buff_var,pl) ->
      let new_buff = Var.get_string buff_var in
      let filename = Printf.sprintf "INNER_PATCH %S" buff_var in
      let result = List.fold_left (fun acc elt ->
          process_patch2 filename game acc elt) new_buff pl
      in 
      (* intentionally throw away the result buff *)
      buff

  | TP_PatchInnerBuffFile(buff_var,pl) ->
      let buff_var = Arch.backslash_to_slash buff_var in
      let filename = Var.get_string buff_var in
      if is_true (eval_pe game (Pred_File_Exists_In_Game(PE_LiteralString filename))) then begin
        let a,b = split filename in
        let new_buff, path =
          Load.load_resource "INNER_PATCH_FILE" game true a b in
        let filename = Printf.sprintf "INNER_PATCH_FILE %S" buff_var in
        let result = List.fold_left (fun acc elt ->
            process_patch2 filename game acc elt) new_buff pl
        in
        ()
      end ;
      buff

  | TP_PatchInnerBuffSave(store_var,buff_var,pl) ->
      let new_buff = Var.get_string buff_var in
      let filename = Printf.sprintf "INNER_PATCH_SAVE %S" buff_var in
      let result = List.fold_left (fun acc elt ->
          process_patch2 filename game acc elt) new_buff pl
      in
      Var.set_string store_var (Var.get_string result);
      buff

  | TP_PatchRandomSeed(i) ->
      eval_pe_warn := false ;
      let _ = try
          let x = (eval_pe game i) in
          Random.init (Int32.to_int x) ;
        with _ ->
          begin
            Random.self_init () ;
          end
      in
      eval_pe_warn := true ;
      buff

  | TP_PatchIf(pred,tb,eb) ->
      let b = ref buff in
      b := List.fold_left (fun acc elt ->
          process_patch2 patch_filename game acc elt) !b
          (if is_true (eval_pe game pred) then tb else eb) ;
      !b

  | TP_Launch_Patch_Macro(str) ->
      let (decl, actions) =
        try
          Hashtbl.find patch_macros str
        with _ ->
          failwith (Printf.sprintf "Unknown macro: -%s-" str)
(*           ( [] , [] ) *)
      in
      List.iter (fun x ->
        match x with
          TP_LocalSet(var,pe) ->
            let pe = (eval_pe game pe) in
            let var = eval_pe_str var in
            Var.add_local_int32 var pe
        | TP_LocalSprint (var,str) ->
            let var = eval_pe_str var in
            let (str : string) = eval_pe_tlk_str game str in
            let str = Var.get_string str in
            Var.add_local_string var str
        | TP_LocalTextSprint (var,str) ->
            let var = eval_pe_str var in
            let str = Var.get_string (eval_pe_str str) in
            Var.add_local_string var str
      ) decl ;
      let b = ref buff in
      b := List.fold_left (fun acc elt ->
        process_patch2 patch_filename game acc elt) !b actions ;
      List.iter (fun x ->
        match x with
          TP_LocalSet(var,_)
        | TP_LocalSprint(var,_) 
        | TP_LocalTextSprint(var,_)->
          let var = eval_pe_str var in
          Var.remove_local var
      ) decl ;
    !b

  | TP_PatchWhile(pred,pl) ->
      let b = ref buff in
      while is_true (eval_pe game pred) do
        b := List.fold_left (fun acc elt ->
            process_patch2 patch_filename game acc elt) !b pl
      done ;
      !b

  | TP_PatchBashFor(where,pl) ->
      let find_list = ref [] in
      List.iter ( fun (directory,exact_match,regexp_string) ->
        try
          let dh = Case_ins.unix_opendir directory in
          let reg = begin match exact_match with
              | Some(true) ->
                Str.regexp_string_case_fold
              | _ ->
                Str.regexp_case_fold
            end regexp_string
          in
          (try
            while true do
              let next = Unix.readdir dh in
              if ((Case_ins.unix_stat (directory ^ "/" ^ next)).Unix.st_kind =
                 Unix.S_REG) && (Str.string_match reg next 0) then
                find_list := (String.uppercase (directory ^ "/" ^ next)) :: !find_list
            done
          with End_of_file -> () );
          Unix.closedir dh ;
        with _ -> ()
      ) where ;
      let the_buff = ref buff in
      List.iter (fun file ->
        let directory = Filename.dirname file in
        let filespec = Filename.basename file in
        Var.set_string "BASH_FOR_DIRECTORY" directory ;
        Var.set_string "BASH_FOR_FILESPEC" file ;
        Var.set_string "BASH_FOR_FILE" filespec ;
        Var.set_string "BASH_FOR_RES"
          (let a,b = split filespec in a) ;
        the_buff := List.fold_left (fun acc elt ->
              process_patch2 patch_filename game acc elt) !the_buff pl
      ) !find_list ;
      !the_buff

  | TP_PatchForEach(var,sl,pl) ->
      let var = eval_pe_str var in
      let sl = List.map Var.get_string sl in
      let the_buff = ref buff in
      List.iter (fun x ->
        Var.set_string var x ;
        the_buff := List.fold_left (fun acc elt ->
              process_patch2 patch_filename game acc elt) !the_buff pl
      ) sl ;
      !the_buff

  | TP_PatchFor(init,guard,inc,body) ->
      let cmd_list = init @ [ TP_PatchWhile(guard,body @ inc) ] in
      let b = ref buff in
      b := List.fold_left (fun acc elt ->
          process_patch2 patch_filename game acc elt) !b cmd_list  ;
      !b

  | TP_PatchSilent -> be_silent := true ; buff
  | TP_PatchVerbose -> be_silent := false ; buff

  | TP_Patch2DALater(str,r,c,value) ->
      let row = (eval_pe game r) in
      let col = (eval_pe game c) in
      let str = Var.get_string str in
      eval_pe_warn := false ;
      let value =
        try
          let x = (eval_pe game value) in
          Int32.to_string x
        with _ ->
          begin
            match value with
            | PE_String(x) -> Var.get_string (eval_pe_str x)
            | _ -> (eval_pe_warn := true ; ignore (eval_pe game value) ; "")
          end
      in
      let iidx = try Var.get_int32 ("%" ^ str ^ "%") with _ -> 0l in
      let idx = Int32.to_string iidx in
      let sidx = str ^ idx in
      Var.set_int32 (sidx ^ "r" ) (row) ;
      Var.set_int32 (sidx ^ "c" ) (col) ;
      Var.set_string (sidx ^ "v" ) value ;
      Var.set_int32 str (Int32.add iidx 1l) ;
      buff

  | TP_Patch2DANow(str, req_col) ->
      let req_col = Int32.to_int (eval_pe game req_col) in
      let str = Var.get_string str in
      let lines = Str.split many_newline_or_cr_regexp buff in
      let max = ref 4 in
      let entries = List.map (fun line ->
        let entry_list = Str.split many_whitespace_regexp line in
        List.iter (fun e ->
          let len = String.length e in
          if len > !max then max := len
        ) entry_list ;
        Array.of_list entry_list) lines
      in
      let updates = Hashtbl.create 255 in
      begin 
      try 
        let count = Int32.to_int (Var.get_int32 ("%" ^ str ^ "%")) in
        Var.set_int32 str 0l ; 
        for i = 0 to pred count do
          let sidx = Printf.sprintf "%%%s%d" str i in
          let row = Int32.to_int (Var.get_int32 (sidx ^ "r%")) in
          let col = Int32.to_int (Var.get_int32 (sidx ^ "c%")) in
          let value = Var.get_string (sidx ^ "v%") in
          Hashtbl.add updates (row,col) value
        done ;
        let buf = Buffer.create (String.length buff) in
        let this_line = ref 0 in 
        List.iter (fun entry_array ->
          let this_col_size = Array.length entry_array in 
          Array.iteri (fun this_col orig_elt -> 
            let elt = 
              if this_col_size >= req_col &&
                 Hashtbl.mem updates (!this_line,this_col) then
                 Hashtbl.find updates (!this_line,this_col)
              else orig_elt 
            in
            Printf.bprintf buf "%-.*s " !max elt
          ) entry_array ;
          Buffer.add_char buf '\n' ;
          if this_col_size >= req_col then incr this_line ;
        ) entries ;
        Buffer.contents buf
      with _ -> buff
      end

  | TP_Read2DANow(str, req_col) ->
      let req_col = Int32.to_int (eval_pe game req_col) in
      let lines = Str.split many_newline_or_cr_regexp buff in
      let str = Var.get_string str in
      let i = ref 0 in
      let j = ref 0 in (* line and column counters *)
      List.iter (fun this_line ->
        let this_line = Str.split many_whitespace_regexp this_line in
        if (List.length this_line) >= req_col then begin
          j := 0;
          List.iter ( fun item ->
            let name = Printf.sprintf "%s_%d_%d" str !i !j in
            Var.set_string name item;
            j := !j + 1;
          ) this_line;
          i := !i + 1;
        end
      ) lines;
      let name = Printf.sprintf "%s" str in
      Var.set_int32 name (Int32.of_int !i);
      buff

  | TP_Read2DAFormer(str, row, col, var) ->
    let row = Int32.to_int (eval_pe game row) in
    let col = Int32.to_int (eval_pe game col) in
    let str = Var.get_string str in
    let str = Printf.sprintf "%%%s_%d_%d%%" str row col in
    if !debug_ocaml then log_and_print "%s reports to %s\n" str (Var.get_string str) ;
    Var.set_string var (Var.get_string_exact str);
    buff

  | TP_PatchInsert2DARow(row,req_col,value) ->
      let row = Int32.to_int (eval_pe game row) in
      let req_col = Int32.to_int (eval_pe game req_col) in
      let lines = Str.split many_newline_or_cr_regexp buff in
      eval_pe_warn := false ;
      let value =
        try
          let x = (eval_pe game value) in
          Int32.to_string x
        with _ ->
          begin
            match value with
            | PE_String(x) -> Var.get_string (eval_pe_str x)
            | _ -> (eval_pe_warn := true ; ignore (eval_pe game value) ; "")
          end
      in
      let before = ref "" in
      let num_rows = ref 0 in
      let found = ref false in
      let avoid_first_nl = ref true in
      List.iter ( fun the_row ->
        before := !before ^ (if !avoid_first_nl then "" else "\r\n") ;
        if !avoid_first_nl then avoid_first_nl := false ;
        if !num_rows < row then begin
          let long_enough = ((List.length (Str.split many_whitespace_regexp the_row)) >= req_col) in
          if long_enough then num_rows := !num_rows + 1 ;
        end else
        if !num_rows = row then begin
          let long_enough = ((List.length (Str.split many_whitespace_regexp the_row)) >= req_col) in
          if long_enough then begin
            num_rows := !num_rows + 1 ;
            before := !before ^ value ^ "\r\n" ;
            found := true ;
          end ;
        end ;
        before := !before ^ the_row
      ) lines ;
      if !num_rows = row then begin
        before := !before ^ "\r\n" ^ value ;
        found := true ;
      end ;
      if not (!found) then failwith "Failed to find enough lines to perform an INSERT_2DA_ROW" ;
      !before ^ "\r\n" ;

  | TP_Patch2DA(row,col,req_col,value) ->
      let row = Int32.to_int (eval_pe game row) in
      let col = Int32.to_int (eval_pe game col) in
      let req_col = Int32.to_int (eval_pe game req_col) in
      let lines = Str.split many_newline_or_cr_regexp buff in

      eval_pe_warn := false ;
      let value =
        try
          let x = (eval_pe game value) in
          Int32.to_string x
        with _ ->
          begin
            match value with
            | PE_String(x) -> Var.get_string (eval_pe_str x)
            | _ -> (eval_pe_warn := true ; ignore (eval_pe game value) ; "")
          end
      in 
      let slv = String.length value in
      let max = ref slv in 
      let entries = List.map (fun line -> 
        let entry_list = Str.split many_whitespace_regexp line in
        List.iter (fun e ->
          let len = String.length e in
          if len > !max then max := len
        ) entry_list ;
        Array.of_list entry_list) lines 
      in
      let newlines_sofar = ref 0 in
      let rec process line_list lines_left = 
        match line_list, lines_left with
          [], 0 -> ()
        | [], _ -> 
          begin
            log_and_print "ERROR: Cannot find %d rows with at least %d columns." row req_col ;
            failwith "Cannot Set 2DA Entry" 
          end
        | hd :: tl, i  when Array.length hd < req_col -> 
            incr newlines_sofar ; 
            process tl i
        | hd :: tl, 0 ->
            hd.(col) <- value 
        | hd :: tl, i -> begin
            incr newlines_sofar ;
            process tl (i - 1)
            end 
      in
      process entries row ;
      (try begin 
        let pos = ref 0 in 
        for n = 1 to !newlines_sofar do
          pos := (Str.search_forward many_newline_or_cr_regexp buff !pos) ;
          pos := Str.match_end () ;
        done ;
        for n = 1 to col do
          pos := (Str.search_forward many_whitespace_regexp buff !pos) ;
          pos := Str.match_end () ; 
        done ;
        let before = !pos in 
        let after = Str.search_forward many_not_whitespace_regexp buff !pos in
        let after = Str.match_end () in
        let before_str = Str.string_before buff before in
        let after_str = Str.string_after buff after in
        Printf.sprintf "%s%s%s" before_str value after_str
      end with _ -> begin 
        log_or_print "WARNING: Fast SET_2DA_ENTRY failed, falling back on old method...\n" ;
        let buf = Buffer.create (slv * 2) in 
        List.iter (fun entry_array ->
          Array.iter (fun elt -> 
            Printf.bprintf buf "%-.*s\t" !max elt
          ) entry_array ;
          Buffer.add_char buf '\n' 
        ) entries ;
        Buffer.contents buf 
        end )
      (*
      List.fold_left (fun line_acc entry_array -> 
        line_acc ^ (Array.fold_left (fun acc elt ->
          Printf.sprintf "%s%-.*s\t" acc !max elt) "" entry_array) ^ "\n") 
        "" entries
        *)

  | TP_PatchStringTextually(case_sens,match_exact,find,what,length) ->
      let find = Var.get_string find in
      let what = Var.get_string what in
      let find_ref = ref find in
      let what_ref = ref what in
      let case_sens = match case_sens with
          None -> false
        | Some(x) -> x
      in
      let match_exact = match match_exact with
          None -> false
        | Some(x) -> x
      in
      begin match length with
          None -> ()
        | Some(x) -> begin
          let y = Int32.to_int (eval_pe game x) in
          while String.length !what_ref < y do
            what_ref := !what_ref ^ (String.make 1 (Char.chr 0)) ;
          done ;
          while String.length !find_ref < y do
            find_ref := !find_ref ^ (String.make 1 (Char.chr 0)) ;
          done ;
          if !debug_ocaml then begin
            for i = 0 to ( String.length !what_ref) - 1 do
              let from_c = String.get !find_ref i in
              let what_c = String.get !what_ref i in
              log_and_print "%c[%d] -> %c[%d] \n" from_c (Char.code from_c) what_c (Char.code what_c) ;
            done
          end ;
        end ;
      end ;
      let find = !find_ref in
      let my_regexp = match case_sens, match_exact with
          false, false -> Str.regexp_case_fold        find
        | true , false -> Str.regexp                  find
        | false, true  -> Str.regexp_string_case_fold find
        | true , true  -> Str.regexp_string           find
      in
      let what = !what_ref in
      Str.global_replace my_regexp what buff

  | TP_PatchStringEvaluate(case_sens,find, pl, replace) ->
      (* REPLACE_EVALUATE ~Give(\([0-9]+\))~ 
                          SET "RESULT" = "%MATCH1%"
                          ~Give(%RESULT%)~
         ( "%MATCH%" / 2 ) *)
      let my_regexp = match case_sens with
          None -> Str.regexp find
        | Some(true) -> Str.regexp find
        | Some(false) -> Str.regexp_case_fold find
      in
      let i = ref 0 in
      let work_buff = ref buff in
      (try
        while true do
          let start_idx = Str.search_forward my_regexp !work_buff !i in 
          for j = 0 to 20 do
            let v = Printf.sprintf "MATCH%d" j in 
            (* Var.remove_var v ;  *) 
            (try let group = Str.matched_group j !work_buff in
                 Var.set_string v group ; 
             with _ -> ()) 
          done ; 
          (try 
            ignore (
              List.fold_left (fun acc elt -> 
                process_patch2 patch_filename game acc elt) !work_buff pl
            ) ;
           with _ -> ());
          let this_replacement = Var.get_string replace in
          let old_before = Str.string_before !work_buff start_idx in
          let old_after = Str.string_after !work_buff start_idx in 
          let new_after = 
            Str.replace_first my_regexp this_replacement old_after
          in
          work_buff := (old_before ^ new_after) ;
          i := start_idx + 1
        done
      with _ -> () ) ;
      !work_buff

  | TP_PatchString(case_sens,match_exact,find,what) ->
      let find = Var.get_string find in
      let case_sens = match case_sens with
          None -> false
        | Some(x) -> x
      in
      let match_exact = match match_exact with
          None -> false
        | Some(x) -> x
      in
      let my_regexp = match case_sens, match_exact with
          false, false -> Str.regexp_case_fold        find
        | true , false -> Str.regexp                  find
        | false, true  -> Str.regexp_string_case_fold find
        | true , true  -> Str.regexp_string           find
      in
      let what = begin match what with
      | Dlg.Trans_String(a) -> what
      | Dlg.Local_String(lse) ->
          Dlg.Local_String ({ lse_male = Var.get_string lse.lse_male ;
                               lse_male_sound = Var.get_string lse.lse_male_sound ;
                                   lse_female = Var.get_string lse.lse_female ;
                             lse_female_sound = Var.get_string lse.lse_female_sound ; })
      | _ -> what
      end
      in
      let new_index = match Dc.resolve_tlk_string game what with
        Dlg.TLK_Index(i) -> i
      | _ -> log_and_print "ERROR: cannot resolve REPLACE patch\n" ; failwith "resolve"
      in 
      let new_string = Printf.sprintf "%d" new_index in
      Str.global_replace my_regexp new_string buff 

  | TP_PatchByte(where,what) -> 
      let where = Int32.to_int (eval_pe game where) in
      let what = Int32.to_int (eval_pe game what) in 
      let what = if what < 0 then what + 256 else what in
      let str = String.make 1 (Char.chr what) in
      bounds_check_write where 1 str ; 
      String.blit str 0 buff where 1 ;
      buff
  | TP_PatchShort(where,what) -> 
      let where = Int32.to_int (eval_pe game where) in 
      let what = Int32.to_int (eval_pe game what) in 
      let str = str_of_short what in
      bounds_check_write where 2 str ; 
      String.blit str 0 buff where 2 ;
      buff
  | TP_PatchLong(where,what) -> 
      let where = Int32.to_int (eval_pe game where) in
      let what = eval_pe game what in
      let str = str_of_int32 what in 
      bounds_check_write where 4 str ;
      String.blit str 0 buff where 4 ;
      buff
  | TP_PatchAppendFile(filename, eval) ->
      let filename = Arch.backslash_to_slash filename in
      let filename = Var.get_string filename in
      let file_buff = load_file filename in
      buff ^ (if eval then Var.get_string file_buff else file_buff)

  | TP_PatchWriteFile(where,filename,insert) ->
      let filename = Arch.backslash_to_slash filename in
      let where = Int32.to_int (eval_pe game where) in 
      let filename = Var.get_string filename in 
      let file_buff = load_file filename in
      if insert then begin
        let before = Str.string_before buff where in
        let after = Str.string_after buff where in
        before ^ file_buff ^ after
      end else begin
        let file_len = String.length file_buff in
        let len = String.length buff in
        if (len - where) < file_len then begin 
          log_and_print "Not enough room for [%s] (%d bytes) if you start at position %d in a %d byte buffer!" (filename) (file_len) (where) (len) ;
          failwith "ERROR: cannot process WRITE_FILE"
        end ;
        String.blit file_buff 0 buff where file_len ;
        buff
      end 

  | TP_PatchASCII(where,what,evaluate,reqdSize) -> 
      let where = Int32.to_int (eval_pe game where) in
      let what = if evaluate then Var.get_string what 
                             else what in 
      let what = match reqdSize with
      | None -> what
      | Some(i) -> 
        let res = String.make i '\000' in
        String.blit what 0 res 0 (min i (String.length what)) ;
        res
      in
      bounds_check_write where (String.length what) what ;
      String.blit what 0 buff where (String.length what) ;
      buff

  | TP_Patch_Gam(cre_name,area,x,y) ->

        (* Patch baldur.gam in save directories *)

        let dlist = list_of_files_in_directory "save" in
        List.iter
        (fun filename ->
        let filename = "save/" ^ filename in
        if is_directory filename && filename <> "save/." && filename <> "save/.." then begin
        log_and_print "Patching baldur.gam in directory %s\n" filename;
        let baldur_buff = load_file (filename ^ "/baldur.gam") in
        if String.sub baldur_buff 0 8 <> "GAMEV1.1" then begin
        failwith "not a valid GAME V1.1 file (wrong sig)"
        end ;
        let non_party_npc_off = int_of_str_off baldur_buff 0x30 in
        let num_non_party_npc = int_of_str_off baldur_buff 0x34 in
        let variable_off = int_of_str_off baldur_buff 0x38 in
        let journal_off = int_of_str_off baldur_buff 0x50 in
        let last_npc_off = (non_party_npc_off + ((num_non_party_npc) * 352)) in

        (* Create the non-party NPC entry *)

        let npc_entry_buff = String.make 352 '\000' in
        write_short npc_entry_buff 0x2 0xFFFF;
        write_int npc_entry_buff 0x4 (variable_off + 352);
        write_int npc_entry_buff 0x8 (String.length buff);
        String.blit area 0 npc_entry_buff 0x18 (String.length area);
        write_short npc_entry_buff 0x20 x;
        write_short npc_entry_buff 0x22 y;

        (* Update the offsets *)

        write_int baldur_buff 0x38 (variable_off + (String.length buff) + 352);
        write_int baldur_buff 0x50 (journal_off + (String.length buff) + 352);
        write_int baldur_buff 0x34 (num_non_party_npc + 1);

        (* Splice in the non-party NPC entry and the cre file *)

        for i = 0 to num_non_party_npc - 1 do
        let cre_off = int_of_str_off baldur_buff (non_party_npc_off + (i * 352) + 0x4) in
        write_int baldur_buff (non_party_npc_off + (i * 352) + 0x4) (cre_off + 352)
        done;

        let before = Str.string_before baldur_buff last_npc_off in
        let after = Str.string_after baldur_buff variable_off in
        let cre_chunk = String.sub baldur_buff last_npc_off (variable_off - last_npc_off) in
        let new_buff = before ^ npc_entry_buff ^ cre_chunk ^ buff ^ after in

        let oc = Case_ins.perv_open_out_bin (filename ^ "/baldur.gam") in
        output_string oc new_buff;
        close_out oc;
    begin (* handle read-only files! *)
      try
        Case_ins.unix_chmod (filename ^ "/baldur.gam") 511 ; (* 511 = octal 0777 = a+rwx *)
      with e -> ()
        (* log_or_print "WARNING: chmod %s : %s\n" filename 
          (Printexc.to_string e) *)
    end ;
        end) dlist;

        (* Patch baldur.gam in mpsave directories *)

        let mpdlist = list_of_files_in_directory "mpsave" in
        List.iter
        (fun filename ->
        let filename = "mpsave/" ^ filename in
        if is_directory filename && filename <> "mpsave/." && filename <> "mpsave/.." then begin
        log_and_print "Patching baldur.gam in directory %s\n" filename;
        let baldur_buff = load_file (filename ^ "/baldur.gam") in
        if String.sub baldur_buff 0 8 <> "GAMEV1.1" then begin
        failwith "not a valid GAME V1.1 file (wrong sig)"
        end ;
        let non_party_npc_off = int_of_str_off baldur_buff 0x30 in
        let num_non_party_npc = int_of_str_off baldur_buff 0x34 in
        let variable_off = int_of_str_off baldur_buff 0x38 in
        let journal_off = int_of_str_off baldur_buff 0x50 in
        let last_npc_off = (non_party_npc_off + ((num_non_party_npc) * 352)) in

        (* Create the non-party NPC entry *)

        let npc_entry_buff = String.make 352 '\000' in
        write_short npc_entry_buff 0x2 0xFFFF;
        write_int npc_entry_buff 0x4 (variable_off + 352);
        write_int npc_entry_buff 0x8 (String.length buff);
        String.blit area 0 npc_entry_buff 0x18 (String.length area);
        write_short npc_entry_buff 0x20 x;
        write_short npc_entry_buff 0x22 y;

        (* Update the offsets *)

        write_int baldur_buff 0x38 (variable_off + (String.length buff) + 352);
        write_int baldur_buff 0x50 (journal_off + (String.length buff) + 352);
        write_int baldur_buff 0x34 (num_non_party_npc + 1);

        (* Splice in the non-party NPC entry and the cre file *)

        for i = 0 to num_non_party_npc - 1 do
        let cre_off = int_of_str_off baldur_buff (non_party_npc_off + (i * 352) + 0x4) in
        write_int baldur_buff (non_party_npc_off + (i * 352) + 0x4) (cre_off + 352)
        done;

        let before = Str.string_before baldur_buff last_npc_off in
        let after = Str.string_after baldur_buff variable_off in
        let cre_chunk = String.sub baldur_buff last_npc_off (variable_off - last_npc_off) in
        let new_buff = before ^ npc_entry_buff ^ cre_chunk ^ buff ^ after in

        let oc = Case_ins.perv_open_out_bin (filename ^ "/baldur.gam") in
        output_string oc new_buff;
        close_out oc;
        begin (* handle read-only files! *)
          try
            Case_ins.unix_chmod (filename ^ "/baldur.gam") 511 ; (* 511 = octal 0777 = a+rwx *)
          with e -> ()
            (* log_or_print "WARNING: chmod %s : %s\n" filename
              (Printexc.to_string e) *)
        end ;
        end) mpdlist;

        (* Patch the baldur.gam in the biff *)

        log_and_print "Patching starting baldur.gam (for new games)...\n";
        let nbaldur_buff,path = Load.load_resource "PATCH_GAM" game true "BALDUR" "GAM" in

        if String.sub nbaldur_buff 0 8 <> "GAMEV1.1" then begin
        failwith "not a valid GAME V1.1 file (wrong sig)"
        end ;
        let party_npc_off = int_of_str_off nbaldur_buff 0x20 in
        let journal_off = int_of_str_off nbaldur_buff 0x50 in
        let num_non_party_npc = int_of_str_off nbaldur_buff 0x34 in

        (* Create the non-party NPC entry *)

        let npc_entry_buff = String.make 352 '\000' in
        String.blit (String.uppercase cre_name) 0 npc_entry_buff 0xc (String.length cre_name);
        String.blit area 0 npc_entry_buff 0x18 (String.length area);
        write_short npc_entry_buff 0x20 x;
        write_short npc_entry_buff 0x22 y;
        for i = 140 to 147 do
        npc_entry_buff.[i] <- Char.chr 0xFF
        done;
        for i = 180 to 185 do
        npc_entry_buff.[i] <- Char.chr 0xFF
        done;

        (* Update all offsets *)

        write_int nbaldur_buff 0x20 (party_npc_off + 352);
        write_int nbaldur_buff 0x28 (party_npc_off + 352);
        write_int nbaldur_buff 0x50 (journal_off + 352);
        write_int nbaldur_buff 0x34 (num_non_party_npc + 1);

        (* Splice in the non-party NPC entry *)

        let before = Str.string_before nbaldur_buff party_npc_off in
        let after = Str.string_after nbaldur_buff party_npc_off in
        let new_buff = before ^ npc_entry_buff ^ after in

        let oc = Case_ins.perv_open_out_bin "override/baldur.gam" in
        output_string oc new_buff;
        close_out oc;
        begin (* handle read-only files! *)
          try
            Case_ins.unix_chmod "override/baldur.gam" 511 ; (* 511 = octal 0777 = a+rwx *)
          with e -> ()
            (* log_or_print "WARNING: chmod %s : %s\n" filename
              (Printexc.to_string e) *)
        end ;
        buff

  | TP_Remove_Known_Spell(sp_list) ->

      (* Check for version *)
      let sp_list = List.map Var.get_string sp_list in
      let cre_v = String.sub buff 0 8 in
      (match cre_v with
      | "CRE V1.0" ->

        let knownspelloff = int_of_str_off buff 0x2a0 in
        let spellmemoff = int_of_str_off buff 0x2a8 in
        let numknownspells = int_of_str_off buff 0x2a4 in

        (* Grab all spells *)

        let spell_list = ref [] in
        let spell_string = String.sub buff knownspelloff (spellmemoff - knownspelloff) in
        for i = 0 to numknownspells - 1 do
        let spell_check = String.sub spell_string (i * 12) 8 in
        let spell_check = get_string_of_size spell_check 0 8 in
        spell_list := (String.uppercase spell_check) :: !spell_list
        done;

        spell_list := List.rev !spell_list;

        let i = ref 0 in
        let match_count = ref 0 in
        let work_buff = ref buff in
        List.iter (fun spell_name ->
        if List.mem spell_name sp_list then begin
        let spellmemoff = int_of_str_off !work_buff 0x2a8 in
        let memspelloff = int_of_str_off !work_buff 0x2b0 in
        let itemsslotoff = int_of_str_off !work_buff 0x2b8 in
        let itemsoff = int_of_str_off !work_buff 0x2bc in
        let effoff = int_of_str_off !work_buff 0x2c4 in
        let numknownspells = int_of_str_off !work_buff 0x2a4 in
        write_int !work_buff 0x2a8 (spellmemoff - 12);
        write_int !work_buff 0x2b0 (memspelloff - 12);
        write_int !work_buff 0x2b8 (itemsslotoff - 12);
        write_int !work_buff 0x2bc (itemsoff - 12);
        write_int !work_buff 0x2c4 (effoff - 12);
        write_int !work_buff 0x2a4 (numknownspells - 1);
        let before_buff = Str.string_before !work_buff (knownspelloff + ((!i - !match_count) * 12)) in
        let after_buff = Str.string_after !work_buff (knownspelloff + ((!i - !match_count) * 12) + 12) in
        work_buff := before_buff ^ after_buff ;
        incr i ;
        incr match_count
        end else
        incr i) !spell_list ;

        !work_buff

      | "CRE V9.0" ->

        let knownspelloff = int_of_str_off buff 0x308 in
        let spellmemoff = int_of_str_off buff 0x310 in
        let numknownspells = int_of_str_off buff 0x30c in

        (* Grab all spells *)

        let spell_list = ref [] in
        let spell_string = String.sub buff knownspelloff (spellmemoff - knownspelloff) in
        for i = 0 to numknownspells - 1 do
        let spell_check = String.sub spell_string (i * 12) 8 in
        let spell_check = get_string_of_size spell_check 0 8 in
        spell_list := (String.uppercase spell_check) :: !spell_list
        done;

        spell_list := List.rev !spell_list;

        let i = ref 0 in
        let match_count = ref 0 in
        let work_buff = ref buff in
        List.iter (fun spell_name ->
        if List.mem spell_name sp_list then begin
        let spellmemoff = int_of_str_off !work_buff 0x310 in
        let memspelloff = int_of_str_off !work_buff 0x318 in
        let itemsslotoff = int_of_str_off !work_buff 0x320 in
        let itemsoff = int_of_str_off !work_buff 0x324 in
        let effoff = int_of_str_off !work_buff 0x32c in
        let numknownspells = int_of_str_off !work_buff 0x30c in
        write_int !work_buff 0x310 (spellmemoff - 12);
        write_int !work_buff 0x318 (memspelloff - 12);
        write_int !work_buff 0x320 (itemsslotoff - 12);
        write_int !work_buff 0x324 (itemsoff - 12);
        write_int !work_buff 0x32c (effoff - 12);
        write_int !work_buff 0x30c (numknownspells - 1);
        let before_buff = Str.string_before !work_buff (knownspelloff + ((!i - !match_count) * 12)) in
        let after_buff = Str.string_after !work_buff (knownspelloff + ((!i - !match_count) * 12) + 12) in
        work_buff := before_buff ^ after_buff ;
        incr i ;
        incr match_count
        end else
        incr i) !spell_list ;

        !work_buff

     | _ -> failwith (Printf.sprintf "ERROR: REMOVE_KNOWN_SPELL: Unknown cre version %s" cre_v));

  | TP_Remove_Memorized_Spell(sp_list) ->
    if !debug_ocaml then List.iter (fun x -> log_and_print "Trying to remove %s\n" x) sp_list ;
    let sp_list = List.map (fun x -> String.uppercase(Var.get_string x)) sp_list in
    if !debug_ocaml then List.iter (fun x -> log_and_print "Trying to remove %s\n" x) sp_list ;

      (* Check for version *)

      let cre_v = String.sub buff 0 8 in
      (match cre_v with
      | "CRE V1.0" ->

        let spellmemoff = int_of_str_off buff 0x2b0 in
        let nummempells = int_of_str_off buff 0x2b4 in
        let effectsoff  = int_of_str_off buff 0x2c4 in

        (* Grab all spells *)

        let spell_list = ref [] in
        let spell_string = String.sub buff spellmemoff (effectsoff - spellmemoff) in
        for i = 0 to nummempells - 1 do
          let spell_check = String.sub spell_string (i * 12) 8 in
          let spell_check = get_string_of_size spell_check 0 8 in
          spell_list := (String.uppercase spell_check) :: !spell_list
        done;

        spell_list := List.rev !spell_list;

        let i = ref 0 in
        let match_count = ref 0 in
        let work_buff = ref buff in
        List.iter (fun spell_name ->
          if !debug_ocaml then log_and_print "I'm doing %s\n" spell_name ;
          if List.mem spell_name sp_list then begin
          let itemsslotoff = int_of_str_off !work_buff 0x2b8 in
          let itemsoff = int_of_str_off !work_buff 0x2bc in
          let effoff = int_of_str_off !work_buff 0x2c4 in
          let nummemspells = int_of_str_off !work_buff 0x2b4 in
          let memoroff = int_of_str_off buff 0x2a8 in
          let memornum = int_of_str_off buff 0x2ac in
          write_int !work_buff 0x2b8 (itemsslotoff - 12);
          write_int !work_buff 0x2bc (itemsoff - 12);
          write_int !work_buff 0x2c4 (effoff - 12);
          write_int !work_buff 0x2b4 (nummemspells - 1);
          let before_buff = Str.string_before !work_buff (spellmemoff + ((!i - !match_count) * 12)) in
          let after_buff = Str.string_after !work_buff (spellmemoff + ((!i - !match_count) * 12) + 12) in
          work_buff := before_buff ^ after_buff ;
          for j = 0 to memornum - 1 do
            let leveloff = int_of_str_off !work_buff (memoroff + 16 * j + 8) in
            let levelnum = int_of_str_off !work_buff (memoroff + 16 * j + 12) in
            if leveloff > !i - !match_count then write_int !work_buff (memoroff + 16 * j + 8) (leveloff - 1)
              else if leveloff + levelnum > !i - !match_count then
                  write_int !work_buff (memoroff + 16 * j + 12) (levelnum - 1) ;
          done ;
          incr i ;
          incr match_count
        end else
        incr i) !spell_list ;

        !work_buff

      | "CRE V9.0" ->

        let spellmemoff = int_of_str_off buff 0x318 in
        let nummempells = int_of_str_off buff 0x31c in
        let effectsoff  = int_of_str_off buff 0x32c in

        (* Grab all spells *)

        let spell_list = ref [] in
        let spell_string = String.sub buff spellmemoff (effectsoff - spellmemoff) in
        for i = 0 to nummempells - 1 do
          let spell_check = String.sub spell_string (i * 12) 8 in
          let spell_check = get_string_of_size spell_check 0 8 in
          spell_list := (String.uppercase spell_check) :: !spell_list
        done;

        spell_list := List.rev !spell_list;

        let i = ref 0 in
        let match_count = ref 0 in
        let work_buff = ref buff in
        List.iter (fun spell_name ->
          if !debug_ocaml then log_and_print "I'm doing %s\n" spell_name ;
          if List.mem spell_name sp_list then begin
          let itemsslotoff = int_of_str_off !work_buff 0x320 in
          let itemsoff = int_of_str_off !work_buff 0x324 in
          let effoff = int_of_str_off !work_buff 0x32c in
          let nummemspells = int_of_str_off !work_buff 0x31c in
          let memoroff = int_of_str_off buff 0x310 in
          let memornum = int_of_str_off buff 0x314 in
          write_int !work_buff 0x320 (itemsslotoff - 12);
          write_int !work_buff 0x324 (itemsoff - 12);
          write_int !work_buff 0x32c (effoff - 12);
          write_int !work_buff 0x31c (nummemspells - 1);
          let before_buff = Str.string_before !work_buff (spellmemoff + ((!i - !match_count) * 12)) in
          let after_buff = Str.string_after !work_buff (spellmemoff + ((!i - !match_count) * 12) + 12) in
          work_buff := before_buff ^ after_buff ;
          for j = 0 to memornum - 1 do
            let leveloff = int_of_str_off !work_buff (memoroff + 16 * j + 8) in
            let levelnum = int_of_str_off !work_buff (memoroff + 16 * j + 12) in
            if leveloff > !i - !match_count then write_int !work_buff (memoroff + 16 * j + 8) (leveloff - 1)
              else if leveloff + levelnum > !i - !match_count then
                  write_int !work_buff (memoroff + 16 * j + 12) (levelnum - 1) ;
          done ;
          incr i ;
          incr match_count
        end else
        incr i) !spell_list ;

        !work_buff

     | _ -> failwith (Printf.sprintf "ERROR: REMOVE_KNOWN_SPELL: Unknown cre version %s" cre_v));

  | TP_Add_Known_Spell(spell,level,sp_type) ->

      (* Create the new known spell *)
      let spell = Var.get_string spell in
      let spell_buff = String.make 12 '\000' in
      String.blit spell 0 spell_buff 0 (String.length spell);
      write_short spell_buff 8 level;

      let new_type =
      (match (String.uppercase sp_type) with
      | "PRIEST" -> 0
      | "INNATE" -> 2
      | "WIZARD" -> 1
      | _ -> log_and_print "WARNING: ADD_NEW_SPELL: Unknown flag %s.  Defaulting to INNATE for spell type.\n" (String.uppercase sp_type);
             2)
      in

      write_short spell_buff 10 new_type ;

      (* Check for version *)
      
      let cre_v = String.sub buff 0 8 in
      (match cre_v with
      | "CRE V1.0" ->

        let spellmemoff = int_of_str_off buff 0x2a8 in
        let memspelloff = int_of_str_off buff 0x2b0 in
        let itemsslotoff = int_of_str_off buff 0x2b8 in
        let itemsoff = int_of_str_off buff 0x2bc in
        let effoff = int_of_str_off buff 0x2c4 in
        write_int buff 0x2a8 (spellmemoff + 12);
        write_int buff 0x2b0 (memspelloff + 12);
        write_int buff 0x2b8 (itemsslotoff + 12);
        write_int buff 0x2bc (itemsoff + 12);
        write_int buff 0x2c4 (effoff + 12);
        let knownspelloff = int_of_str_off buff 0x2a0 in
        let numknownspells = int_of_str_off buff 0x2a4 in
        write_int buff 0x2a4 (numknownspells + 1);
        let before_buff = Str.string_before buff knownspelloff in
        let after_buff = Str.string_after buff knownspelloff in
        before_buff ^ spell_buff ^ after_buff

     | "CRE V9.0" ->

        let spellmemoff = int_of_str_off buff 0x310 in
        let memspelloff = int_of_str_off buff 0x318 in
        let itemsslotoff = int_of_str_off buff 0x320 in
        let itemsoff = int_of_str_off buff 0x324 in
        let effoff = int_of_str_off buff 0x32c in
        write_int buff 0x310 (spellmemoff + 12);
        write_int buff 0x318 (memspelloff + 12);
        write_int buff 0x320 (itemsslotoff + 12);
        write_int buff 0x324 (itemsoff + 12);
        write_int buff 0x32c (effoff + 12);
        let knownspelloff = int_of_str_off buff 0x308 in
        let numknownspells = int_of_str_off buff 0x30c in
        write_int buff 0x30c (numknownspells + 1);
        let before_buff = Str.string_before buff knownspelloff in
        let after_buff = Str.string_after buff knownspelloff in
        before_buff ^ spell_buff ^ after_buff

     | _ -> failwith (Printf.sprintf "ERROR: ADD_NEW_SPELL: Unknown cre version %s" cre_v));

  | TP_Add_Map_Note(m) ->

      (* Turn the colours into something usable *)

      let colour_flags =
      (match (String.uppercase m.colour) with
      | "GRAY" -> 0
      | "VIOLET" -> 1
      | "GREEN" -> 2
      | "ORANGE" -> 3
      | "RED" -> 4
      | "BLUE" -> 5
      | "DARKBLUE" -> 6
      | "LIGHTGRAY" -> 7
      | _ -> log_and_print "WARNING: ADD_MAP_NOTE: Unknown flag %s.  Defaulting to GRAY for flags.\n" (String.uppercase m.colour);
             0) in

      (* Create the new automap *)

      let auto_buff = String.make 52 '\000' in
      write_short auto_buff 0 m.xcoord;
      write_short auto_buff 2 m.ycoord;
      write_byte auto_buff 8 1;
      write_byte auto_buff 10 colour_flags;

      let new_index = match Dc.resolve_tlk_string game m.mstring with
          Dlg.TLK_Index(i) -> i
      | _ -> log_and_print "ERROR: ADD_MAP_NOTE: cannot resolve SAY patch\n" ; failwith "resolve"
      in
      let new_string = str_of_int new_index in
      String.blit new_string 0 auto_buff 4 4 ;

      (* Check for version *)

        let area_v = String.sub buff 0 8 in
      (match area_v with
       | "AREAV1.0" ->
         let automap_off = int_of_str_off buff 0xc4 in
         let auto_num = int_of_str_off buff 0xc8 in
         if automap_off = 0 then begin
         let new_automap_off = String.length buff in
         write_int buff 0xc4 new_automap_off;
         write_int buff 0xc8 (auto_num + 1)
         end else
         write_int buff 0xc8 (auto_num + 1)
       | "AREAV9.1" ->
         let automap_off = int_of_str_off buff 0xd4 in
         let auto_num = int_of_str_off buff 0xd8 in
         if automap_off = 0 then begin
         let new_automap_off = String.length buff in
         write_int buff 0xd4 new_automap_off;
         write_int buff 0xd8 (auto_num + 1)
         end else
         write_int buff 0xd8 (auto_num + 1)
       | _ -> failwith (Printf.sprintf "ADD_MAP_NOTE: Unknown area version %s" area_v));

      (* Splice it all together *)

      buff ^ auto_buff

  | TP_PatchPrint(msg) ->
        let str = Dc.single_string_of_tlk_string game msg in
        be_silent := false ;
        let str = Var.get_string str in
        log_and_print "\n%s\n" str ;
        buff

  | TP_PatchSprint(name,msg) ->
        let name = eval_pe_str name in
        let (str : string) = eval_pe_tlk_str game msg in
        let value = Var.get_string str in
        Var.set_string name value  ;
        buff

  | TP_PatchTextSprint (var,str) ->
      let var = eval_pe_str var in
      let str = Var.get_string (eval_pe_str str) in
      Var.add_local_string var str ;
      buff

  | TP_PatchToLower(name) ->
        let name = eval_pe_str name in
        let value = Var.get_string_exact ("%" ^ name ^ "%") in
        Var.set_string name (String.lowercase value) ;
        buff

  | TP_PatchToUpper(name) ->
        let name = eval_pe_str name in
        let value = Var.get_string_exact ("%" ^ name ^ "%") in
        Var.set_string name (String.uppercase value) ;
        buff

  | TP_PatchSnprint(size,name,msg) ->
        let name = eval_pe_str name in
        let str = eval_pe_tlk_str game msg in 
        let value = Var.get_string str in
        let wanted_size = Int32.to_int (eval_pe game size) in 
        let actual_size = String.length value in 
        let substr =
          if actual_size > wanted_size then
            Str.first_chars value wanted_size
          else 
            value
        in 
        Var.set_string name substr ; 
        buff

  | TP_Add_Cre_Item(i) ->

      (* Create the new item *)
      let i = { i with item_name = 
        Var.get_string i.item_name ;
      } in

      let item_buff = String.make 20 '\000' in
      String.blit i.item_name 0 item_buff 0 (String.length i.item_name);
      write_short item_buff 0xa i.i_charge1;
      write_short item_buff 0xc i.i_charge2;
      write_short item_buff 0xe i.i_charge3;

      (* Make the item type into something usable *)
      let i = { i with item_slot = List.fold_left (
          fun acc (from, into) ->
            Str.global_replace (Str.regexp_case_fold from) into acc)
            (" " ^ i.item_slot ^ " ")
              [("[ \t]QITEM[ \t]"," QITEM1 QITEM2 QITEM3 ");
               ("[ \t]QUIVER[ \t]"," QUIVER1 QUIVER2 QUIVER3 ");
               ("[ \t]RING[ \t]"," LRING RRING ");
               ("[ \t]INV[ \t]",
                  " INV1 INV2 INV3 INV4 INV5 INV6 INV7 INV8 INV9 INV10 INV11 INV12 INV13 INV14 INV15 INV16 ")]
               ;
        }
      in
      let type_word_list = Str.split many_whitespace_regexp i.item_slot in
      let possible_slots = List.map (fun str -> match String.uppercase str with
      | "HELMET" -> 0
      | "ARMOR" -> 2
      | "SHIELD" -> 4
      | "GLOVES" -> 6
      | "LRING" -> 8
      | "RRING" -> 10 
      | "AMULET" -> 12
      | "BELT" -> 14
      | "BOOTS" -> 16
      | "WEAPON1" -> 18
      | "WEAPON2" -> 20
      | "WEAPON3" -> 22
      | "WEAPON4" -> 24
      | "QUIVER1" -> 26
      | "QUIVER2" -> 28
      | "QUIVER3" -> 30
      | "QUIVER4" -> 32
      | "CLOAK" -> 34
      | "QITEM1" -> 36
      | "QITEM2" -> 38
      | "QITEM3" -> 40
      | "INV1" -> 42
      | "INV2" -> 44
      | "INV3" -> 46
      | "INV4" -> 48
      | "INV5" -> 50
      | "INV6" -> 52
      | "INV7" -> 54
      | "INV8" -> 56
      | "INV9" -> 58
      | "INV10" -> 60
      | "INV11" -> 62
      | "INV12" -> 64
      | "INV13" -> 66
      | "INV14" -> 68
      | "INV15" -> 70
      | "INV16" -> 72
      | _ -> log_and_print "WARNING: ADD_CRE_ITEM: Unknown slot %s.  Default to INV15 for placement.\n" (String.uppercase str) ; 70) type_word_list 
      in

        let rec find_empty_slot islot_off lst = match lst with
        | [] -> List.hd possible_slots
        | hd :: tl ->
          let in_slot = short_of_str_off buff (islot_off + hd) in
          if in_slot = 65535 then hd
          else find_empty_slot islot_off tl 
        in

      (* Make the flags into something useable *)

      let new_flags =
      (match (String.uppercase i.i_flags) with
      | "NONE" -> 0
      | "IDENTIFIED" -> 1
      | "UNSTEALABLE" -> 2
      | "STOLEN" -> 4
      | "UNDROPPABLE" -> 8
      | "IDENTIFIED&STOLEN" -> 5
      | "IDENTIFIED&UNSTEALABLE" -> 3
      | "IDENTIFIED&UNDROPPABLE" -> 9
      | "UNSTEALABLE&UNDROPPABLE" -> 10
      | "STOLEN&UNDROPPABLE" -> 12
      | "IDENTIFIED&STOLEN&UNDROPPABLE" -> 13
      | "IDENTIFIED&UNSTEALABLE&UNDROPPABLE" -> 11
      | _ -> log_and_print "WARNING: ADD_CRE_ITEM: Unknown flag %s.  Defaulting to NONE for flags.\n" (String.uppercase i.i_flags);
             0)
      in

      write_int item_buff 0x10 new_flags;

      let write_empty_inv_slot buff off itm =
        let i = ref 0 in
        while !i < 16 do
          let slot = short_of_str_off buff (off + (!i * 2)) in
          if slot = 65535 then begin
            write_short buff (off + (!i * 2)) itm;
            i := 16 ;
          end else
            incr i
        done
      in

      (* Check for version *)

      let cre_v = String.sub buff 0 8 in
      (match cre_v with
      | "CRE V1.0" ->

        (* Read in the offsets we need to update and grab the plus state *)

        let islot_off = int_of_str_off buff 0x2b8 in
        let itype = find_empty_slot islot_off possible_slots in
        let items_off = int_of_str_off buff 0x2bc in
        let num_items = int_of_str_off buff 0x2c0 in
        let possible_plus = i.twohanded_weapon in
        write_int buff 0x2b8 (islot_off + 20);
        write_int buff 0x2c0 (num_items + 1);

        (* Check and see if we want to equip the item *)
        if i.equip then begin
          (* If it's a two-handed weapon, empty the shield slot. *)
          if not possible_plus then begin
            let shield_slot = short_of_str_off buff (islot_off + 4) in
            write_empty_inv_slot buff (islot_off + 42) shield_slot ;
            write_short buff (islot_off + 4) 65535 ;
            let slot = short_of_str_off buff (islot_off + itype) in
            if slot <> 65535 then begin
              write_empty_inv_slot buff (islot_off + 42) slot;
              write_short buff (islot_off + itype) num_items;
            end else
              write_short buff (islot_off + itype) num_items;
          end else (* possible_plus *) begin
            (* Not a two-handed weapon, just empty the slot and move any item
               in the slot to the inventory *)
            let slot = short_of_str_off buff (islot_off + itype) in
            if slot = 65535 then
              write_short buff (islot_off + itype) num_items
            else begin
              write_empty_inv_slot buff (islot_off + 42) slot ;
              write_short buff (islot_off + itype) num_items;
            end;
          end;
        end (* end of: if equip *) else begin
          (* We're not equipping anything here, so just move any
             item that may be in the slot to inventory *)
          let slot = short_of_str_off buff (islot_off + itype) in
          if slot <> 65535 then begin
            write_empty_inv_slot buff (islot_off + 42) slot;
            write_short buff (islot_off + itype) num_items
          end else
            write_short buff (islot_off + itype) num_items
        end;

        (* Check if it's a weapon and select the appropriate slot *)
        (* let real_slot = find_empty_slot islot_off possible_slots in *)
        let real_slot = itype in
        if i.equip && real_slot >= 18 && real_slot <= 24 then begin
          let i_num = (real_slot - 16) / 2 in
          write_short buff (islot_off + 76) (i_num - 1)
        end ;
        (* Splice it all together *)
        let splice_off = items_off + (num_items * 20) in
        let before_buff = Str.string_before buff splice_off in
        let after_buff = Str.string_after buff splice_off in
        before_buff ^ item_buff ^ after_buff

      | "CRE V9.0" ->

        (* Read in the offsets we need to update and grab the plus state *)

        let islot_off = int_of_str_off buff 0x320 in
        let items_off = int_of_str_off buff 0x324 in
        let num_items = int_of_str_off buff 0x328 in
        let possible_plus = i.twohanded_weapon in
        write_int buff 0x320 (islot_off + 20);
        write_int buff 0x328 (num_items + 1);

        let itype = find_empty_slot islot_off possible_slots in 

        (* Check and see if we want to equip the item *)

        if i.equip then begin


        (* If it's a two-handed weapon, empty the shield slot. *)

        if not possible_plus then begin
        let shield_slot = short_of_str_off buff (islot_off + 4) in
        write_empty_inv_slot buff (islot_off + 42) shield_slot ;
        write_short buff (islot_off + 4) 65535 ;
        let slot = short_of_str_off buff (islot_off + itype) in
        if slot <> 65535 then begin
        write_empty_inv_slot buff (islot_off + 42) slot;
        write_short buff (islot_off + itype) num_items;
        end else
        write_short buff (islot_off + itype) num_items;
        end else begin

        (* Not a two-handed weapon, just empty the slot and move any item
           in the slot to the inventory *)

        let slot = short_of_str_off buff (islot_off + itype) in
        if slot = 65535 then
        write_short buff (islot_off + itype) num_items
        else begin
        let slot = short_of_str_off buff (islot_off + itype) in
        write_empty_inv_slot buff (islot_off + 42) slot ;
        write_short buff (islot_off + itype) num_items;
        end;
        end;
        end

        (* We're not equipping anything here, so just move any
           item that may be in the slot to inventory *)

        else begin
          let slot = short_of_str_off buff (islot_off + itype) in
          if slot <> 65535 then begin
          write_empty_inv_slot buff (islot_off + 42) slot;
          write_short buff (islot_off + itype) num_items
          end else
          write_short buff (islot_off + itype) num_items
        end;

        (* Check if it's a weapon and select the appropriate slot *)
        (* let real_slot = find_empty_slot islot_off possible_slots in *)
        let real_slot = itype in
        if i.equip && real_slot >= 18 && real_slot <= 24 then begin
          let i_num = (real_slot - 16) / 2 in
          write_short buff (islot_off + 76) (i_num - 1)
        end ;

        (* Splice it all together *)

        let splice_off = items_off + (num_items * 20) in
        let before_buff = Str.string_before buff splice_off in
        let after_buff = Str.string_after buff splice_off in
        before_buff ^ item_buff ^ after_buff

     | _ -> failwith ("ERROR: ADD_CRE_ITEM: Unknown cre version: " ^ cre_v)) ;



  | TP_Replace_Cre_Item(i) ->

      (* Create the new item *)
      let i = { i with item_name =
        Var.get_string i.item_name ;
      } in

      let item_buff = String.make 20 '\000' in
      String.blit i.item_name 0 item_buff 0 (String.length i.item_name);
      write_short item_buff 0xa i.i_charge1;
      write_short item_buff 0xc i.i_charge2;
      write_short item_buff 0xe i.i_charge3;

      (* Make the item type into something usable *)
      let i = { i with item_slot = List.fold_left (
          fun acc (from, into) ->
            Str.global_replace (Str.regexp_case_fold from) into acc)
            (" " ^ i.item_slot ^ " ")
              [("[ \t]QITEM[ \t]"," QITEM1 QITEM2 QITEM3 ");
               ("[ \t]QUIVER[ \t]"," QUIVER1 QUIVER2 QUIVER3 ");
               ("[ \t]RING[ \t]"," LRING RRING ");
               ("[ \t]INV[ \t]",
                  " INV1 INV2 INV3 INV4 INV5 INV6 INV7 INV8 INV9 INV10 INV11 INV12 INV13 INV14 INV15 INV16 ")]
               ;
        }
      in
      let type_word_list = Str.split many_whitespace_regexp i.item_slot in
      let possible_slots = List.map (fun str -> match String.uppercase str with
      | "HELMET" -> 0
      | "ARMOR" -> 2
      | "SHIELD" -> 4
      | "GLOVES" -> 6
      | "LRING" -> 8
      | "RRING" -> 10
      | "AMULET" -> 12
      | "BELT" -> 14
      | "BOOTS" -> 16
      | "WEAPON1" -> 18
      | "WEAPON2" -> 20
      | "WEAPON3" -> 22
      | "WEAPON4" -> 24
      | "QUIVER1" -> 26
      | "QUIVER2" -> 28
      | "QUIVER3" -> 30
      | "QUIVER4" -> 32
      | "CLOAK" -> 34
      | "QITEM1" -> 36
      | "QITEM2" -> 38
      | "QITEM3" -> 40
      | "INV1" -> 42
      | "INV2" -> 44
      | "INV3" -> 46
      | "INV4" -> 48
      | "INV5" -> 50
      | "INV6" -> 52
      | "INV7" -> 54
      | "INV8" -> 56
      | "INV9" -> 58
      | "INV10" -> 60
      | "INV11" -> 62
      | "INV12" -> 64
      | "INV13" -> 66
      | "INV14" -> 68
      | "INV15" -> 70
      | "INV16" -> 72
      | _ -> log_and_print "WARNING: ADD_CRE_ITEM: Unknown slot %s.  Default to INV15 for placement.\n" (String.uppercase str) ; 70) type_word_list
      in

        let rec find_empty_slot islot_off lst = match lst with
        | [] -> List.hd possible_slots
        | hd :: tl ->
          let in_slot = short_of_str_off buff (islot_off + hd) in
          if in_slot = 65535 then hd
          else find_empty_slot islot_off tl
        in

      (* Make the flags into something useable *)

      let new_flags =
      (match (String.uppercase i.i_flags) with
      | "NONE" -> 0
      | "IDENTIFIED" -> 1
      | "UNSTEALABLE" -> 2
      | "STOLEN" -> 4
      | "UNDROPPABLE" -> 8
      | "IDENTIFIED&STOLEN" -> 5
      | "IDENTIFIED&UNSTEALABLE" -> 3
      | "IDENTIFIED&UNDROPPABLE" -> 9
      | "UNSTEALABLE&UNDROPPABLE" -> 10
      | "STOLEN&UNDROPPABLE" -> 12
      | "IDENTIFIED&STOLEN&UNDROPPABLE" -> 13
      | "IDENTIFIED&UNSTEALABLE&UNDROPPABLE" -> 11
      | _ -> log_and_print "WARNING: ADD_CRE_ITEM: Unknown flag %s.  Defaulting to NONE for flags.\n" (String.uppercase i.i_flags);
             0)
      in

      write_int item_buff 0x10 new_flags;

      let write_empty_inv_slot buff off itm =
        let i = ref 0 in
        while !i < 16 do
          let slot = short_of_str_off buff (off + (!i * 2)) in
          if slot = 65535 then begin
            write_short buff (off + (!i * 2)) itm;
            i := 16 ;
          end else
            incr i
        done
      in

      (* Check for version *)

      let cre_v = String.sub buff 0 8 in
      (match cre_v with
      | "CRE V1.0" ->
        let islot_off = int_of_str_off buff 0x2b8 in
        let items_off = int_of_str_off buff 0x2bc in
        let num_items = int_of_str_off buff 0x2c0 in
        let onefilled = ref false in
        (* straight-replace the current item *)
        List.iter (fun this ->
          if not !onefilled then begin
            let this_slot = short_of_str_off buff (islot_off + this) in
            if this_slot <> 65535 then begin
              onefilled := true;
              (* put item_buff in the cre *)
              for i = 0 to 4 do
                let part = int_of_str_off item_buff (i * 4) in
                write_int buff (items_off + 20 * this_slot + i*4) part;
              done ;
              if i.equip then begin
                (* If it's a two-handed weapon, empty the shield slot. *)
                if not i.twohanded_weapon then begin
                  write_short buff (islot_off + 4) 65535 ;
                  write_short buff (islot_off + 76) ((this/2 - 9));
                end else (* possible_plus *) begin
                  (* Not a two-handed weapon, just empty the slot and move any item
                     in the slot to the inventory *)
                  write_short buff (islot_off + 76) ((this/2 - 9));
                end;
              end
            end
          end ;
        ) possible_slots ;
        if !onefilled then buff else process_patch2 patch_filename game buff (TP_Add_Cre_Item(i));
      | "CRE V9.0" ->
        let islot_off = int_of_str_off buff 0x320 in
        let items_off = int_of_str_off buff 0x324 in
        let num_items = int_of_str_off buff 0x328 in
        let onefilled = ref false in
        List.iter (fun this ->
          if not !onefilled then begin
            let this_slot = short_of_str_off buff (islot_off + this) in
            if this_slot <> 65535 then begin
              onefilled := true;
              for i = 0 to 4 do
                let part = int_of_str_off item_buff (i * 4) in
                write_int buff (items_off + 20 * this_slot + i*4) part;
              done ;
              if i.equip then begin
                (* If it's a two-handed weapon, empty the shield slot. *)
                if not i.twohanded_weapon then begin
                  write_short buff (islot_off + 4) 65535 ;
                  write_short buff (islot_off + 76) ((this/2 - 9));
                end else (* possible_plus *) begin
                  (* Not a two-handed weapon, just empty the slot and move any item
                     in the slot to the inventory *)
                  write_short buff (islot_off + 76) ((this/2 - 9));
                end;
              end
            end
          end
        ) possible_slots ;
        if !onefilled then buff else process_patch2 patch_filename game buff (TP_Add_Cre_Item(i));

     | _ -> failwith ("ERROR: ADD_CRE_ITEM: Unknown cre version: " ^ cre_v)) ;



  | TP_CompileBAFtoBCS ->
        (try
        let bcs = handle_script_buffer (patch_filename ^ ".BAF") buff in
        let out_buff = Buffer.create 40960 in
        Bcs.save_bcs game (Bcs.Save_BCS_Buffer(out_buff)) bcs ;
        Buffer.contents out_buff
         with _ -> buff)

  | TP_CompileDLGtoD ->
        handle_dlg_buffer game patch_filename buff

  | TP_CompileDtoDLG ->
        handle_d_buffer game patch_filename buff

  | TP_CompileBCStoBAF ->
        (try
        let bcs = handle_script_buffer (patch_filename ^ ".BCS") buff in
        let out_buff = Buffer.create 40960 in
        Bcs.print_script_text game (Bcs.Save_BCS_Buffer(out_buff))
          (Bcs.BCS_Print_Script(bcs)) false None ;
        Buffer.contents out_buff
         with _ -> buff)

  | TP_EvaluateBuffer -> Var.get_string buff
  
  | TP_PatchReadLN(x) ->
      if !interactive then begin
        let y = read_line ()in
        Var.set_string (eval_pe_str x) y;
        readln_strings:= (x,y) :: !readln_strings;
      end else begin
        match !readln_strings with
        | (a,b) :: tl ->
            Var.set_string (eval_pe_str x) b;
            readln_strings := tl
        | [] ->
            log_and_print "Not enough backed up entries in your replies";
            failwith "Missing READLN strings"
      end ;
      buff

  | TP_Add_S_Item(store_args,item,where,charge1,charge2,charge3,flags,stock,supply) ->

      (* Create the new item *)
      let item = Var.get_string item in
      let item = String.uppercase item in
      let before_what = ref "" in
      let before_or_after = ref 0 in
      List.iter ( fun (name, pos) -> before_what := Var.get_string name ; before_or_after := pos ) where ;
      let before_what_list = Str.split (Str.regexp "[\t ]") !before_what in

      let item_buff = String.make 28 '\000' in
      String.blit item 0 item_buff 0 (String.length item);
      write_short item_buff 0xa charge1;
      write_short item_buff 0xc charge2;
      write_short item_buff 0xe charge3;

      let new_flags =
      (match (String.uppercase flags) with
      | "NONE" -> 0
      | "IDENTIFIED" -> 1
      | "UNSTEALABLE" -> 2
      | "STOLEN" -> 4
      | "IDENTIFIED&STOLEN" -> 5
      | "IDENTIFIED&UNSTEALABLE" -> 3
      | _ -> log_and_print "WARNING: ADD_STORE_ITEM: Unknown flag %s.  Defaulting to 0 for flags.\n" (String.uppercase flags);
             0)
      in

      let supply_tog = match supply with
      | None -> 0
      | Some(str) -> match (String.uppercase str) with
        | "LIMITED" -> 0
        | "UNLIMITED" -> 1
        | _ -> log_and_print "WARNING: ADD_STORE_ITEM: Unknown value %s. Defaulting to 0 (limited) for supply.\n"
                 (String.uppercase str) ; 0
      in

      let stock = if supply_tog = 1 then 1 else stock in

      write_int item_buff 0x10 new_flags;
      write_int item_buff 0x14 stock;
      write_int item_buff 0x18 supply_tog;

      (* Grab the state of the + and uppercase the item string *)

      let possible_plus = store_args.overwrite_store_item in

      (* Read in the offsets that we need to update *)

      let drinksoffset = int_of_str_off buff 0x4c in
      let ipurchasedoffset = int_of_str_off buff 0x2c in
      let isaleoffset = int_of_str_off buff 0x34 in
      let cureoffset = int_of_str_off buff 0x70 in

      (* Read in the # of items for sale *)

      let numisale = int_of_str_off buff 0x38 in


      (* Grab all items *)

      let items_list = ref [] in
      if !debug_ocaml then log_and_print "Trying to String.sub buff %d %d\n" isaleoffset (ipurchasedoffset - isaleoffset);
      let items_string = String.sub buff isaleoffset (numisale * 28) in
      if !debug_ocaml then log_and_print "Done String.sub\n" ;
      for i = 0 to numisale - 1 do
      let item_check = String.sub items_string (i * 28) 8 in
      let item_check = get_string_of_size item_check 0 8 in
      items_list := item_check :: !items_list
      done;

      items_list := List.rev !items_list;

      (* If we have a + sign and the item already exists *)

      if not possible_plus && (List.mem item !items_list) then begin
      let i = ref 0 in
      let str_before = ref "" in
      let str_after = ref "" in
      let new_buff = ref "" in
      List.iter (fun patch_it ->

      if patch_it = item then begin

      (* Overwrite the item *)

      str_before := Str.string_before buff (isaleoffset + (!i * 28));
      str_after := Str.string_after buff (isaleoffset + (!i * 28) + 28);
      new_buff := !str_before ^ item_buff ^ !str_after;
      end
      else
      incr i
      ) !items_list;
      !new_buff
      end

      (* No + sign and the item exists *)

      else if possible_plus && (List.mem item !items_list) then begin
      log_and_print "%s.ITM is already in the store.  Skipping...\n" item;
      buff
      end

      (* No + sign and the item doesn't exist *)

      else begin
        log_and_print "Patching %s.ITM into store...\n" (String.uppercase item);
        (* Update the offsets by 28 bytes *)
        if ipurchasedoffset >= isaleoffset then write_int buff 0x2c (ipurchasedoffset + 28);
        if cureoffset >= isaleoffset then write_int buff 0x70 (cureoffset + 28);
        if drinksoffset >= isaleoffset then write_int buff 0x4c (drinksoffset + 28);
        (* Add 1 to the #items for sale *)
        write_int buff 0x38 (numisale + 1);
        let added = ref false in
        let buff_ref = ref (Str.string_before buff (isaleoffset)) in
        if (!before_or_after = 1) or (!before_or_after = 2) then begin
          let item_reg = List.map (fun x ->
            let y = if String.length x > 8 then Str.string_before x 8 else x in
            (y ^ String.make (8 - (String.length y)) '.')
          ) before_what_list in
(*           if !debug_ocaml then begin
            log_and_print "item_reg is -%s-\n" item_reg ;
          end ; *)
          let item_reg = List.map Str.regexp_case_fold item_reg in
          let item_string = Str.string_after (Str.string_before buff (isaleoffset + 28 * numisale)) isaleoffset in
          let length = (String.length item_string) / 28 in
(*           if !debug_ocaml then log_and_print "item_string:\"%s\", with %d items\n" item_string length ; *)
          List.iter (fun this_test ->
            if not !added then
              for i = 0 to length - 1 do
                let this_one_long = Str.string_before (Str.string_after item_string (i * 28)) 28 in
                if not !added then begin
                  let this_one = Str.string_before (this_one_long) 8 in
                  if !debug_ocaml then log_and_print "Item %d is \"%s\"\n" i this_one ;
                  if Str.string_match this_test this_one 0 then begin
                    buff_ref := Printf.sprintf "%s%s%s" !buff_ref
                    (* t_i_l i_b
                       i_b t_i_l *)
                      (if !before_or_after = 1 then this_one_long else item_buff)
                      (if !before_or_after = 1 then item_buff else this_one_long) ;
                    added := true
                  end
                  else buff_ref := !buff_ref ^ this_one_long
                end else buff_ref := !buff_ref ^ this_one_long ;
              done ;
          ) item_reg ;
          if not !added then log_and_print "Not found space for %s %s %s.\n" item (if !before_or_after = 1
              then "after" else "before") !before_what ;
        end ;
        let after = Str.string_after buff (match !added, !before_or_after with
          | (true,_)
          | (false,3) -> isaleoffset + 28 *numisale
          | _     -> isaleoffset
        ) in
        if not !added then begin
          let before = Str.string_before buff (match !before_or_after with
            | 3 -> isaleoffset + 28 *numisale
            | _ -> isaleoffset
          ) in
          buff_ref := before ^ item_buff
        end ;
        buff_ref := !buff_ref ^ after ;
        !buff_ref
      end

  | _ -> buff
  (* end: process_patch2 *)
  ) ()

(*************************************************************************
                           process_action
 *************************************************************************)
  and process_action tp a =
    Var.set_int32 "INTERACTIVE" (Int32.of_int (if !interactive then 1 else 0)) ;
    let str = action_to_str a in
    Stats.time str (fun () ->
    try
    (match a with
    | TP_Require_File(file,error_msg) ->
        log_and_print "Checking for required files ...\n" ;
        let file = Arch.backslash_to_slash file in
        let size = file_size file in
        let test = ref false in
        test := bigg_file_exists file game.Load.key ;
        if !test then begin
          log_or_print "[%s] found: %d bytes\n" file size;
        end
        else begin
          log_and_print "[%s] not found\n" file ;
          log_and_print "\n%s\n" (Dc.single_string_of_tlk_string game error_msg) ;
          failwith file ;
        end

    | TP_Inlined_File(name,contents) ->
        let name = Var.get_string (Arch.backslash_to_slash name) in
        log_only "Defined Inlined File [%s] (length %d)\n"
          name (String.length contents) ;
        Hashtbl.add inlined_files (Arch.backslash_to_slash name) contents

    | TP_Biff(dest,what) ->
        let find_list = ref [] in
        game.Load.key_mod <- true ;
        List.iter ( fun (directory,exact_match,regexp_string) ->
          try
            let dh = Case_ins.unix_opendir directory in
            let reg = begin match exact_match with
                | Some(true) ->
                  Str.regexp_string_case_fold
                | _ ->
                  Str.regexp_case_fold
              end regexp_string
            in
            (try
              while true do
                let next = Unix.readdir dh in
                if ((Case_ins.unix_stat (directory ^ "/" ^ next)).Unix.st_kind =
                   Unix.S_REG) && (Str.string_match reg next 0) then
                  find_list := (String.uppercase (directory ^ "/" ^ next)) :: !find_list
              done
            with End_of_file -> () );
            Unix.closedir dh ;
          with _ -> ()
        ) what ;
        if !find_list <> [] then begin
          let filename = "data/" ^ dest ^ ".bif" in
          let new_key = Biff.save_biff game.Load.key filename !find_list in
          let oc = open_for_writing "CHITIN.KEY" true in
          Key.save_key new_key oc ;
          close_out oc ;
          (* re-load the chitin *)
          let keyname = Load.find_file_in_path "." "^chitin.key$" in
          let keybuff = load_file keyname in
          game.Load.key <- Key.load_key keyname keybuff ;
          game.Load.loaded_biffs <- Hashtbl.create 5 ;
        end

    | TP_ActionBashFor(where,al) ->
        let find_list = ref [] in
        List.iter ( fun (directory,exact_match,regexp_string) ->
          try
            let dh = Case_ins.unix_opendir directory in
            let reg = begin match exact_match with
                | Some(true) ->
                  Str.regexp_string_case_fold
                | _ ->
                  Str.regexp_case_fold
              end regexp_string
            in
            (try
              while true do
                let next = Unix.readdir dh in
                if ((Case_ins.unix_stat (directory ^ "/" ^ next)).Unix.st_kind =
                   Unix.S_REG) && (Str.string_match reg next 0) then
                  find_list := (String.uppercase (directory ^ "/" ^ next)) :: !find_list
              done
            with End_of_file -> () );
            Unix.closedir dh ;
          with _ -> ()
        ) where ;
        List.iter (fun file ->
          let directory = Filename.dirname file in
          let filespec = Filename.basename file in
          Var.set_string "BASH_FOR_DIRECTORY" directory ;
          Var.set_string "BASH_FOR_FILESPEC" file ;
          Var.set_string "BASH_FOR_FILE" filespec ;
          Var.set_string "BASH_FOR_RES"
            (let a,b = split filespec in a) ;
          List.iter (process_action tp) al ;
        ) !find_list ;

    | TP_Action_For_Each(var,sl,al) ->
        let var = eval_pe_str var in
        let sl = List.map Var.get_string sl in
        List.iter (fun x ->
          Var.set_string var x ;
          List.iter (process_action tp) al ;
        ) sl ;

    | TP_Outer_For(init,guard,inc,body) ->
        let cmd_list = init @ [ TP_PatchWhile(guard,[(TP_PatchInnerAction body)] @ inc) ] in
        let b = ref "" in
        let patch_filename = "" in
        b := List.fold_left (fun acc elt ->
            process_patch2 patch_filename game acc elt) !b cmd_list  ;
       ()

    | TP_Outer_While(guard,body) ->
        let cmd_list = [ TP_PatchWhile(guard,[TP_PatchInnerAction body])] in
        let b = ref "" in
        let patch_filename = "" in
        b := List.fold_left (fun acc elt ->
            process_patch2 patch_filename game acc elt) !b cmd_list  ;
       ()

    | TP_Outer_Inner_Buff(buff_var,pl) ->
        let new_buff = Var.get_string buff_var in
        let filename = Printf.sprintf "INNER_PATCH %S" buff_var in
        let dummy = List.fold_left (fun acc elt ->
            process_patch2 filename game acc elt) new_buff pl in
        ()

      | TP_Outer_Inner_Buff_Save(store_var,buff_var,pl) ->
        let new_buff = Var.get_string buff_var in
        let filename = Printf.sprintf "INNER_PATCH_SAVE %S" buff_var in
        let result = List.fold_left (fun acc elt ->
            process_patch2 filename game acc elt) new_buff pl
        in
        Var.set_string store_var (Var.get_string result)

      | TP_Outer_Set(name,value) ->
        let value = (eval_pe game value) in
        let name = eval_pe_str name in
        Var.set_int32 name value ;

    | TP_Outer_Sprint(name,msg) ->
        let name = eval_pe_str name in
        let (str : string) = eval_pe_tlk_str game msg in
        let value = Var.get_string str in
        Var.set_string name value  ;

    | TP_Outer_Text_Sprint (var,str) ->
        let var = eval_pe_str var in
        let str = Var.get_string (eval_pe_str str) in
        Var.add_local_string var str
  
    | TP_Forbid_File(file,error_msg) ->
        log_and_print "Checking for forbidden files ...\n" ;
        let file = Arch.backslash_to_slash file in
        let size = file_size file in
        let test = ref false in
        test := bigg_file_exists file game.Load.key ;
       if !test then begin
          log_and_print "[%s] found: %d bytes\n" file size ;
          log_and_print "\n%s\n" (Dc.single_string_of_tlk_string game
          error_msg) ;
          failwith file
        end else begin
          log_or_print "[%s] not found (as desired)\n" file ;
        end

    | TP_Print(msg) ->
        let str = Dc.single_string_of_tlk_string game msg in
        be_silent := false ;
        let str = Var.get_string str in
        log_and_print "\n%s\n" str

    | TP_Fail(msg) ->
        let str = Dc.single_string_of_tlk_string game msg in
        log_and_print "FAILURE:\n%s\n" str ;
        failwith str

    | TP_If(p,al1,al2) ->
        let res = is_true (eval_pe game p) in
        (* log_or_print "IF evaluates to %b\n" res ; *)
        if res then begin
          List.iter (process_action tp) al1
        end else begin
          List.iter (process_action tp) al2
        end

    | TP_Define_Action_Macro(str,decl,al) ->
        Hashtbl.replace action_macros str (decl, al)
    
    | TP_Define_Patch_Macro(str,decl,al) ->
        Hashtbl.replace patch_macros str (decl, al)

    | TP_Launch_Action_Macro(str) ->
      let (decl, actions) =
        try
          Hashtbl.find action_macros str
        with _ ->
          failwith (Printf.sprintf "Unknown Macro: -%s-" str)
(*           ( [] , [] ) *)
      in
      List.iter (fun x -> match x with
        TP_LocalSet(var,pe) ->
          let pe = (eval_pe game pe) in
          let var = eval_pe_str var in
          Var.add_local_int32 var pe
        | TP_LocalSprint (var,str) ->
          let var = eval_pe_str var in
          let (str : string) = eval_pe_tlk_str game str in
          let str = Var.get_string str in
          Var.add_local_string var str
        | TP_LocalTextSprint (var,str) ->
            let var = eval_pe_str var in
            let str = Var.get_string (eval_pe_str str) in
            Var.add_local_string var str
      ) decl ;
      List.iter (process_action tp) actions ;
      List.iter (fun x -> match x with
        TP_LocalSet(var,_)
      | TP_LocalSprint (var,_)
      | TP_LocalTextSprint (var,_) ->
        let var = eval_pe_str var in
        Var.remove_local var
      ) decl ;

    | TP_Reinclude(string_list) ->
      let string_list = List.map Var.get_string string_list in
      List.iter (fun file ->
        let tph_parsed =
          if !debug_ocaml then log_and_print "Loading %s...\n" file ;
          let x = handle_tph_filename file in
          Hashtbl.replace loaded_tph file x ;
          x
        in
        List.iter (process_action tp) tph_parsed ;
      ) string_list ;

    | TP_Include(string_list) ->
      let string_list = List.map Var.get_string string_list in
      List.iter (fun file ->
        let tph_parsed = if Hashtbl.mem loaded_tph file
          then begin
            if !debug_ocaml then log_and_print "%s is already loaded, restoring the image...\n" file ;
            Hashtbl.find loaded_tph file
          end
          else begin
            if !debug_ocaml then log_and_print "Loading %s...\n" file ;
            let x = handle_tph_filename file in
            Hashtbl.replace loaded_tph file x ;
            x
          end
        in
        List.iter (process_action tp) tph_parsed ;
      ) string_list ;

    | TP_Action_ReadLN(x) ->
        if !interactive then begin
          let y = read_line ()in
          Var.set_string (eval_pe_str x) y;
          readln_strings:= (x,y) :: !readln_strings;
        end else begin
          match !readln_strings with
          | (a,b) :: tl ->
              Var.set_string (eval_pe_str x) b;
              readln_strings := tl
          | [] ->
          log_and_print "Not enough backed up entries in your replies"; failwith "Missing READLN strings"
        end;

    | TP_Uninstall_Now(name,comp) ->
        let comp = Int32.to_int (eval_pe game comp) in
        let name = Var.get_string name in
        if already_installed name comp then begin
          if uninstall game handle_tp2_filename name comp false then
            ()
          else
            failwith
              (Printf.sprintf "unable to uninstall %s component #%d"
                name comp)
        end else
          log_or_print "%s component #%d not present, good.\n"
            name comp

    | TP_RandomSeed(i) ->
      eval_pe_warn := false ;
      let _ = try
          let x = Int32.to_int (eval_pe game i) in
          Random.init x ;
        with _ ->
          begin
            Random.self_init () ;
          end
      in
      eval_pe_warn := true ;

    | TP_ClearMemory ->
      log_and_print "Clearing the memory.\n" ;
      let temp_author = Var.get_string "%TP2_AUTHOR%" in
      let temp_lang = Var.get_string "%LANGUAGE%" in
      let temp_name = Var.get_string "%TP2_FILE_NAME%" in
      Var.clear_var () ;
      Var.all_the_assoc () ;
      Var.set_string "TP2_AUTHOR" temp_author ;
      Var.set_string "LANGUAGE" temp_lang ;
      Var.set_string "TP2_FILE_NAME" temp_name ;
      Arch2.associate_these () ;

    | TP_Clear_Ids_Map ->
      log_and_print "Clearing the IDS map.\n" ;
      Bcs.clear_ids_map game ;

    | TP_Silent -> be_silent := true ;
    | TP_Verbose -> be_silent := false ;

    | TP_CopyRandom(slistlist,plist,wlist) ->
        List.iter (fun slist ->
        log_and_print "Randomizing %d file(s) ...\n" (List.length slist) ;
        let final_dlist = ref [] in
        let finished = ref false in
        while not !finished do
          let dlist = List.map (fun s -> (Random.bits (), s) ) slist in
          let dlist = List.sort (fun (a,b) (a',b') -> compare a a') dlist in
          final_dlist := dlist ;
          finished := true ;
          if List.length dlist > 1 then
            List.iter2 (fun (a,b) c -> if b = c then finished := false)
              dlist slist
        done ;
        let dlist = List.map (fun (a,b) -> "override/" ^ b) !final_dlist in
        let sdlist = List.combine slist dlist in
        let copy_args = {
          copy_get_existing = true;
          copy_use_regexp = false;
          copy_use_glob = false;
          copy_file_list = sdlist ;
          copy_patch_list = plist ;
          copy_constraint_list = wlist ;
          copy_backup = true;
          copy_at_end = true;
          copy_save_inlined = false;
        } in
        process_action tp (TP_Copy(copy_args))
        ) slistlist

    | TP_CopyAllGamFiles(pl,wl) ->
        (* Patch the baldur.gam in the biff *)
        let my_copy_args = {
          copy_get_existing = true;
          copy_use_regexp = false;
          copy_use_glob = false ;
          copy_file_list = [ ("baldur.gam", "override") ] ;
          copy_patch_list = pl ;
          copy_constraint_list = wl ;
          copy_backup = true ;
          copy_at_end = false ;
          copy_save_inlined = false ;
        } in
        process_action tp (TP_Copy(my_copy_args)) ;
        let file_list = ref [] in
        let get_gam_list base_dir =
          let dlist = list_of_files_in_directory base_dir in
          List.iter (fun filename ->
            let filename = base_dir ^ "/" ^ filename in
            if is_directory filename          &&
               filename <> base_dir ^ "/."    &&
               filename <> base_dir ^ "/.."
            then
              file_list := !file_list @ [(filename ^ "/baldur.gam", filename ^ "/baldur.gam")] ;
          ) dlist ;
          ()
        in
        get_gam_list "save" ;
        get_gam_list "mpsave" ;
        let my_copy_args = {
          copy_get_existing = false;
          copy_use_regexp = false;
          copy_use_glob = false ;
          copy_file_list = !file_list ;
          copy_patch_list = pl ;
          copy_constraint_list = wl ;
          copy_backup = false ;
          copy_at_end = false ;
          copy_save_inlined = false ;
        } in
        process_action tp (TP_Copy(my_copy_args)) ;

    | TP_Copy(copy_args) ->
        if !debug_ocaml then log_and_print "Welcome to TP_Copy()\n" ;
        let get_existing = copy_args.copy_get_existing in
        let use_reg = copy_args.copy_use_regexp in
        let use_glob = copy_args.copy_use_glob in
        let slist = copy_args.copy_file_list in
        let plist = copy_args.copy_patch_list in
        let clist = copy_args.copy_constraint_list in
        let make_a_backup = copy_args.copy_backup in
        let copy_at_end = copy_args.copy_at_end in
        let save_inlined = copy_args.copy_save_inlined in
        if true then begin
          let worklist = ref [] in

          let bts1, bts2 = match get_existing, use_reg with
            | false, false -> true,  true
            | true,  false -> false, true
            | true,  true  -> false, false
            | _ -> failwith "COPY type internal"
          in
          let slist = List.map (fun (x,y) ->
            (if bts1 then Arch.backslash_to_slash x else x),
            (if bts2 then Arch.backslash_to_slash y else y)
          ) slist
          in

          let slist = List.map (fun (x,y) -> (Var.get_string x, Var.get_string y)) slist in

          let slist =
            if get_existing = true && use_reg = true then begin
              let files_in_chitin = Key.list_of_key_resources game.Load.key in
              let files_in_override =
                if use_glob then begin
                  try
                    let dh = Case_ins.unix_opendir "override" in
                    let lst = ref [] in
                    (try
                      while true do
                        let next = Unix.readdir dh in
                        if ((Case_ins.unix_stat ("override/" ^ next)).Unix.st_kind =
                           Unix.S_REG) then
                          lst := (String.uppercase next) :: !lst
                      done
                    with End_of_file -> () );
                    Unix.closedir dh ;
                    !lst
                  with _ -> []
                end else []
              in
              let new_list = List.map (fun (s,p) ->
                let regexp = Str.regexp_case_fold s in
                let matches = ref [] in
                List.iter (fun possible ->
                  if Str.string_match regexp possible 0 then begin
                    matches := (possible, p ^ "/" ^ possible) :: !matches
                  end
                ) (files_in_chitin @ files_in_override);
                let matches = List.sort compare !matches in
                let matches =
                  let rec nodup lst = match lst with
                  | [] -> []
                  | [hd] -> lst
                  | a :: b :: tl -> if (a = b) then nodup (b :: tl)
                                               else a :: (nodup (b :: tl))
                  in nodup matches
                in
                if (matches = []) then
                  [(s,p)]
                else
                  matches
              ) slist in
              List.flatten new_list
            end else if use_glob then begin
              let res = ref [] in
              List.iter (fun (s,p) ->
                let myfun (glob_s : string) = begin
                  log_only "Callback from Arch.glob: %s\n" glob_s;
                  res := (glob_s, p) :: !res
                end in
                log_only "Calling Arch.glob: %s\n" s;
                let res = (Arch.glob s myfun) in
                log_only "Arch.glob returned.\n";
                res
              ) slist ;
              !res
            end else
              slist
          in

          let len = List.length slist in
          log_and_print "Copying%s %d file%s ...\n"
            (if List.length plist > 0 then " and patching" else "")
            len
            (if len = 1 then "" else "s") ;

          let copy_one_file src dest = begin
            let dest =
              if use_reg || use_glob then
                dest
              else
                Var.get_string dest
            in
            let src =
              if use_reg || use_glob then
                src
              else
                Var.get_string src
            in
            if String.uppercase src = "DIALOG.TLK" then
              log_and_print "\n\nUse COPY_LARGE rather than COPY on dialog.tlk!\n\n\n" ;

            let src_dir = Filename.dirname src in
            Var.set_string "SOURCE_DIRECTORY" src_dir ;
            Var.set_string "SOURCE_FILESPEC" src ;
            Var.set_string "SOURCE_FILE" (Filename.basename src) ;
            Var.set_string "SOURCE_RES"
              (let a,b = split (Filename.basename src) in a) ;
            let dest_dir = Filename.dirname dest in
            Var.set_string "DEST_DIRECTORY" dest_dir ;
            Var.set_string "DEST_FILESPEC" dest ;
            Var.set_string "DEST_FILE" (Filename.basename dest) ;
            Var.set_string "DEST_RES"
              (let a,b = split (Filename.basename dest) in a) ;
            let buff =
              if not get_existing then
                load_file src
              else
                let a,b = split src in
                let buff,path = Load.load_resource "COPY" game true a b in
                buff
            in
            Var.set_int32 "SOURCE_SIZE" (Int32.of_int (String.length buff));
            let orig_buff = String.copy buff
            in

            (* if (buff <> "") then *) begin
              if (!has_if_eval_bug) then begin try
                  List.iter (fun p -> process_patch1 src game buff p) plist
                with _ -> ()
              end;
              let ok_to_copy = List.fold_left (fun acc elt -> acc &&
                match elt with
                | TP_Contains(s) -> begin
                    let my_regexp = Str.regexp_case_fold s in
                    try let _ = Str.search_forward my_regexp buff 0 in
                      true
                    with _ ->
                      log_only "Not copying [%s] to [%s] because it does NOT contain [%s]\n" src dest s ;
                      false
                    end
                | TP_NotContains(s) -> begin
                    let my_regexp = Str.regexp_case_fold s in
                    try let _ = Str.search_forward my_regexp buff 0 in
                      log_only "Not copying [%s] to [%s] because it DOES contain [%s]\n" src dest s ;
                      false
                    with _ ->
                      true
                    end
                | TP_IfSizeIs(size) ->
                    if String.length buff = size then
                      true
                    else begin
                      log_only "Not copying [%s] to [%s] because size is %d, NOT %d\n" src dest (String.length buff) size ;
                      false
                    end
                | TP_Eval(pe) ->
                    let v = eval_pe game pe in
                    if v = Int32.zero then begin
                      log_only "Not copying [%s] to [%s] because condition evaluates to %ld\n" src dest v ; false
                    end else true
                | TP_ButOnlyIfItChanges -> true
              ) true clist in
              if ok_to_copy then begin
                let result_buff =
                  List.fold_left (fun acc elt ->
                    try process_patch2 src game acc elt
                    with e -> log_and_print
                      "ERROR: [%s] -> [%s] Patching Failed (COPY) (%s)\n"
                      src dest (Printexc.to_string e); raise e)
                  buff plist
                in
                let dest =
                  if is_directory dest then
                    dest ^ "/" ^ (Filename.basename src)
                  else
                    dest
                in
                let it_changed =
                  Stats.time "BUT_ONLY" (fun () ->
                  if List.mem TP_ButOnlyIfItChanges clist then
                    begin
                      let changed =
                        ((String.length result_buff) <>
                         (String.length orig_buff)) ||
                        ((String.compare orig_buff result_buff) <> 0)
                      in
                      let newsrc, newdes =
                        ( Filename.basename src, Filename.basename dest )
                      in
                      if !debug_boiic then log_and_print "[%s]%s->[%s]%s did%s change\n"
                          newsrc (String.make (12 - (String.length newsrc)) ' ')
                          newdes (String.make (12 - (String.length newdes)) ' ')
                          (if changed then "" else "n't") ;
                      changed
                    end else begin
                      if (!debug_change) && (plist <> []) then begin
                        let changed =
                          ((String.length result_buff) <>
                           (String.length orig_buff)) ||
                          ((String.compare orig_buff result_buff) <> 0)
                        in
                        let newsrc, newdes =
                          ( Filename.basename src, Filename.basename dest )
                        in
                        if !debug_change && not changed then log_and_print "Patches on [%s]%s->[%s]%s don't alter the file\n"
                            newsrc (String.make (12 - (String.length newsrc)) ' ')
                            newdes (String.make (12 - (String.length newdes)) ' ') ;
                        true
                      end else true ;
                      (* always copy *)
                    end ) ()
                in
                if (it_changed) then begin
                  let doit () = Stats.time "saving files" (fun () ->
                    if not (save_inlined) then begin
                      let out =
                        try open_for_writing_internal make_a_backup dest true
                        with e -> log_and_print
                          "ERROR: COPY ~%s~ ~%s~ FAILED: cannot open target\n"
                            src dest ;
                          raise e
                      in
                      output_string out result_buff ;
                      close_out out ;
                      begin (* handle read-only files! *)
                        try
                          Case_ins.unix_chmod dest 511 ; (* 511 = octal 0777 = a+rwx *)
                        with e -> ()
                          (* log_or_print "WARNING: chmod %s : %s\n" filename
                            (Printexc.to_string e) *)
                      end ;
                      if make_a_backup then
                        log_only "Copied [%s] to [%s]\n" src dest
                      else
                        log_only "Copied [%s] to [%s] (NO BACKUP MADE!)\n" src dest
                    end else begin
                      log_only "Defined Inlined File [%s] (length %d)\n"
                      dest (String.length result_buff) ;
                      Hashtbl.add inlined_files (Arch.backslash_to_slash dest) result_buff
                    end
                  ) ()
                  in
                  if copy_at_end then worklist := doit :: !worklist
                  else doit ()
                end else begin
                  log_only
                    "Not copying [%s] to [%s] because it did not change\n"
                    src dest
                end
              end (* end: if ok_to_copy *)
            end (* (* end: if buff <> "" *)
            else log_or_print "Not copying [%s] to [%s] because it is empty\n"
              src dest
              *)
          end (* end: let copy_one_file = begin *)
          in
          List.iter (fun (src,dest) ->
            let src = Var.get_string src in
            if is_directory src then begin
              let dh = Case_ins.unix_opendir src in
              (try
                while true do
                  let base = Unix.readdir dh in
                  let source = src ^ "/" ^ base in
                  if not (is_directory source) then
                    copy_one_file source (dest ^ "/" ^ base)
                done
              with _ -> () ) ;
              Unix.closedir dh
            end else copy_one_file src dest
          ) slist ;
          List.iter (fun f -> f ()) !worklist
        end

    | TP_CopyLarge(copy_large_args) ->
        if !debug_ocaml then log_and_print "Welcome to TP_CopyLarge()\n" ;
        let use_glob = copy_large_args.copy_large_use_glob in
        let slist = copy_large_args.copy_large_file_list in
        let make_a_backup = copy_large_args.copy_large_backup in
        let worklist = ref [] in

        let slist =
          List.map (fun (x,y) -> (Arch.backslash_to_slash x, Arch.backslash_to_slash y)) slist
        in

        let slist =
          if use_glob then begin
            let res = ref [] in
            List.iter (fun (s,p) ->
              let myfun (glob_s : string) = begin
                log_only "Callback from Arch.glob: %s\n" glob_s;
                res := (glob_s, p) :: !res
              end in
              log_only "Calling Arch.glob: %s\n" s;
              let res = (Arch.glob s myfun) in
              log_only "Arch.glob returned.\n";
              res
            ) slist ;
            !res
          end else
            slist
        in

        let len = List.length slist in
        log_and_print "Copying%s %d file%s ...\n" "" len
          (if len = 1 then "" else "s") ;

        let copy_one_file src dest = begin
          let dest =
            if use_glob then
              dest
            else
              Var.get_string dest
          in
          let src =
            if use_glob then
              src
            else
              Var.get_string src
          in
          let dest =
            if is_directory dest then
              dest ^ "/" ^ (Filename.basename src)
            else
              dest
          in
          match String.uppercase (snd (split dest)) with
          | ".IDS" -> Bcs.clear_ids_map game
          | _ -> ()
          ;
          Stats.time "saving files" (fun () ->
            if Hashtbl.mem inlined_files (Arch.backslash_to_slash src) then begin
              let copy_args = {
                copy_get_existing = false;
                copy_use_regexp = false;
                copy_use_glob = use_glob;
                copy_file_list = [(src,dest)] ;
                copy_patch_list = [] ;
                (* C_L is <> [] to avoid looping *)
                copy_constraint_list = [TP_Eval(get_pe_int "1")] ;
                copy_backup = make_a_backup ;
                copy_at_end = false ;
                copy_save_inlined = false ;
              } in
              (* we came here from tp_copy, now we go back  :) *)
              process_action tp (TP_Copy(copy_args))
            end else begin
              if make_a_backup then
                backup_if_extant dest ;
              if !debug_ocaml then log_and_print "Calling Util.copy_large_file() due to COPY_LARGE in source\n" ;
              copy_large_file src dest "doing a COPY_LARGE" ;
              if make_a_backup then
                log_only "Copied [%s] to [%s]\n" src dest
              else
                log_only "Copied [%s] to [%s] (NO BACKUP MADE!)\n" src dest
            end
            ) ()
        end (* end: let copy_one_file = begin *)
        in
        List.iter (fun (src,dest) ->
          let src = Var.get_string src in
          if is_directory src then begin
            let dh = Case_ins.unix_opendir src in
            (try
              while true do
                let base = Unix.readdir dh in
                let source = src ^ "/" ^ base in
                if not (is_directory source) then
                  copy_one_file source (dest ^ "/" ^ base)
              done
            with _ -> () ) ;
            Unix.closedir dh
          end else copy_one_file src dest
        ) slist ;
        List.iter (fun f -> f ()) !worklist

    | TP_Add_Music(m) -> begin
      if is_true (eval_pe game
        (PE_And(
        (Pred_File_Exists(PE_LiteralString "override/SONGLIST.2DA")),
        (Pred_File_Contains(PE_LiteralString "override/SONGLIST.2DA",
          PE_LiteralString ("^" ^ m.music_name ^ "\\b")))))) then begin
        let music_base_name = Filename.basename m.music_file in
        Var.set_int32 (music_base_name) (Bcs.int_of_sym game "SONGLIST.2DA" music_base_name) ;
        log_and_print "\n\nERROR: MUS [%s] already present! Skipping!\n\n"
        music_base_name
      end else begin
        log_and_print "Adding %s Music ...\n" m.music_name;
        let this_music_number = get_next_line_number "SONGLIST.2DA" in
        let music_base_name = Filename.basename m.music_file in
        let str_to_append = Printf.sprintf "%d %s %s"
          this_music_number m.music_name music_base_name in

        let a1 = TP_Append("SONGLIST.2DA",str_to_append,[],true) in

        let dest_music_file = "music/" ^ music_base_name in
        let a2 = TP_Copy(
          { copy_get_existing = false;
            copy_use_regexp = false;
            copy_use_glob = false;
            copy_file_list = [(m.music_file,dest_music_file)] ;
            copy_patch_list = [] ;
            copy_constraint_list = [] ;
            copy_backup = true;
            copy_at_end = false;
            copy_save_inlined = false;
            } ) in
        let action_list = [ a1 ; a2 ] in
        List.iter (process_action tp) action_list ;
        Var.set_int32 (m.music_name) (Int32.of_int this_music_number) ;
        log_and_print "Added %s Music\n" m.music_name;
      end
    end

    | TP_Add_Projectile(p) -> begin
      if is_true (eval_pe game
        (PE_And(
        (Pred_File_Exists(PE_LiteralString "override/PROJECTL.IDS")),
        (Pred_File_Contains(PE_LiteralString "override/PROJECTL.IDS",
          PE_LiteralString ("^" ^ p.pro_file ^ "\\b")))))) then begin
        let this_pro_name = Filename.chop_extension (Filename.basename p.pro_file) in
        Var.set_int32 (this_pro_name) (Bcs.int_of_sym game "PROJECTL.IDS" this_pro_name) ;
        log_and_print "\n\nERROR: PRO [%s] already present! Skipping!\n\n"
        this_pro_name
      end else begin
        log_and_print "Adding projectile file %s ...\n" p.pro_file;
        let this_pro_name = Filename.chop_extension (Filename.basename p.pro_file) in
        let this_pro_number = get_next_line_number "PROJECTL.IDS" in
        let a1 = TP_Append("PROJECTL.IDS",
                           (Printf.sprintf "%d %s" this_pro_number this_pro_name),[],true) in
        let dest_pro_file = "override/" ^ (Filename.basename p.pro_file) in
        let a2 = TP_Copy(
          { copy_get_existing = false;
            copy_use_regexp = false;
            copy_use_glob = false;
            copy_file_list = [(p.pro_file, dest_pro_file)] ;
            copy_patch_list = [] ;
            copy_constraint_list = [] ;
            copy_backup = true;
            copy_at_end = false;
            copy_save_inlined = false;
            } ) in
        let action_list = [ a1 ; a2 ] in
        List.iter (process_action tp) action_list ;
        Var.set_int32 (this_pro_name) (Int32.of_int (this_pro_number + 1)) ;
        log_and_print "Added projectile file %s\n" p.pro_file;
      end
    end

    | TP_Add_Spell(file,kind,level,ids_name,pl,cl) ->
      log_and_print "Adding spell %s\n" ids_name;
      let file = Var.get_string file in
      let ids_name = Var.get_string ids_name in
      let kind = eval_pe game kind in
      let level = eval_pe game level in
      let kind = Int32.to_int kind in
      let level = Int32.to_int level in
      let memo = match kind with
        | 1 -> "PR"
        | 2 -> "WI"
        | 3 -> "IN"
        | 4 -> "CL"
        | _ -> failwith "ADD_SPELL with spell type not in 1-4 range."
      in
      if level < 0 || level > 9 then failwith "ADD_SPELL with level outside the 0-9 range.";
      let rec try_it n =
        if n = 100 then failwith "Couldn't find space for the spell."
        else if not (Load.file_exists_in_game game (Printf.sprintf "SP%s%d%s%d.SPL" memo level (if n < 10 then "0" else "") n)) then
          n
        else try_it (n + 1)
      in
      let n = try_it 1 in
      if !debug_ocaml then log_and_print "installed as SP%s%d%s%d.SPL\n" memo level 
                                                (if n < 10 then "0" else "") n;
      let number = kind * 1000 + level * 100 + n in
        let a1 = TP_Append("SPELL.IDS",
                           (Printf.sprintf "%d %s" number ids_name),[],true) in
        let dest_file = "override/" ^ (Printf.sprintf "SP%s%d%s%d.SPL" memo level
                                                (if n < 10 then "0" else "") n) in
        let a2 = TP_Copy(
          { copy_get_existing = false;
            copy_use_regexp = false;
            copy_use_glob = false;
            copy_file_list = [(file, dest_file)] ;
            copy_patch_list = pl ;
            copy_constraint_list = cl ;
            copy_backup = true;
            copy_at_end = false;
            copy_save_inlined = false;
            } ) in
        let action_list = [ a1 ; a2 ] in
        List.iter (process_action tp) action_list ;
        Var.set_int32 (ids_name) (Int32.of_int (number)) ;
        Bcs.clear_ids_map game ;
        log_and_print "Added spell %s\n" ids_name;



    | TP_Add_Kit(k) -> begin
      log_and_print "Adding %s Kit ...\n" k.kit_name;

      if is_true (eval_pe game
        (PE_And(
        (Pred_File_Exists(PE_LiteralString "override/kitlist.2da")),
        (Pred_File_Contains(PE_LiteralString "override/kitlist.2da",
          PE_LiteralString ("^" ^ k.kit_name ^ "\\b")))))) then begin
        Var.set_int32 (k.kit_name) (Bcs.int_of_sym game "KITLIST.2DA" k.kit_name) ;
        log_and_print "\n\nERROR: Kit [%s] already present! Skipping!\n\n"
        k.kit_name
      end else begin
        let a1 = TP_Append("CLASWEAP.2DA",k.clasweap,[],true) in
        let a2 = TP_Append_Col("WEAPPROF.2DA",k.weapprof,2,[]) in
        let a3 = TP_Append("ABCLASRQ.2DA",k.abclasrq,[],true) in
        let a4 = TP_Append("ABDCDSRQ.2DA",k.abdcdsrq,[],true) in
        let a5 = TP_Append("ABDCSCRQ.2DA",k.abdcscrq,[],true) in
        let a_e1 = TP_Append("ABCLSMOD.2DA",k.abclsmod,[],true) in
        let a_e2 = TP_Append("DUALCLAS.2DA",k.dualclas,[],true) in
        let a6 = TP_Append("ALIGNMNT.2DA",k.alignmnt,[],true) in
        let abil_file = String.uppercase (Filename.basename k.ability_file) in
        let abil_file_no_ext = Filename.chop_extension abil_file in
        let dest_abil_file = "override/" ^ abil_file in
        let a7 = TP_Copy(
          { copy_get_existing = false ;
            copy_use_regexp = false;
            copy_use_glob = false;
            copy_file_list = [(k.ability_file,dest_abil_file)] ;
            copy_patch_list = [] ;
            copy_constraint_list = [] ;
            copy_backup = true ;
            copy_at_end = false;
            copy_save_inlined = false;
            } ) in
        let include_list = split_apart k.include_in in
        let lower_index = match Dc.resolve_tlk_string game k.lower with
          Dlg.TLK_Index(i) -> i
        | _ -> log_and_print "ERROR: cannot resolve KIT lower string\n" ;
                failwith "resolve"
        in
        let mixed_index = match Dc.resolve_tlk_string game k.mixed with
          Dlg.TLK_Index(i) -> i
        | _ -> log_and_print "ERROR: cannot resolve KIT mixed string\n" ;
                failwith "resolve"
        in
        let help_index = match Dc.resolve_tlk_string game k.help with
          Dlg.TLK_Index(i) -> i
        | _ -> log_and_print "ERROR: cannot resolve KIT help string\n" ;
                failwith "resolve"
        in
        let this_kit_number = get_next_line_number "KITLIST.2DA" in
        let this_kit_prof_number = get_next_col_number "WEAPPROF.2DA" in
        let append_to_kitlist = Printf.sprintf
          "%d  %s %d %d %d %s %d %s"
          this_kit_number k.kit_name
          lower_index mixed_index help_index
          abil_file_no_ext this_kit_prof_number
          k.unused_class in
        let a8 = TP_Append("KITLIST.2DA",append_to_kitlist,[],true) in
        let include_actions = List.map (fun file ->
          let num = get_next_line_number (file ^ ".2DA" ) in
          let str = Printf.sprintf "%d  %d" num this_kit_number in
          TP_Append(file ^ ".2DA",str,[],true)
        ) include_list in
        let abbr = Printf.sprintf  "%s     %s" k.kit_name k.tob_abbr in
        let a9 = TP_Append("LUABBR.2DA",abbr,[],true) in
        let a10 = TP_Set_Col("25STWEAP.2DA",
          ("" :: "" :: k.kit_name :: k.tob_start),this_kit_prof_number+2) in
        let a11 = TP_Append("KIT.IDS",
          (Printf.sprintf "0x%x %s" (0x4000 + this_kit_number)
            k.kit_name),[],true) in
        let fix2da1 = TP_Copy ({
            copy_get_existing = true;
            copy_use_regexp = false;
            copy_use_glob = false;
            copy_file_list = ["weapprof.2da", "override";
                              "25stweap.2da", "override";
            ] ;
            copy_patch_list = [
              TP_Read2DA(get_pe_int "2", get_pe_int "0", get_pe_int "0", "tb#kit_temp2");
              TP_Patch2DA(get_pe_int "2", get_pe_int "0", get_pe_int "0", get_pe_int "%tb#kit_temp2% %tb#kit_temp2%");
              TP_PatchSet(get_pe_string "tb#kit_this_is_a_temp_var",get_pe_int "0");
              TP_Patch2DANow("tb#kit_this_is_a_temp_var", get_pe_int "0");
            ] ;
            copy_constraint_list = [] ;
            copy_backup = true;
            copy_at_end = false;
            copy_save_inlined = false;
        }) in
        let action_list = a1 :: a2 :: a3 :: a4 :: a5 :: a6 :: a7 :: a8 ::
          a9 :: a10 :: a11 :: a_e1 :: a_e2 :: fix2da1 :: include_actions  in
        let old_allow_missing = !Load.allow_missing in
        Load.allow_missing :=
          "LUABBR.2DA" :: "25STWEAP.2DA" :: old_allow_missing ;
        (* actually do it! *)
        List.iter (process_action tp) action_list ;
        let fix2da2 = TP_Copy ({
            copy_get_existing = true;
            copy_use_regexp = false;
            copy_use_glob = false;
            copy_file_list = ["weapprof.2da", "override";
                              "25stweap.2da", "override";
            ] ;
            copy_patch_list = [
              TP_PatchStringTextually(None,None,"^%tb#kit_temp2%",
                              String.make (String.length (Var.get_string_exact "%tb#kit_temp2%")) ' ',None);
            ] ;
            copy_constraint_list = [] ;
            copy_backup = true;
            copy_at_end = false;
            copy_save_inlined = false;
        }) in
        process_action tp fix2da2;
        Load.allow_missing := old_allow_missing ;
        Var.set_int32 (k.kit_name) (Int32.of_int this_kit_number) ;
        log_and_print "Added %s Kit\n" k.kit_name;
        Bcs.clear_ids_map game ;
      end
    end

    | TP_String_Set_Range(min,max,tra_file) ->
      begin
        Dc.push_copy_trans ();
        try
          let tra_file = Var.get_string tra_file in
          handle_tra_filename (Arch.backslash_to_slash tra_file) ;
          for i = min to max do
              Dc.set_string game i (Dlg.Trans_String(i)) false 
          done ;
          Dc.pop_trans ()
        with e -> 
          (Dc.pop_trans (); raise e)
      end

    | TP_String_Set(lst,tra_file_opt) -> begin
      (match tra_file_opt with
      | None -> ()
      | Some(tra_file) ->
        begin
        let tra_file = Var.get_string tra_file in
        Dc.push_copy_trans ();
        handle_tra_filename (Arch.backslash_to_slash tra_file)
        end
      ) ;
      List.iter (fun (s1,str) ->
        let i = try [int_of_string s1]
            with _ -> begin
              try
                Hashtbl.find_all game.Load.dialog_search s1
              with _ ->
                log_and_print "ERROR: Cannot find string [%s]\n" s1 ;
                failwith "ERROR: STRING_SET"
             end
        in
        List.iter (fun i -> Dc.set_string game i str false) i
      ) lst ;
      (match tra_file_opt with
      | None -> ()
      | Some(_) -> Dc.pop_trans ())
      end

    | TP_String_Set_Evaluate(lst,tra_file_opt) -> begin
      (match tra_file_opt with
      | None -> ()
      | Some(tra_file) ->
        begin
        let tra_file = Var.get_string tra_file in
        Dc.push_copy_trans ();
        handle_tra_filename (Arch.backslash_to_slash tra_file)
        end
      ) ;
      List.iter (fun (pe,str) ->
        let i = Int32.to_int (eval_pe game pe) in
        if !debug_ocaml then log_and_print "Setting %d to \"%s\"\n" i
          (Dc.single_string_of_tlk_string game str) ;
        Dc.set_string game i str false ;
      ) lst ;
      (match tra_file_opt with
      | None -> ()
      | Some(_) -> Dc.pop_trans ())
      end

    | TP_Load_Tra(str_l) -> begin
        let str_l = List.map (fun x -> Arch.backslash_to_slash (Var.get_string x)) str_l in
        let numtra = List.length str_l in
        log_and_print "loading %d tra file%s\n" numtra
         (if numtra = 1 then "" else "s") ;
        List.iter (fun str ->
            handle_tra_filename (Var.get_string str)
        ) str_l
      end

    | TP_Mkdir(str_l) -> begin

      let str_l = List.map (fun x -> Arch.backslash_to_slash x) str_l in
      let numdir = List.length str_l in
      log_and_print "Creating %d director%s\n" numdir
       (if numdir = 1 then "y" else "ies") ;
      List.iter (fun str ->
        try
          recursive_mkdir (Var.get_string str) 511 (* 511 = octal 0777 = a+rwx *)
        with e ->
          (
            log_and_print "Problem %s on %s : tp.ml\n" (Printexc.to_string e) str
          )
      ) str_l
      end

    | TP_Compile(eval,dlg_l,tra_l) -> begin
        let dlg_l =  List.map (fun x -> Arch.backslash_to_slash x) dlg_l in
        let tra_l =  List.map (fun x -> Arch.backslash_to_slash x) tra_l in
        let numd = ref 0 in
        let nums = ref 0 in 
        let handle_one_d_file filespec = match split 
          (String.uppercase filespec) with
        | _,"BAF" -> incr nums 
        | _,"D" -> incr numd
        | _,_ -> () 
        in 
        List.iter (fun filespec ->
          if is_directory filespec then begin
            let dh = Case_ins.unix_opendir filespec in 
            (try
              while true do
                let base = Unix.readdir dh in 
                let source = filespec ^ "/" ^ base in
                if not (is_directory source) then
                  handle_one_d_file source
              done
            with _ -> () ) ;
            Unix.closedir dh 
          end else handle_one_d_file filespec 
        ) dlg_l ; 
        log_and_print "Compiling " ;
        (if (!numd > 0) then begin
          log_and_print "%d dialogue file%s " !numd
            (if (!numd > 1) then "s" else "") ;
          if (!nums > 0) then
            log_and_print "and "
          end
        ) ;
        (if (!nums > 0) then begin
          log_and_print "%d script%s " !nums
            (if (!nums > 1) then "s" else "") ;
          end
        ) ; 
        log_and_print "...\n" ;

        Dc.push_copy_trans ();
        begin 
          match !our_lang with
            Some(l) -> List.iter (fun path ->
                let my_regexp = Str.regexp_string "%s" in
                let tra_file = Str.global_replace 
                  my_regexp l.lang_dir_name (Var.get_string path) in
                handle_tra_filename tra_file ;
              ) tra_l 
          | _ -> List.iter (fun tra_file -> 
                handle_tra_filename tra_file ;
              ) tra_l 
        end ;
        let handle_one_d_file d = 
          (* handle AUTO_TRA "solarom/%s" *)
          Dc.push_copy_trans () ;
          List.iter (fun f -> 
            match f,!our_lang with
            Auto_Tra(path),Some(l) -> 
                  let my_regexp = Str.regexp_string "%s" in
                  let tra_file_dir = Str.global_replace
                    my_regexp l.lang_dir_name path in
                  let d_base,_ = split (Filename.basename d) in
                  let tra_file = tra_file_dir ^ "/" ^ d_base ^ ".TRA" in
                  handle_tra_filename tra_file ;
          | Auto_Tra(path),None ->
                  let d_base,_ = split (Filename.basename d) in
                  let tra_file = path ^ "/" ^ d_base ^ ".TRA" in 
                  handle_tra_filename tra_file
          | _ -> () 
          ) tp.flags ;
          Dc.ok_to_resolve_strings_while_loading := Some(game) ; 
          (try 
            begin 
            match split (String.uppercase (Filename.basename d)) with
            | _,"BAF" -> compile_baf_filename d
            | _,"D" -> handle_d_filename d 
            | _,_ -> () 
            end 
          with e -> 
            begin 
              Dc.ok_to_resolve_strings_while_loading := None ; 
              Dc.pop_trans () ; 
              log_and_print "ERROR: compiling [%s]!\n" d ;
              raise e
            end
          ) ;
          Dc.ok_to_resolve_strings_while_loading := None ; 
          Dc.pop_trans () ; 
        in 
        List.iter (fun filespec -> 
          if is_directory filespec then begin
            let dh = Case_ins.unix_opendir filespec in
            (try
              while true do
                let base = Unix.readdir dh in
                let source = filespec ^ "/" ^ base in
                if not (is_directory source) then
                  handle_one_d_file source
              done
            with End_of_file -> () ) ;
            Unix.closedir dh ;
          end else handle_one_d_file filespec
        ) dlg_l ;
        log_or_print "Processing %d dialogues/scripts ...\n"
          (List.length dlg_l) ;
        emit_dlg_files game "override" ;
        Dc.pop_trans () ;
        end

    | TP_Set_Col(file,new_col_list,col_num) ->
        log_and_print "Setting game text column-wise ...\n" ;
        let eight,three = split (String.uppercase file) in
        let buff,loaded_path = Load.load_resource "SET_COLUMN" game true eight three in
        if buff = "" then
          log_or_print "[%s]: empty or does not exist\n" file
        else begin
          let dest = "override/" ^ file in
          let buff_as_lines = Str.split many_newline_or_cr_regexp buff in
          if List.length buff_as_lines <> List.length new_col_list then begin
            log_and_print "Cannot set column-wise because there are %d lines in %s but I was only given %d things to append\n" (List.length buff_as_lines) file (List.length new_col_list)  ;
            failwith ("cannot set column-wise to " ^ file)
          end ;
          Stats.time "saving files" (fun () -> 
          let out = open_for_writing dest true in
          List.iter2 (fun orig_line new_col ->
            let line_as_cols = Str.split many_whitespace_regexp orig_line in 
            let i = ref 0 in
            List.iter (fun orig_col -> 
              (if !i = col_num then 
                Printf.fprintf out "%-20s" new_col
              else
                Printf.fprintf out "%-20s" orig_col) ;
              incr i;
            ) line_as_cols ;
            (if (!i <= col_num) then 
                Printf.fprintf out "%-20s" new_col) ; 
            output_string out "\r\n" 
          ) buff_as_lines new_col_list ; 
          close_out out ;
          begin (* handle read-only files! *)
            try
              Case_ins.unix_chmod dest 511 ; (* 511 = octal 0777 = a+rwx *)
            with e -> ()
              (* log_or_print "WARNING: chmod %s : %s\n" filename
                (Printexc.to_string e) *)
          end ;) () ;
          log_or_print "Set text in [%s] column-wise\n" file
        end 

    | TP_Append_Col(file,src,count_prepend,con_l) ->
        let file = Arch.backslash_to_slash file in
        let temp = Var.get_string src in
        let src_list = Str.split (Str.regexp "[ \t]+") temp in
        let src_list = List.map (fun elt -> if elt = "$" then "" else elt) src_list in
        let rec prepend i string_list =
            if i = 1 then
              ("" :: string_list)
            else begin
              if i > 1 then
                "" :: (prepend (i-1) string_list)
              else  (* i <= 0 *)
                   string_list
               end
        in

        let src_list = prepend count_prepend src_list in
        log_and_print "Appending to files column-wise ...\n" ;
        let eight,three = split (String.uppercase file) in
        let buff,loaded_path = Load.load_resource "APPEND_COLUMN" game true eight three in
        if buff = "" then
          log_or_print "[%s]: empty or does not exist\n" file
        else begin 
        let okay = List.fold_left (fun acc elt -> acc &&
          match elt with 
            TP_Contains(s) -> begin
              let my_regexp = Str.regexp_case_fold s in
              try let _ = Str.search_forward my_regexp buff 0 in
                log_only "Appending cols to [%s] because it DOES contain [%s]\n" file s  ; 
                true
              with _ -> 
                log_only "Not appending cols to [%s] because it does NOT contain [%s]\n" file s ; 
                false
              end
          | TP_NotContains(s) -> begin
              let my_regexp = Str.regexp_case_fold s in
              try let _ = Str.search_forward my_regexp buff 0 in 
                log_only "Not appending cols to [%s] because it DOES contains [%s]\n" file s ; 
                false
              with _ -> 
                log_only "Appending cols to [%s] because it does NOT contain [%s]\n" file s ; 
                true
              end
          | TP_IfSizeIs(size) -> String.length buff = size 
          | TP_ButOnlyIfItChanges -> true 
          | TP_Eval(pe) ->
                let v = eval_pe game pe in 
                if v = Int32.zero then begin
                  log_only "Not appending cols to [%s] because condition evaluates to %ld\n" file v ; false
                end else begin 
                  log_only "Appending cols to [%s] because condition evaluates to %ld\n" file v ; true
                end 
        ) true con_l in 
        if okay then begin (* do the append *) 
          let dest = "override/" ^ file in 
          let buff_as_lines = Str.split many_newline_or_cr_regexp buff in
          if List.length buff_as_lines <> List.length src_list then begin
            log_and_print "Cannot append column-wise because there are %d lines in %s but I was only given %d things to append\n" (List.length buff_as_lines) file (List.length src_list)  ;
            failwith ("cannot append column-wise to " ^ file)
          end ;
          Stats.time "saving files" (fun () -> 
          let out = open_for_writing dest true in
          List.iter2 (fun orig app ->
            output_string out orig ;
            output_string out " " ;
            output_string out app ;
            output_string out "\r\n" 
          ) buff_as_lines src_list ; 
          close_out out ;
          begin (* handle read-only files! *)
            try
              Case_ins.unix_chmod dest 511 ; (* 511 = octal 0777 = a+rwx *)
            with e -> ()
              (* log_or_print "WARNING: chmod %s : %s\n" filename
                (Printexc.to_string e) *)
          end ;) () ;
          log_or_print "Appended text to [%s] column-wise\n" file
        end 
        end

    | TP_Append(file,src,con_l,frombif) ->
        let file = Arch.backslash_to_slash file in
        if Filename.check_suffix(String.lowercase file) "ids" then Bcs.clear_ids_map game ;
        log_and_print "Appending to files ...\n" ;
        let file = Var.get_string file in
        let src = Var.get_string src in
        let buff = if frombif then begin
            let eight,three = split (String.uppercase file) in
            let the_buff,loaded_path = Load.load_resource "APPEND" game true eight three in
            the_buff
          end else begin
            load_file file
          end
        in
        let okay = List.fold_left (fun acc elt -> acc &&
          match elt with 
            TP_Contains(s) -> begin
              let my_regexp = Str.regexp_case_fold s in
              try let _ = Str.search_forward my_regexp buff 0 in
                log_only "Appending [%.10s...] to [%s] because it DOES contain [%s]\n" src file s  ; 
                true
              with _ -> 
                log_only "Not appending [%.10s...] to [%s] because it does NOT contain [%s]\n" src file s ; 
                false
              end
          | TP_NotContains(s) -> begin
              let my_regexp = Str.regexp_case_fold s in
              try let _ = Str.search_forward my_regexp buff 0 in 
                log_only "Not appending [%.10s...] to [%s] because it DOES contains [%s]\n" src file s ; 
                false
              with _ -> 
                log_only "Appending [%.10s...] to [%s] because it does NOT contain [%s]\n" src file s ; 
                true
              end
          | TP_IfSizeIs(size) -> String.length buff = size 
          | TP_ButOnlyIfItChanges -> true
          | TP_Eval(pe) -> 
                let v = eval_pe game pe in 
                if v = Int32.zero then begin
                  log_only "Not appending [%.10s] to [%s] because condition evaluates to %ld\n" src file v ; false
                end else begin 
                  log_only "Appending [%.10s] to [%s] because condition evaluates to %ld\n" src file v ; true
                end 
        ) true con_l in 
        if okay then begin (* do the append *) 
          let dest =
            if frombif then "override/" ^ file
            else file
          in
          Stats.time "saving files" (fun () -> 
          let out = open_for_writing dest true in
          let nice_newlines = Str.global_replace
            many_newline_or_cr_regexp "\r\n" (buff ^ "\r\n") in
          output_string out nice_newlines ; 
          (* output_string out buff ;
          output_string out "\r\n" ; *)
          output_string out src ;
          output_string out "\r\n" ;
          close_out out ;
          begin (* handle read-only files! *)
            try
              Case_ins.unix_chmod dest 511 ; (* 511 = octal 0777 = a+rwx *)
            with e -> ()
              (* log_or_print "WARNING: chmod %s : %s\n" filename
                (Printexc.to_string e) *)
          end ;) () ;
          log_or_print "Appended text to [%s]\n" file
        end 

    | TP_Extend_Top(use_reg,dest,src,pl,tra_l)
    | TP_Extend_Bottom(use_reg,dest,src,pl,tra_l) -> begin
        let dest = if not use_reg then Arch.backslash_to_slash dest else dest in
        let tra_l = List.map (fun x -> Arch.backslash_to_slash x) tra_l in
        log_and_print "Extending game scripts ...\n" ;
        let dest = (Var.get_string dest) in
        let src = (Var.get_string src) in
        let src = Arch.backslash_to_slash src in
        let dlist =
          if use_reg = false then
            [dest]
          else begin
            let files_in_chitin = Key.list_of_key_resources game.Load.key in
            let regexp = Str.regexp_case_fold dest in
            let matches = ref [] in
            List.iter (fun possible ->
              if Str.string_match regexp possible 0 then begin
                matches := (possible) :: !matches
              end
            ) files_in_chitin ;
            if (!matches = []) then
              [dest]
            else
              !matches
          end
        in
        Dc.push_copy_trans () ;
        begin
          match !our_lang with
            Some(l) -> List.iter (fun path ->
                let my_regexp = Str.regexp_string "%s" in
                let tra_file = Str.global_replace
                  my_regexp l.lang_dir_name  (Var.get_string path) in
                handle_tra_filename tra_file ;
              ) tra_l 
          | _ -> List.iter (fun tra_file -> 
                handle_tra_filename tra_file ; 
              ) tra_l 
        end ;
        let src_script = 
          try 
            let src_buff = load_file src in
            List.iter (fun p -> process_patch1 src game src_buff p) pl ;
            let src_buff = List.fold_left (fun acc elt ->
                  try process_patch2 src game acc elt
                  with e -> 
                    log_and_print "ERROR: [%s] -> [%s] Patching Failed (EXTEND_TOP/BOTTOM)\n"
                      src dest ; raise e) 
                src_buff pl 
            in 
            Dc.ok_to_resolve_strings_while_loading := Some(game) ; 
            (try 
              let res = handle_script_buffer src src_buff in
              Dc.ok_to_resolve_strings_while_loading := None ; 
              res
            with e -> 
              begin 
              Dc.ok_to_resolve_strings_while_loading := None ; 
              raise e
              end
            ) ;
          with _ -> []
        in 
        List.iter (fun dest -> 
          let eight,three = split (String.uppercase dest) in 
          let dest_script = 
            let old_a_m = !Load.allow_missing in 
            Load.allow_missing := dest :: old_a_m ;
            let dest_buff, dest_path = 
              try 
                Load.load_resource "EXTEND_TOP/EXTEND_BOTTOM" game true eight three
              with _ -> 
                begin 
                log_only "[%s] not found, treating as empty.\n" dest ;
                "",""
                end 
            in
            Load.allow_missing := old_a_m ; 
            handle_script_buffer dest dest_buff
          in 

          let destpath = "override/" ^ dest in 
          Stats.time "saving files" (fun () -> 
          let out = open_for_writing destpath true in
          Bcs.save_bcs game (Bcs.Save_BCS_OC(out)) (match a with
            TP_Extend_Top(_,_,_,_,_) -> src_script @ dest_script
                                 | _ -> dest_script @ src_script) ;
          close_out out ;
          begin (* handle read-only files! *)
            try
              Case_ins.unix_chmod destpath 511 ; (* 511 = octal 0777 = a+rwx *)
            with e -> ()
              (* log_or_print "WARNING: chmod %s : %s\n" filename
                (Printexc.to_string e) *)
          end ;) () ;
          log_or_print "Extended script [%s] with [%s]\n" dest src
        ) dlist ;
        Dc.pop_trans () ; 
        end

    | TP_At_Interactive_Exit(str) ->
        if !interactive then process_action tp (TP_At_Exit(str))
    | TP_At_Interactive_Uninstall(str) ->
        if !interactive then process_action tp (TP_At_Uninstall(str))
    | TP_At_Interactive_Now(str) ->
        if !interactive then process_action tp (TP_At_Now(str))

    | TP_At_Exit(str) ->
        begin
        let str = Var.get_string str in
        let a,b = split (String.uppercase str) in 
        match b with
        | "TP2" -> (!handle_recursive_tp2) str
        | _ ->
          let str = Arch.handle_view_command str !skip_at_view in
          if List.mem str !execute_at_exit then
            ()
          else
            execute_at_exit := str :: !execute_at_exit
        end

    | TP_At_Now(str) ->
        begin
        let str = Var.get_string str in
        let a,b = split (String.uppercase str) in 
        match b with 
        | "TP2" -> (!handle_recursive_tp2) str 
        | _ ->
          let str = Arch.handle_view_command str !skip_at_view in
          ignore (Unix.system (Arch.slash_to_backslash str))
        end

    | TP_At_Uninstall(str) -> ()
    )
    with e -> (* from: let rec process_action = try *)
      (if !continue_on_error then begin
        log_and_print "WARNING: Continuing despite [%s]\n"
          (Printexc.to_string e)
      end else begin
        return_value := return_value_error_tp2_component_install ; 
        log_and_print "Stopping installation because of error.\n" ; 
        raise e
      end)
      ) () 
  in


  let subcomp_group the_comp =
    let rec walk lst = match lst with
    | TPM_SubComponents(ts,_,_) :: tl -> Some(ts)
    | hd :: tl -> walk tl
    | [] -> None
    in walk the_comp.mod_flags
  in
  let subcomp_predicate the_comp =
    let rec walk lst = match lst with
    | TPM_SubComponents(_,p,_) :: tl -> is_true (eval_pe game p)
    | hd :: tl -> walk tl
    | [] -> false
    in walk the_comp.mod_flags
  in
  let subcomp_forced the_comp=
    let rec walk lst = match lst with
    | TPM_SubComponents(_,_,b) :: tl -> b
    | hd :: tl -> walk tl
    | [] -> false
    in walk the_comp.mod_flags
  in
  let fails_requirements m = List.exists (fun f -> match f with
      | TPM_RequireComponent(s,i,warn) ->
          not (already_installed s i)
      | TPM_ForbidComponent(s,i,warn) ->
          (already_installed s i)
      | TPM_Deprecated(warn) ->
          true
      | TPM_RequirePredicate(p,warn) ->
          not (is_true (eval_pe game p))
      | _ ->
          false
    ) m.mod_flags
    ||
    List.exists (fun f -> match f with
      | TP_Require_File(file,warn) ->
          not (bigg_file_exists file game.Load.key)
      | TP_Forbid_File(file,warn) ->
          bigg_file_exists file game.Load.key
      | _ -> false
    ) m.mod_parts

  in

  let last_module_index = get_last_module_index tp in
  let comp_num = ref 0 in
  let comp_ht = Hashtbl.create 255 in
  List.iter (fun the_mod ->
    incr comp_num ; 
    (match subcomp_group the_mod with
    | Some(ts) -> 
        if Hashtbl.mem comp_ht ts then
          decr comp_num
        else
          Hashtbl.add comp_ht ts true 
    | None -> ()) 
  ) tp.module_list ; 
  let comp_num = !comp_num in
  (* comp_num = number of user visible "component chunks", each of which
   * can have multiple sub-components *)
   
  let any_member_of_subcomp_group_installed the_comp = 
    let i = ref 0 in
    List.exists (fun the_mod ->
      List.iter (fun f -> match f with
      | TPM_Designated(i') -> i := i' ;
      | _ -> ()
      ) the_mod.mod_flags ; 
      match subcomp_group the_mod with
      | Some(ts) -> if ts = the_comp &&
                       already_installed this_tp2_filename !i then
                      true
                    else (incr i ; false)
      | None -> (incr i; false)
    ) tp.module_list
  in


  let any_already_installed = ref false in 
  let any_not_yet_installed = ref false in 

  let subcomp_installed = Hashtbl.create 255 in 

  for i = 0 to last_module_index do 
    try let m = get_nth_module tp i false in
    if already_installed this_tp2_filename i then begin 
      any_already_installed := true ;
      match subcomp_group m with
      | Some(ts) -> Hashtbl.add subcomp_installed ts true
      | None -> ()
    end else begin
      let deprecated = List.exists (fun f -> match f with 
        | TPM_Deprecated(warn) -> true
        | _ -> false) m.mod_flags 
      in 
      if not deprecated then begin
        let group_already =
          match subcomp_group m with
          | Some(ts) -> Hashtbl.mem subcomp_installed ts
          | None -> false
        in
        if not group_already then
          any_not_yet_installed := true
      end
    end
    with Not_found -> ()
  done ;

  let rec hasgroup x =
    match x with
    | TPM_Group(c)::b -> true
    | a::b -> hasgroup b
    | [] -> false
  in

  let module_defaults = Array.init (last_module_index+1) (fun i ->
      try
        let cli_uninstall = ref (List.exists (fun h -> h = i) !force_uninstall_these)  in
        let cli_install   = ref (List.exists (fun h -> h = i) !force_install_these  )  in
        if !always_yes or !cli_install then
          TP_Install
        else if (!always_uninstall or !cli_uninstall) && (already_installed this_tp2_filename i) then
          TP_Uninstall
        else if (!always_uninstall or !cli_uninstall) then
          TP_Skip
        else if !sometimes_reinstall && (already_installed this_tp2_filename i) then
          TP_Install
        else if !sometimes_reinstall then
          TP_Skip
        else if (!force_install_these <> []) || (!force_uninstall_these <> []) then
          TP_Skip
        else if (hasgroup (get_nth_module tp i false).mod_flags) then
          TP_Skip
        else
          TP_Ask
      with Not_found -> TP_Ask)
  in

  let get_trans i = Dc.single_string_of_tlk_string game (Dlg.Trans_String(i)) in

  let handle_letter tp answer can_uninstall package_name m finished i =
    let subgroup_already =
      match subcomp_group m with
      | Some(ts) ->
          let res = any_member_of_subcomp_group_installed ts in
          res
      | None ->
          false
    in

    match answer with
    | "Q" -> begin
        for i = 0 to last_module_index do
          try
            let c = get_nth_module tp i false in
            if subcomp_forced c && not (fails_requirements c) then module_defaults.(i) <- TP_Install
            else module_defaults.(i) <- TP_Skip ;
            if !debug_ocaml then log_and_print "component %d is %s\n" i (match module_defaults.(i) with
                                                                         | TP_Install -> "Install"
                                                                         | TP_Skip -> "Skip"
                                                                         | _ -> "Internal Error");
          with Not_found -> ();
          finished := true
        done
        end

    | "I" | "Y" when subgroup_already && not (already_installed tp.tp_filename i) ->
        log_or_print "Skipping [%s] because another subcomponent of [%s] is already installed.\n" package_name
        (match subcomp_group m with
          | Some(ts) -> Dc.single_string_of_tlk_string game  ts
          | None -> "???") ;
        finished := true

    | "I" | "Y" | "R" -> begin
        if can_uninstall then begin
          try
            (* log_and_print
            "\nRemoving old installation of [%s] first ...\n" *)
            log_and_print "\n%s%s%s\n"
              ((get_trans (-1013))) package_name
              ((get_trans (-1014))) ;
            (if not (uninstall game handle_tp2_filename this_tp2_filename i !interactive) then failwith "uninstallation error");
            log_and_print
              (* "\nSUCCESSFULLY REMOVED OLD [%s]\n\n"  *)
              "\n%s [%s]\n\n"
              ((get_trans (-1015))) package_name ;
          with e ->
            log_and_print "WARNING: unable to uninstall: %s\n"
              (Printexc.to_string e)
        end ;
        set_backup_dir tp.backup i ;

        let strset_backup_filename =
          Printf.sprintf "%s/%d/UNSETSTR.%d" tp.backup i i
        in

        let args_backup_filename =
          Printf.sprintf "%s/%d/ARGS.%d" tp.backup i i
        in
        
        let readln_backup_filename =
          Printf.sprintf "%s/%d/READLN.%d" tp.backup i i
        in

        if !interactive && (!Var.cli_variables = None) then begin
          let counter = ref 0 in
          let continue = ref true in
          Var.cli_variables := Some([]) ;
          while !continue do
            let varname = Printf.sprintf "%%argv[%d]%%" !counter in
            let varcont = Var.get_string varname in
            if varcont = varname then begin
              continue := false
            end else begin
              if !debug_ocaml then log_and_print "%%argv[%d]%% = %s\n" !counter varcont;
              Var.cli_variables := Some(varcont :: (value_of_option !Var.cli_variables)) ;
              incr counter;
            end ;
          done ;
          Var.cli_variables := Some(value_of_option !Var.cli_variables)
        end ;
        if !interactive then begin
          let outchan = Case_ins.perv_open_out_bin args_backup_filename in
          Marshal.to_channel outchan (List.rev (value_of_option !Var.cli_variables)) [] ;
          close_out outchan ;
        end ;
        if (* not !interactive && *) file_exists args_backup_filename then begin
        (* re-loading when interactive and clearing of old argv[] variables is intended. *)
          let counter = ref 0 in
          let continue = ref true in
          while !continue do
            let varname = Printf.sprintf "%%argv[%d]%%" !counter in
            let varcont = Var.get_string varname in
            if varcont = varname then begin
              continue := false
            end else begin
              if !debug_ocaml then log_and_print "removing the pair %%argv[%d]%% = %s\n" !counter varcont;
              Hashtbl.remove Var.variables varname ;
              incr counter;
            end ;
          done ;
          try
            let infile = Case_ins.perv_open_in_bin args_backup_filename in
            let record : string list =
              Marshal.from_channel infile
            in
            close_in infile ;
            let counter = ref 0 in
            List.iter (fun s ->
              Var.set_string ("argv[" ^ (string_of_int !counter) ^ "]") s;
              incr counter
            ) record
          with e ->
            log_and_print "WARNING: Unable to read --args references from [%s]: %s\n"
                                         args_backup_filename (Printexc.to_string e)
        end ;

        log_and_print "\n%s [%s]\n"
          (* "\nInstalling [%s]\n"  *)
              ((get_trans (-1016)))
          package_name ;
        (try
          Var.set_int32 "COMPONENT_NUMBER" (Int32.of_int i) ;

          List.iter (fun flag -> match flag with
            Always(al) -> List.iter (process_action tp) al
          | TP_No_If_Eval () -> has_if_eval_bug := false ;
          | Define_Action_Macro(str,decl,al) ->
            Hashtbl.replace action_macros str (decl, al)
          | Define_Patch_Macro(str,decl,al) ->
            Hashtbl.replace patch_macros str (decl, al)
          | Load_Macro(str_list) ->
            List.iter (fun str ->
              match str with
              | "tb#factorial" ->
                Hashtbl.replace patch_macros "tb#factorial" (
                  [
                    TP_LocalSet(get_pe_string("tb#factorial_index"), get_pe_int("tb#factorial_index"))
                  ], [
                    TP_PatchIf (PE_Not(PE_Equal(get_pe_int("tb#factorial_index"),get_pe_int("1"))),
                    [
                      TP_PatchSet(get_pe_string("tb#factorial_result"),
                        PE_Mul(get_pe_int("tb#factorial_result"),get_pe_int("tb#factorial_index"))) ;
                      TP_PatchSet(get_pe_string("tb#factorial_index"),
                        PE_Sub(get_pe_int("tb#factorial_index"),get_pe_int("1"))) ;
                      TP_Launch_Patch_Macro("tb#factorial")
                    ],[])
                  ]
                ) ;
              | "tb#fix_file_size" ->
                Hashtbl.replace action_macros "tb#fix_file_size" (
                  [
                    TP_LocalSprint(get_pe_string "file",get_pe_tlk_string "") ;
                    TP_LocalSprint(get_pe_string "old",get_pe_tlk_string "") ;
                    TP_LocalSet(get_pe_string "i",get_pe_int "0") ;
                  ] ,
                  [
                    TP_If (
                      PE_Not(Pred_File_Exists_In_Game(get_pe_string "tb#removedF%tb#fix_file_size_category%.txt")) ,
                      [
                        TP_Inlined_File("tb#fix_file_size_base.txt", " ") ;
                        TP_Copy({
                          copy_get_existing = false ;
                          copy_use_regexp = false;
                          copy_use_glob = false ;
                          copy_file_list = [("tb#fix_file_size_base.txt", "override/tb#removedF%tb#fix_file_size_category%.txt")] ;
                          copy_patch_list = [] ;
                          copy_constraint_list = [] ;
                          copy_backup = true ;
                          copy_at_end = false ;
                          copy_save_inlined = false ;
                        }) ;
                        TP_Copy({
                          copy_get_existing = true ;
                          copy_use_regexp = false;
                          copy_use_glob = true ;
                          copy_file_list = [("%tb#fix_file_size_regexp%", "override")] ;
                          copy_patch_list = [
                            TP_PatchIf (
                              PE_LT(get_pe_int "%SOURCE_SIZE%", get_pe_int "%tb#fix_file_size_min%"),
                              [
                                TP_PatchSprint(get_pe_string "file",
                                  get_pe_tlk_string "%SOURCE_FILE%") ;
                                TP_PatchInnerAction(
                                  [
                                    TP_Append("tb#removedF%tb#fix_file_size_category%.txt","%file%",[],true)
                                  ]
                                ) ;
                              ] ,
                              []
                            ) ;
                          ] ;
                          copy_constraint_list = [ TP_ButOnlyIfItChanges ] ;
                          copy_backup = true ;
                          copy_at_end = false ;
                          copy_save_inlined = false ;
                        }) ;
                        TP_Copy({
                          copy_get_existing = true ;
                          copy_use_regexp = false;
                          copy_use_glob = false ;
                          copy_file_list = [("tb#removedF%tb#fix_file_size_category%.txt", "override")] ;
                          copy_patch_list = [
                            TP_Read2DANow("tb#fix_file_size_list",get_pe_int "1") ;
                            TP_PatchFor(
                              [
                                TP_PatchSet(get_pe_string "i", get_pe_int "0");
                              ],
                              PE_LT(get_pe_int "%i%",get_pe_int "%tb#fix_file_size_list%"),
                              [
                                TP_PatchSet(get_pe_string "i", PE_Add(get_pe_int "%i%",get_pe_int "1")) ;
                              ],
                              [
                                TP_Read2DAFormer("tb#fix_file_size_list", get_pe_int "%i%", get_pe_int "0", "old") ;
                                TP_PatchInnerAction([
                                  TP_Copy({
                                    copy_get_existing = true ;
                                    copy_use_regexp = false;
                                    copy_use_glob = false ;
                                    copy_file_list = [("%tb#fix_file_size_target%", "override/%old%")] ;
                                    copy_patch_list = [] ;
                                    copy_constraint_list = [] ;
                                    copy_backup = true ;
                                    copy_at_end = false ;
                                    copy_save_inlined = false ;
                                  }) ;
                                ]) ;
                              ]
                            ) ;
                          ] ;
                          copy_constraint_list = [ TP_ButOnlyIfItChanges ] ;
                          copy_backup = true ;
                          copy_at_end = false ;
                          copy_save_inlined = false ;
                        }) ;
                      ] , []) ;
                  ]
                ) ;
                | _ ->
                  failwith (Printf.sprintf "Unknown standard macro: -%s-" str) ;
            ) str_list
          | _ -> ()
          ) tp.flags ;
          if file_exists readln_backup_filename && not !interactive then begin
            try
              let infile = Case_ins.perv_open_in_bin readln_backup_filename in
              let record : (tp_pe_string * string) list =
                Marshal.from_channel infile
              in
              close_in infile ;
              readln_strings := record
            with e ->
              log_and_print "WARNING: Unable to read readln references from [%s]: %s\n"
                                           args_backup_filename (Printexc.to_string e)
          end ;
          List.iter (process_action tp) m.mod_parts ;
          if !interactive then begin
            let outchan = Case_ins.perv_open_out_bin readln_backup_filename in
            Marshal.to_channel outchan (List.rev !readln_strings) [] ;
            close_out outchan ;
          end ;
          readln_strings := [] ;
          be_silent := false ;
        with e -> begin
          be_silent := false ;
          return_value := return_value_error_tp2_component_install ;
          log_and_print "\n%s%s%s\n"
          (*  "\nERROR Installing [%s], rolling back to previous state\n" *)
              ((get_trans (-1017))) package_name ((get_trans (-1018))) ;
          Dc.clear_state () ;
          record_strset_uninstall_info game strset_backup_filename ;
          (match !backup_list_chn with
            Some(chn) -> close_out chn ; backup_list_chn := None
          | None -> ()) ;
          uninstall_tp2_component game tp this_tp2_filename i false ;
          print_log () ;
          raise e
        end );
        log_and_print "\n\n" ;
        record_strset_uninstall_info game strset_backup_filename ;
        strings_to_print_at_exit :=
          (Printf.sprintf "\n%s [%s]\n"
           (*  "\nSUCCESSFULLY INSTALLED [%s]\n"  *)
              ((get_trans (-1019)))
            package_name) ::
          !strings_to_print_at_exit ;
        (* add this successful install to the log! *)
        begin
        if List.find_all (fun x -> x = TPM_NotInLog) m.mod_flags = [] then
          the_log := !the_log @
            [ ((String.uppercase this_tp2_filename),!our_lang_index,i,Some(package_name),Installed) ]
        else log_and_print "NOT adding a WeiDU.log record. You cannot uninstall this.\n"
        end ;
        finished := true
      end
    | "N" ->
        log_and_print "\n%s [%s]\n"
         (* "\nSkipping [%s]\n" *)
              ((get_trans (-1020)))
          package_name ;
        finished := true
    | "U" when not can_uninstall ->
      log_and_print "\nYou can't uninstall the non-installed component [%s] (component #%d)\n"
               package_name i ;
      finished := false ;
    | "U" ->
        log_and_print "\n%s%s%s%d%s\n"
         (* "\nRemoving [%s] (component #%d)\n" *)
              ((get_trans (-1021)))
              package_name
              ((get_trans (-1022)))
              i
              ((get_trans (-1023))) ;
        (if not (uninstall game handle_tp2_filename this_tp2_filename i !interactive ) then failwith "uninstallation error" );
        log_and_print "\n\n%s%s%s%d%s\n"
         (* "\n\nSUCCESSFULLY REMOVED [%s] (component #%d)\n\n" *)
              ((get_trans (-1024)))
              package_name
              ((get_trans (-1022)))
              i
              ((get_trans (-1023))) ;
        finished := true
    | _ -> ()
  in

  let specify = ref false in

  let rec findgroup flags =
    match flags with
      | TPM_Group(x)::b -> Some(x)
      | a::b -> findgroup b
      | [] -> None
  in

  let groups = ref [] in
  List.iter (fun x ->
     match findgroup x.mod_flags with
    | Some(a) -> if not (List.mem a !groups) then begin
      groups := a :: !groups;
      end ;
    | None -> ()) (List.rev tp.module_list) ;
  let hasgroups = !groups <> [] in

  let has_ask_every = !ask_all ||
    List.exists (fun a -> a = Ask_Every_Component) tp.flags
  in

  (* for big mods, ask about things in general first *)
  if comp_num > 4 && not !always_yes && not !always_uninstall &&
     not !sometimes_reinstall &&
     not (has_ask_every) && not (!specified_specific_components) && not hasgroups
  then begin
  (* add (-1000) "\nThis mod has %d distinct optional components.\nTo save time, you can choose what to do with them at a high level rather\nthan being asked about each one.\n" ; *)
      log_and_print "\n%s %d %s" (Var.get_string(get_trans (-1000))) comp_num (get_trans (-1001)) ;
      let finished = ref false in

      if !any_not_yet_installed then
      while not !finished do
        finished := true ;
        (* log_and_print "\nWhat should be done with all components that are NOT YET installed?\n[I]nstall them, [S]kip them, [A]sk about each one? " ; *)
        log_and_print "\n%s" ((get_trans (-1002)));
        match String.uppercase(read_line ()) with
        | "R"
        | "I" ->
          for i = 0 to last_module_index do
            try
            let the_comp = get_nth_module tp i false in
            match subcomp_group the_comp with
            | Some(x) ->
             (* if not (any_member_of_subcomp_group_installed x) then *)
                module_defaults.(i) <- TP_Ask
            | None ->
              if not (already_installed this_tp2_filename i) then
                module_defaults.(i) <- TP_Install
            with Not_found -> ()
          done
        | "S"
        | "Q" ->
          for i = 0 to last_module_index do
            try
            let the_comp = get_nth_module tp i false in
            let c = get_nth_module tp i false in
            if subcomp_forced c && not (fails_requirements c) then module_defaults.(i) <- TP_Install
            else begin
              match subcomp_group the_comp with
              | Some(x) ->
                if not (any_member_of_subcomp_group_installed x) then
                  module_defaults.(i) <- TP_Skip
              | None ->
                if not (already_installed this_tp2_filename i) then
                  module_defaults.(i) <- TP_Skip
            end
            with Not_found -> ()
          done
        | "A" -> ()
        | "X" -> specify := true
        | _ -> finished := false
      done ;

      finished := false ;
      if !any_already_installed then
      while not !finished do
        finished := true ;
        (* log_and_print "\nWhat should be done with all components that are ALREADY installed?\n[R]e-install them, [U]ninstall them, [S]kip them, [A]sk about each one? " ; *)
        log_and_print "\n%s" ((get_trans (-1003)));
        match String.uppercase(read_line ()) with
        | "I"
        | "R" ->
          for i = 0 to last_module_index do
            try
            let the_comp = get_nth_module tp i false in
            match subcomp_group the_comp with
            | Some(x) ->
             (* if not (any_member_of_subcomp_group_installed x) then *)
                module_defaults.(i) <- TP_Ask
            | None ->
              if (already_installed this_tp2_filename i) then
                module_defaults.(i) <- TP_Install
            with Not_found -> ()
          done
        | "S"
        | "Q" ->
          for i = 0 to last_module_index do
            try let the_comp = get_nth_module tp i false in
            let c = get_nth_module tp i false in
            if subcomp_forced c && not (fails_requirements c) then module_defaults.(i) <- TP_Install
            else begin
              match subcomp_group the_comp with
              | Some(x) ->
                if (any_member_of_subcomp_group_installed x) then
                  module_defaults.(i) <- TP_Skip
              | None ->
                if (already_installed this_tp2_filename i) then
                  module_defaults.(i) <- TP_Skip
            end
            with Not_found -> ()
          done
        | "U" ->
          for i = 0 to last_module_index do
            try let the_comp = get_nth_module tp i false in
            ( match subcomp_group the_comp with
            | Some(x) ->
              if (any_member_of_subcomp_group_installed x) then
                module_defaults.(i) <- TP_Skip
            | None -> () ) ;
              if (already_installed this_tp2_filename i) then
                module_defaults.(i) <- TP_Uninstall
            with Not_found -> ()
          done
        | "A" -> ()
        | "X" -> specify := true
        | _ -> finished := false
      done ;
  end ;

  let is_my_group the_comp group =
    let rec walk lst = match lst with
    | TPM_Group(ts) :: tl when ts = group -> true
    | hd :: tl -> walk tl
    | [] -> false
    in walk the_comp.mod_flags
  in

  (* now ask about groups *)
  if hasgroups  && not !always_yes && not !always_uninstall &&
     not !sometimes_reinstall && not (!specified_specific_components)
    then List.iter (fun this_grp ->
      let finished = ref false in
      while not !finished do
        finished := true ;
        log_and_print "\n%s%s%s" (Var.get_string(get_trans (-1028)))
                (Dc.single_string_of_tlk_string game this_grp) (get_trans (-1029)) ;
        match String.uppercase(read_line ()) with
        | "Y" ->
          for i = 0 to last_module_index do
            try
            let the_comp = get_nth_module tp i false in
            if is_my_group the_comp this_grp then module_defaults.(i) <- TP_Ask ;
            with Not_found -> ()
          done
        | "N" -> ()
        | _ -> finished := false
      done ;
  ) !groups ;

  let handle_error_generic always_yes specified_specific_components finished package_name = (fun e ->
    return_value := return_value_error_tp2_component_install ;
    log_and_print "ERROR: %s\n" (Printexc.to_string e) ;
    Dc.clear_state () ;
    (if (!log_file <> "") then
      log_and_print "%s %s %s %s\n" ((get_trans (-1004))) !log_file
                                    (get_trans (-1005)) tp.author) ;
    (* log_and_print "PLEASE email the file %s to %s\n" !log_file tp.author);*)
    if !always_yes or !specified_specific_components then begin
      log_and_print "Automatically Skipping [%s] because of error.\n"
        package_name ;
      finished := true
    end ;
    lang_init ())
  in

  let asked_about_comp = Hashtbl.create 255 in

  let ask_about_module_with_subcomp current m subcomp handle_error = begin
    let subcomp_group_str = Dc.single_string_of_tlk_string game subcomp in
    if Hashtbl.mem asked_about_comp subcomp then
      log_or_print "Already Asked About [%s]\n" subcomp_group_str
    else begin
      Hashtbl.add asked_about_comp subcomp true ;
    let any_already = any_member_of_subcomp_group_installed subcomp in
    let finished = ref false in
    while not !finished do
      try
        let choice_num = ref 1 in
        let choice_ht = Hashtbl.create 255 in
        let already_ht = Hashtbl.create 255 in
        let uninstalled_because_preditate = ref false in
        let is_forced = ref false in
        let at_least_one_OK = ref false in
        for i = 0 to last_module_index do
          try let m = get_nth_module tp i false in
          match subcomp_group m with
          | Some(ts) when ts = subcomp && not(subcomp_predicate m) && already_installed this_tp2_filename i ->
            let can_uninstall = already_installed this_tp2_filename i in
            uninstalled_because_preditate := true ;
            let package_name = Dc.single_string_of_tlk_string
              game m.mod_name in
            log_and_print "Uninstalling the component %s because its predicate is no longer true.\n"
              package_name ;
            handle_letter tp "U" can_uninstall package_name m finished i ;
          | Some(ts) when ts = subcomp ->
              is_forced := !is_forced || (subcomp_forced m);
              at_least_one_OK := !at_least_one_OK || (subcomp_predicate m)
          | _ -> ()
          with Not_found -> ()
        done ;
        let any_already = any_already && not !uninstalled_because_preditate in
        if !at_least_one_OK then begin
          (if (any_already) then begin
            if not !is_forced then
              log_and_print "\n%s%s%s\n" (get_trans (-1006)) subcomp_group_str
                (get_trans (-1025))
            else log_and_print "\n%s%s%s\n"  (get_trans (-1006)) subcomp_group_str
                (get_trans (-1031))
          end else begin
            (if !is_forced then
              log_and_print "\n%s%s%s\n" (get_trans (-1006)) subcomp_group_str
                (get_trans (-1030))
            else
              log_and_print "\n%s%s%s\n" (get_trans (-1006)) subcomp_group_str
                (get_trans (-1026)) )
          end ) ;
          for i = 0 to last_module_index do
            try let m = get_nth_module tp i false in
            match subcomp_group m with
            | Some(ts) when ts = subcomp && (subcomp_predicate m) ->
              let this_subcomp_name = Dc.single_string_of_tlk_string
                game m.mod_name in
              log_and_print "%2d] %s" !choice_num this_subcomp_name ;
              (if already_installed this_tp2_filename i then begin
                log_and_print "%s\n" (get_trans (-1027)) ;
                Hashtbl.add already_ht () (m,i) ;
                Hashtbl.add choice_ht !choice_num (m,i,true)
              end else begin
                log_and_print "\n" ;
                Hashtbl.add choice_ht !choice_num (m,i,false) ;
              end
              );
              incr choice_num
            | _ -> ()
            with Not_found -> ()
          done ;
          let answer = String.uppercase (read_line ()) in
          (match answer with
          | "U" when !is_forced -> finished := false ;
          | "U" | "R" ->
              if Hashtbl.mem already_ht () then begin
                let (m,i) = Hashtbl.find already_ht () in
                let can_uninstall = already_installed this_tp2_filename i in
                let package_name =
                  Dc.single_string_of_tlk_string game m.mod_name in
                handle_letter tp answer can_uninstall
                  package_name m finished i ;
                finished := true
              end
          | "N" when any_already || not !is_forced ->
              finished := true;
          | "Q" ->
            for i = 0 to last_module_index do
              try
                let c = get_nth_module tp i false in
                if subcomp_forced c && not (fails_requirements c) then module_defaults.(i) <- TP_Install
                else module_defaults.(i) <- TP_Skip ;
              with Not_found -> ();
              finished := true
            done
          | _ ->
            begin
              let choice, ok = try (int_of_string answer,true)
                               with _ -> (0,false)
              in
              if (ok) then
                if Hashtbl.mem choice_ht choice then begin
                  let (m,i,already) = Hashtbl.find choice_ht choice in
                  if (already) then
                    finished := true
                  else begin
                    log_or_print "Uninstalling All Other Subcomponents of [%s]\n"
                      subcomp_group_str ;
                    Hashtbl.iter (fun () (m,i) ->
                      let can_uninstall = already_installed
                        this_tp2_filename i in
                      let package_name =
                        Dc.single_string_of_tlk_string game m.mod_name in
                      handle_letter tp "U" can_uninstall
                        package_name m finished i
                    ) already_ht ;
                    log_or_print "Done Uninstalling All Other Subcomponents of [%s]\n" subcomp_group_str ;
                    let can_uninstall = already_installed
                      this_tp2_filename i in
                    let package_name =
                      Dc.single_string_of_tlk_string game m.mod_name in
                    handle_letter tp "Y" can_uninstall
                      package_name m finished i ;
                    finished := true
                  end
                end
            end
          )
        end else finished := true ;
      with e -> handle_error e

    done
    end
  end in

  let ask_about_module current m package_name handle_error = begin
    let finished = ref false in
        while not !finished do try
          let can_uninstall = already_installed this_tp2_filename !current in
            if can_uninstall then
              (* log_and_print "\nInstall Component [%s]\n[R]e-Install, [N]o Change or [U]ninstall or [Q]uit? "  package_name *)
              log_and_print "\n%s%s%s" (get_trans (-1006)) package_name (get_trans (-1007))
            else
             (* log_and_print "\nInstall Component [%s]\n[Y]es or [N]o or [Q]uit? "  package_name ; *)
              log_and_print "\n%s%s%s" (get_trans (-1006)) package_name (get_trans (-1008)) ;
            begin
            let answer = String.uppercase(read_line ()) in
            handle_letter tp answer can_uninstall package_name m
                        finished !current ;
            Dc.clear_state () ;
            end
        with e -> handle_error e
        done
  end in

  print_log () ;

  let original_menu_style () =
    let current = ref (-1) in
    List.iter (fun m ->
      incr current ;
      List.iter (fun f -> match f with
      | TPM_Designated(i) -> current := i ;
      | _ -> ()
      ) m.mod_flags ;
      let def = ref (module_defaults.(!current)) in
      let can_uninstall = already_installed this_tp2_filename !current in
      List.iter (fun f -> match f with
      | TPM_InstallByDefault ->
        if !def = TP_Ask && (not can_uninstall) then def := TP_Install
      | _ -> ()
      ) m.mod_flags ;
      let package_name = Dc.single_string_of_tlk_string game m.mod_name in

      let preproc_fail msg warn already is_forbid_file =
        if (!def <> TP_Skip) && (!def <> TP_Uninstall) then begin
          (* WW: the "FORBID_FILE" bug -- don't uninstall things here! *)
        if (already) then
            def := (if is_forbid_file then TP_Ask else TP_Uninstall)
          else begin
            let warn = Dc.single_string_of_tlk_string game warn in
            log_and_print "\n%s: [%s]\n\t%s\n" msg package_name warn ;
            def := TP_Skip
          end
        end
      in

      List.iter (fun f -> match f with
      | TPM_Deprecated(warn) ->
        begin
          if can_uninstall then begin
            let warn = Dc.single_string_of_tlk_string game warn in
            log_and_print "\nNOTE: [%s] is deprecated. Uninstalling!\n\t%s\n"
              package_name warn ;
            def := TP_Uninstall ;
          end else begin
            def := TP_Skip ;
          end
        end
      | TPM_RequireComponent(s,i,warn) ->
        begin
          if already_installed s i &&
             not (temporarily_uninstalled s i) then
            () (* good! *)
          else preproc_fail "SKIPPING" warn can_uninstall false
        end
      | TPM_ForbidComponent(s,i,warn) ->
        begin
          if already_installed s i &&
             not (temporarily_uninstalled s i) then
            preproc_fail "SKIPPING" warn can_uninstall true
          else
            () (* good! *)
        end
      | TPM_RequirePredicate(p,warn) ->
        begin
          if is_true (eval_pe game p) then
            ()
          else preproc_fail "SKIPPING" warn can_uninstall true
        end
      | TPM_SubComponents(ts,_,_) -> ()
      | TPM_Designated(i) -> () (* handled above *)
      | TPM_InstallByDefault
      | TPM_NotInLog
      | TPM_Group (_)  -> ()
      ) m.mod_flags ;
      List.iter (fun a -> match a with
        | TP_Require_File(file,warn) ->
          begin
            if (bigg_file_exists file game.Load.key) then
              ()
            else preproc_fail "SKIPPING" warn can_uninstall false
          end
        | TP_Forbid_File(file,warn) ->
          begin
            if (bigg_file_exists file game.Load.key) then
              preproc_fail "SKIPPING" warn can_uninstall true
            else
              ()
          end
        | _ -> ()
      ) m.mod_parts ;
      let finished = ref false in
      let handle_error = handle_error_generic always_yes specified_specific_components finished
        package_name in
      Dc.clear_state () ;
      match !def with
      | TP_Install ->
        begin try
          handle_letter tp "Y" can_uninstall package_name m finished !current ;
          with e -> handle_error e
        end
      | TP_Uninstall ->
        begin try
          handle_letter tp "U" can_uninstall package_name m finished !current ;
          with e -> handle_error e
        end
      | TP_Skip ->
            (*
            log_and_print "\nSkipping Component [%s] %d\n"
            (package_name) !current ;
            *)
        begin
          match subcomp_group m with
          | Some(ts) -> Hashtbl.add asked_about_comp ts true  
          | None -> ()
        end
      | TP_Ask ->
        begin
          match subcomp_group m with
          | Some(ts) -> ask_about_module_with_subcomp current m ts handle_error
          | None -> ask_about_module current m package_name handle_error
        end
    ) tp.module_list
  in 
  
  if !specify then begin
    let finished = ref false in
    while not !finished do
      log_and_print "\n%s" (get_trans (-1009)); 
      let line = read_line () in
      if line = "" then 
        finished := true
      else begin
        let reg = Str.regexp_string_case_fold line in 
        let current = ref (-1) in 
        List.iter (fun m ->
          incr current ;
          let rec process lst = match lst with
          | TPM_Designated(i) :: tl -> current := i ;
          | hd :: tl -> process tl
          | [] -> ()
          in process m.mod_flags ;
          let package_name = Dc.single_string_of_tlk_string game m.mod_name in 
          let handle_error = handle_error_generic (ref false) (ref false) (ref true)
            package_name in
          try 
            let _ = Str.search_forward reg package_name 0 in
            ask_about_module current m package_name handle_error
          with Not_found -> () 
        ) tp.module_list
      end
    done 
  end else begin
    original_menu_style () ; 
  end ;

  interactive := false ;

  (* now we must handle every temporarily-uninstalled mods *)
  let re_installed = ref [] in 
  let rec process lst = match lst with
    [] -> []
  | (_,_,_,_,Installed) as head :: tl -> head :: (process tl)
  | (_,_,_,_,Permanently_Uninstalled) as head :: tl -> head :: (process tl)
  | (a,b,c,sopt,Temporarily_Uninstalled) as head :: tl ->
    begin 
    try 
      (* we must re-install it! *)
      begin
        (* log_and_print "\nRe-Installing [%s] component %d %s\n" 
          a c (str_of_str_opt sopt); *)
        log_and_print "\n%s%s%s %d %s\n" 
          (get_trans (-1010)) a (get_trans (-1011)) c 
          (str_of_str_opt sopt); 
        let tp_file = a in 
        let tp2 = handle_tp2_filename tp_file in 
        Load.allow_missing := !Load.allow_missing @ 
          List.fold_left (fun acc elt -> match elt with
            Allow_Missing(lst) -> lst @ acc
              | _ -> acc) [] tp2.flags ; 
        (* load their chosen language *)
        Dc.clear_state () ;
        Dc.push_trans ();
        init_default_strings () ; 
        Var.set_string "TP2_AUTHOR" tp2.author ;
        Var.set_string "TP2_FILE_NAME" tp2.tp_filename;
        (try
          let l = List.nth tp2.languages b in
          our_lang := Some(l) ;
          our_lang_index := b ;
        (*  log_and_print "Re-Installing Using Language [%s]\n" l.lang_name ;*)
          log_and_print "%s [%s]\n" ((get_trans (-1012))) l.lang_name ;
          Var.set_string "LANGUAGE" l.lang_dir_name ;
          Var.set_string "TP2_FILE_NAME" tp2.tp_filename ;
          List.iter handle_tra_filename (List.map Arch.backslash_to_slash l.lang_tra_files)
        with _ -> 
          our_lang := None ; 
          our_lang_index := 0 ;
          () ) ;
        let m = get_nth_module tp2 c true in 
        let package_name = Dc.single_string_of_tlk_string game m.mod_name in 
        let fails_requirements = List.exists (fun f -> match f with
          | TPM_RequireComponent(s,i,warn) ->
            begin
              if already_installed s i then
                false
              else begin
                log_and_print "\n[%s] component %d %s fails component requirements, *not* Re-Installing.\n" a c (str_of_str_opt sopt); 
                true
              end 
            end
          | TPM_ForbidComponent(s,i,warn) ->
            begin
              if not (already_installed s i) then
                false
              else begin
                log_and_print "\n[%s] component %d %s fails component requirements, *not* Re-Installing.\n" a c (str_of_str_opt sopt);
                true
              end
            end
          | TPM_Deprecated(warn) -> begin
              log_and_print "\n[%s] component %d %s is deprecated, *not* Re-Installing.\n" a c (str_of_str_opt sopt);
              true
            end
          | TPM_RequirePredicate(p,warn) ->
            begin
              if is_true (eval_pe game p) then
                false
              else begin
                log_and_print "\n[%s] component %d %s fails component requirements, *not* Re-Installing.\n" a c (str_of_str_opt sopt);
                true
              end
            end
          | _ -> false) m.mod_flags
          ||
          List.exists (fun f -> match f with
            | TP_Require_File(file,warn) ->
              begin
                if (bigg_file_exists file game.Load.key) then
                  false
                else begin
                  log_and_print "\n[%s] component %d %s fails component requirements, *not* Re-Installing.\n" a c (str_of_str_opt sopt);
                  true
                end
              end
            | TP_Forbid_File(file,warn) ->
              begin
                if (bigg_file_exists file game.Load.key) then begin
                  log_and_print "\n[%s] component %d %s fails component requirements, *not* Re-Installing.\n" a c (str_of_str_opt sopt);
                  true
                end
                else
                  false
              end
            | _ -> false
          ) m.mod_parts

        in
        begin
        if fails_requirements then begin
          handle_letter tp2 "U" false package_name m (ref false) c ;
          re_installed := !re_installed @ [(a,b,c,sopt,Permanently_Uninstalled)] ;
        end else begin 
          handle_letter tp2 "R" false package_name m (ref false) c ;
          re_installed := !re_installed @ [(a,b,c,sopt,Installed)] ;
        end
        end ; 
        Dc.clear_state () ; 
        Dc.pop_trans ();
      end ;
      (process tl)  
    with e ->
      log_and_print "ERROR Re-Installing [%s] component %d %s\nTry to re-install it manually.\n%s\n" a c (str_of_str_opt sopt) (Printexc.to_string e) ;
      return_value := return_value_error_tp2_component_install ; 
      (a,b,c,sopt,Permanently_Uninstalled) :: (process tl)
    end 
  in 
  let result = (process !the_log) @ !re_installed in

  the_log := result ; 

  save_log game handle_tp2_filename handle_tra_filename get_tra_list_filename ;

  Load.allow_missing := old_allow_missing ;
  game.Load.script_style <- old_script_style ; 
end
